#include <strings.h>
#include <unistd.h>

#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <sys/types.h>

#include <pkcs15container.h>

#define RND_EXTRA_LEN	15
#define TLEN 100000
#define TNAME1 "p15.np"
#define TNAME2 "p15.p"
#define PASSWD "123"


/*****************************************************************************/
void print_errors(p15_info_t* errs)
{
	int i = 0;

	if (!errs) {
		printf("Success\n");
	}
	while (errs) {
		i++;
		printf("\n");
		printf("%d:\n", i);
		printf("Type: ");
		switch (errs->type) {
		case P15_ERROR:
			printf("Error\n");
			break;
		case P15_WARNING:
			printf("Warning\n");
			break;
		case P15_INTERNAL_ERROR:
			printf("Internal Error\n");
			break;
		default:
			printf("Unknown (%d)\n", errs->type);
		}
		printf("Code: ");
		if (errs->code < 0 || errs->code > P15_RC_MAX) 
			printf("Unknown (%d)\n", errs->code);
		else 
			printf("%s\n", p15_rc_dsc[errs->code]);
		if (errs->text) 
			printf("Details: %s\n", errs->text);
		if (errs->consumed)
			printf("Consumed bytes: %d\n", (int)errs->consumed);
		if (errs->asn_type_name)
			printf("ASN type name: %s\n", errs->asn_type_name);
		errs = errs->next;
	}
}
/*****************************************************************************/
int ckrc(char* func, int rc)
{
	if (!rc) return 0;

	printf("\n%s returned code: ", func);
	if (rc < 0 || rc > P15_RC_MAX) 
		printf("Unknown (%d)\n", rc);
	else 
		printf("%s\n", p15_rc_dsc[rc]);
	return rc;
}
/*****************************************************************************/
int init_random(p15_ctx_t* ctx)
{
	int f = 0;
	uint8_t rnd[64];

	f = open("/dev/urandom", O_RDONLY);
	if (f < 0) {
		printf("Error opening /dev/urandom.\n");
		f = 0;
		return 1;
	}
	if (read(f, rnd, 64) != 64) {
		printf("Error reading /dev/urandom.\n");
		close(f);
		return 1;
	}
	close(f);
	f = 0;
	return ckrc("p15_init_random", p15_init_random(ctx, rnd, rnd + 32));
}
/*****************************************************************************/
int add_cert(p15_ctx_t* ctx, char* certfile, int trusted, p15_id_t* id, 
	char* label)
{
	int f = 0;
	int l;
	uint8_t buf[TLEN];
	int rc;
	p15_info_t* errs = NULL;

	f = open(certfile, O_RDONLY);
	if (f < 0) {
		printf("Cannot read file %s\n", certfile);
		return 1;
	}
	l = read(f, buf, TLEN);
	close(f);

	rc = ckrc("p15_add_cert", p15_add_cert(ctx, trusted, trusted, id, NULL,
		label, buf, l, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc) return 1;
	
	return 0;			
}
/*****************************************************************************/
int add_crl(p15_ctx_t* ctx, char* crlfile, char* label)
{
	int f = 0;
	int l;
	uint8_t buf[TLEN];
	int rc;
	p15_info_t* errs = NULL;

	f = open(crlfile, O_RDONLY);
	if (f < 0) {
		printf("Cannot read file %s\n", crlfile);
		return 1;
	}
	l = read(f, buf, TLEN);
	close(f);

	rc = ckrc("p15_add_crl", p15_add_crl(ctx, NULL, NULL, label, buf, l, 
		0, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc) return 1;
	
	return 0;	
}
/*****************************************************************************/
int add_key(p15_ctx_t* ctx, char* cubafile, int pkuz, char* label, p15_id_t** nid)
{
	int f = 0;
	uint8_t kbuf[36];
	int rc;
	p15_info_t* errs = NULL;

	f = open(cubafile, O_RDONLY);
	if (f < 0) {
		printf("Cannot read file %s\n", cubafile);
		return 1;
	}
	if (read(f, kbuf, 36) != 36) {
		printf("Wrong cuba key format (%s)\n", cubafile);
		close(f);
		return 1;
	}
	close(f);
	rc = ckrc("p15_add_private_key", p15_add_private_key(ctx, NULL, nid, 
		label, 0, P15_KA_DEFAULT, pkuz, NULL, NULL, &errs, 1, kbuf + 4));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc) return 1;
	return 0;
}
/*****************************************************************************/
int write_token(p15_ctx_t* ctx, char* fname, char* pass)
{
	size_t tlen;
	uint8_t token[TLEN];
	int rc;
	p15_info_t* errs = NULL;
	int f;

	tlen = TLEN;
	rc = ckrc("p15_save_token", p15_save_token(ctx, token, &tlen, 
		(uint8_t*)pass, pass ? strlen(pass) : 0, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc > 1) return 1;
	unlink(fname);
	f = open(fname, O_WRONLY | O_CREAT, 0644);
	if (f < 0) {
		printf("Failed to create token file.\n");
		return 1;
	}
	if (write(f, token, tlen) != tlen) {
		printf("Failed to write token.\n");
		return 1;
	}
	close(f);
	return 0;
}
/*****************************************************************************/
int main(int argc, char** argv) {
	int rc = 0;
	p15_ctx_t* ctx = NULL;
	p15_info_t* errs = NULL;
	p15_id_t* id1 = NULL;
	p15_id_t* id2 = NULL;

	ctx = p15_new_ctx();
	if (!ctx) { 
		rc = 1;
		printf("Error creating ctx.\n");
		goto end;
	}
	rc = init_random(ctx);
	if (rc) { rc = 1; goto end; }

	if (add_key(ctx, "d1.key", P15_PKIPAR_CPRO_A, "Key For d1", &id1)) { rc = 1; goto end; }

	if (add_key(ctx, "d2.key", P15_PKIPAR_CPRO_A, NULL, &id2)) { rc = 1; goto end; }
	/* #00000002 */

	rc = add_cert(ctx, "ca.cer", 1, NULL, "Удостоверяющий центр"); 
	if (!rc) rc = add_cert(ctx, "d1.cer", 0, id1, NULL); /* :00000001 */
	if (!rc) rc = add_cert(ctx, "d2.cer", 0, id2, "cert for d2"); 
	if (rc) goto end;

	rc = ckrc("p15_set_private_key_subject", p15_set_pkey_subject(ctx, id1, NULL));
	if (!rc) rc = ckrc("p15_set_private_key_subject", 
		p15_set_pkey_subject(ctx, id2, NULL));
	if (rc) goto end; 
	rc = add_crl(ctx, "crl.old", "Bad Crl");
	if (rc) goto end;

	if (write_token(ctx, TNAME1, NULL)) { rc = 1; goto end; }
	if (write_token(ctx, TNAME2, PASSWD)) { rc = 1; goto end; }


end:
	if (errs) 
		print_errors(errs);
	p15_free_infos(&errs);
	p15_free_ctx(&ctx);
	p15_free_id(&id1);
	p15_free_id(&id2);
	if (rc) 
		printf("TEST FAILED\n");
	else
		printf("TEST PASSED\n");
	return rc;
}

