#if defined(WIN32) || defined(WIN64)
#include <io.h>
#else
#include <strings.h>
#include <unistd.h>

#endif

#include <stdarg.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include "p15_mem.h"
#include "pkcs15container.h"
#include "p15_internal.h"
#include "gostc_sym.h"
#include "gostc_sym_uz.h"
#include "gostc_asym_mask.h"
#include "gostc_hash.h"
#include "gostc_pkcs5.h"
#include "gostc_sym_keywrap.h"
#include "P15-AuthenticatedData.h"
#include "P15-PKCS15Token.h"
#include "P15-CRLContainer.h"
#include "P15-Gost28147-89-KeyWrapParameters.h"
#include "P15-Gost28147-89-EncryptedKey.h"
#include "P15-Gost28147-89-Parameters.h"
#include "P15-RandomInitValue.h"
#include "P15-FactorTSVersion.h"
#include "P15-PBKDF2-params.h"
#include "P15-GostPrivateKey.h"
#include "P15-GostSecretKeyValue.h"



#define URANDOM	"/dev/urandom"

#define P15_OIT_ANY	-1
#define P15_OIT_UNKNOWN	0
#define P15_OIT_CERT	1
#define P15_OIT_RCERT	2
#define P15_OIT_PRKEY	3
#define P15_OIT_DATA	4
#define P15_OIT_PUBKEY	5
#define P15_OIT_SECKEY	6



typedef struct _p15_obj_it_t p15_obj_it_t;
struct _p15_obj_it_t {
	int type; /* P15_OIT_ const */
	int st; /* 0-at start, 1-at object, 2-at end */
	int objsi; /* index of current PKCS15Objects */
	int obji; /* index of current object in PathOrObject */
};


static void p15_free_ctx_content(p15_ctx_t* ctx);
static int p15_add_object(p15_ctx_t* ctx, int type, void* obj);
static p15_info_t* p15_new_info(int type, int rc, char* text);
static p15_info_t* p15_add_error(p15_info_t **errors, int type, 
	int rc, char* text);

static void p15_add_error_type(p15_info_t* info, asn_enc_rval_t erv);
static int p15_consume_bytes(const void *buffer, size_t size, void *appkey);
static p15_der_t* p15_new_der_from_sbuf(p15_sbuf_t* sbuf);
static int p15_name_to_rdn(P15_Name_t* name, p15_der_t** rdn);

static int p15_set_node_id(p15_ctx_t* ctx, p15_id_t* nodeid, p15_id_t* id, 
	p15_id_t** newid);

static void p15_refine_id(p15_id_t*);
static int p15_id_to_ostr(p15_id_t* id, OCTET_STRING_t* ostr);

static int p15_ck_id_exist(p15_ctx_t* ctx, p15_id_t* id);
/* 0-not exist(ok), 1-exists(bad), >1 - P15_RC_CONST */

static int p15_process_obj_id(p15_ctx_t* ctx, P15_Identifier_t* asnid, int objtype, 
	p15_id_t* objid);
/* ret: inv args, cert_id/crl_id/key_id/id in use, inv token, int err 
 * objtype - P15_OT_ const
 */

static int p15_mk_kekri(p15_ctx_t* ctx, P15_RecipientInfo_t** pri, 
	uint8_t **pkey, uint8_t **pkmsk, uint8_t* pwkey, uint8_t* pwmsk,
	p15_info_t** errors);

static int p15_encuz_to_oid(int uz, OBJECT_IDENTIFIER_t* oid);
static int p15_pkipar_to_oid(int uz, OBJECT_IDENTIFIER_t* oid);

static int p15_oid_to_encuz(OBJECT_IDENTIFIER_t* oid, int* uz);
/* ret 0-ok, P15_RC_UNSUPPORTED_ALGORITHM, P15_RC_INTERNAL_ERROR */

static int p15_oid_to_pkipar(OBJECT_IDENTIFIER_t* oid, int* uz, int only2001);
/* ret 0-ok, P15_RC_UNSUPPORTED_ALGORITHM, P15_RC_INTERNAL_ERROR */

static int p15_str_to_oid(const char* str, OBJECT_IDENTIFIER_t* oid);

static int p15_oid_str_eq(const char* str, OBJECT_IDENTIFIER_t* oid);
/* ret 0-eq, 1-not eq, >1-P15_RC_ const */

static int p15_oid_eq(OBJECT_IDENTIFIER_t* o1, OBJECT_IDENTIFIER_t* o2);
/* ret 0-eq, 1-not eq, >1-P15_RC_ const */

static p15_oid_t* p15_new_oid_from_asn(OBJECT_IDENTIFIER_t* aoid);

static int p15_get_pwkey(p15_ctx_t* ctx, uint8_t* pw, int pwlen, 
	uint8_t** pwkey, uint8_t** pwkmsk, uint8_t** salt);

static int p15_mk_enveloped_data(p15_ctx_t* ctx, P15_EnvelopedData_t* edata,
	uint8_t* data, int len, uint8_t* pwkey, uint8_t* pwkmsk, 
	p15_info_t** errors);

static int p15_mk_rng_do(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, 
	uint8_t* pwkey, uint8_t* pwkmsk, p15_info_t** errors);

static int p15_mk_factor_version(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, 
	p15_info_t** errors);

static int p15_mk_encap_token(p15_ctx_t* ctx, uint8_t* pwkey, uint8_t* pwkmsk, 
	uint8_t* salt, OCTET_STRING_t** petoken, uint8_t* tokenhash, int* warns,
	p15_info_t** errors);
/* ret: int err, P15_RC_NO_PRIVATE_KEY_DATA, P15_RC_NO_OBJECT_DATA, crypto2001remask */

static int p15_get_safe_random(p15_ctx_t* ctx, uint8_t* rnd, int len);

static int p15_asn_encode(asn_TYPE_descriptor_t* td, void* el, 
	uint8_t** ebuf, size_t* ebuflen, p15_info_t** errors);

static int p15_asn_decode(uint8_t* der, size_t len, asn_TYPE_descriptor_t* td,
	void** el, p15_info_t** errors);
/* ret INTERNAL_ERROR or other errors if not internal. */

static int p15_load_kmi(p15_ctx_t* ctx, P15_KeyManagementInfo_t* kmis, 
	uint8_t* password, int passlen, p15_id_t** kekid,
	uint8_t** pwkey, uint8_t** pwkmsk, p15_info_t** errors);

static int p15_load_kekris(p15_ctx_t* ctx, const char* upname,
	P15_RecipientInfos_t* ris, p15_id_t* kekid, 
	uint8_t* pwkey, uint8_t* pwkmsk, 
	uint8_t** akey, uint8_t** akmsk,
	int set_enc_uz,
	p15_info_t** errors);
/* ret 0, 1-warns, P15_RC_WRONG_PASSWORD, P15_RC_INTERNAL_ERROR, P15_RC_NO_AUTH_AND_DECRYPT,
 * other not important errors
 */

static int p15_ck_ANY_NULL(ANY_t* any);
/* 0-ANY=NULL, 1-not NULL, >1 - err */

static int p15_load_aattrs(int hash2012, P15_AuthAttributes_t* aattrs, 
	uint8_t* hash, p15_info_t** errors);
/* ret 0, 1-warns, int err, no auth, other errs (terr) */

static int p15_load_aattr(int hash2012, P15_Attribute_t* attr, int* oidfound, 
	int* hashfound, uint8_t* hash, int* ign, int* terr, p15_info_t** errors);
/* ret !0-internal error */

static p15_obj_it_t* p15_new_obj_it(int type);
static void p15_free_obj_it(p15_obj_it_t** pit);
static int p15_obj_iterate(p15_obj_it_t* it, P15_PKCS15Token_t* token, void** obj, 
	int* type, int* warns, p15_info_t** errors);
/* ret: 0-ok (object returned), 1-no objects left, >1-internal error 
 * when it->type==P15_OIT_ANY --> actual type is returned to *type 
 */

static void* p15_oit_get_obj(P15_PKCS15Objects_t* os, int oi, int* type,
	int* warns, p15_info_t** errors);

static int p15_oit_ck_type(int type, P15_PKCS15Objects_t* os);
/* ret: 0-good, 1-bad, 2-unknown */

static int p15_iterate_data_object(void* obj,
	P15_Label_t** label, uint8_t* flags, 
	OBJECT_IDENTIFIER_t** appoid, OBJECT_IDENTIFIER_t** dataoid,
	ANY_t** oany, P15_EnvelopedData_t** edata, int* warns, int* errs,
	p15_info_t** errors);
/* ret: 0-found, 1-continue, int err 
 * Returned data should NOT be freed 
 * label and flags may be NULL, oany ^^ edata == true
 */

static int p15_decrypt_enveloped_data(p15_ctx_t* ctx, p15_id_t* kekid, 
	const char* upname, P15_EnvelopedData_t* edata, 
	uint8_t** data,	int* len, uint8_t* pwkey, uint8_t* pwkmsk, 
	int* warns, int* errs, 	p15_info_t** errors);
/* ret: 0-ok, int err, no auth */

static int p15_get_random_container(p15_ctx_t* ctx, P15_PKCS15Token_t* token, 
	p15_id_t* kekid, uint8_t* pwkey, uint8_t* pwkmsk,
	P15_RandomInitValue_t** rndiv, int* warns, int* errs, p15_info_t** errors);
/* ret: 0-ok, int err, P15_RC_NO_RANDOM_CONTAINER, no auth */

static int p15_process_data_object(p15_ctx_t* ctx, P15_Label_t* label, uint8_t flags, 
	OBJECT_IDENTIFIER_t* appoid, OBJECT_IDENTIFIER_t* dataoid,
	ANY_t* oany, P15_EnvelopedData_t* edata, 
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid,
	int* warns, int* errs, p15_info_t** errors);
/* ret: 0-ok, int_err, no_auth */

static int p15_save_object(p15_ctx_t* ctx, p15_object_t* object, 
	P15_PKCS15Token_t* token, uint8_t* pwkey, uint8_t* pwkmsk, 
	int* warns, p15_info_t** errors);
/* ret: 0-ok, int err, no key data, no obj data, crypto2001mask */

static int p15_load_objects(p15_ctx_t* ctx, P15_PKCS15Token_t* ptoken, 
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid, int* warns, int* errs, 
	p15_info_t** errors);
/* ret: 0-ok, no_auth, int_err */

static int p15_utctime_to_gt(GeneralizedTime_t* gt, p15_utctime_t* in);

static int p15_gt_to_utctime(GeneralizedTime_t* gt, p15_utctime_t** out);
/* ret: int err, inv token */

static int p15_mk_prkey_obj(p15_ctx_t* ctx, P15_PrivateKeys_t* prkeys, 
	p15_private_key_t* pk, uint8_t* pwkey, uint8_t* pwkmsk, int* warns, 
	p15_info_t** errors);
/* ret int err, P15_RC_NO_PRIVATE_KEY_DATA, P15_RC_CRYPTOLIB_MASK2001 */

static int p15_mk_seckey_obj(p15_ctx_t* ctx, P15_SecretKeys_t* skeys,
	p15_secret_key_t* sk, uint8_t* pwkey, uint8_t* pwkmsk, int* warns,
	p15_info_t** errors);
/* ret int err, P15_RC_NO_SECRET_KEY_DATA */

static int p15_mk_pubkey_obj(p15_ctx_t* ctx, P15_PublicKeys_t* pubkeys, 
	p15_public_key_t* pbk, int* warns, p15_info_t** errors);
/* ret: int err */

static int p15_mk_oiddo_obj(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, 
	p15_oiddo_t* oiddo, uint8_t* pwkey, uint8_t* pwkmsk, 
	p15_info_t** errors);
/* ret: 0, int err, no obj data */

static int p15_process_private_key_object(p15_ctx_t* ctx, void* pkobj,
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid, int* warns, int* errs, 
	p15_info_t** errors);
/* ret: int err, no auth */

static int p15_process_secret_key_object(p15_ctx_t* ctx, void* skobj,
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid, int* warns, int* errs, 
	p15_info_t** errors);
/*ret: int err, no auth */

static int p15_process_public_key_object(p15_ctx_t* ctx, void* pkobj,
	int* warns, int* errs, p15_info_t** errors);
/* ret: int err */

static int p15_process_key_params01(p15_ctx_t* ctx, char* parent, 
	P15_Gost01KeyParameters_t* pars, int* keypar, 
	int* warns, int* errs, p15_info_t** errors);
/* ret: int err, unsupported alg, inv token */

static int p15_process_key_params12(p15_ctx_t* ctx, char* parent, 
	P15_Gost12KeyParameters_t* pars, int* keypar, 
	int* warns, int* errs, p15_info_t** errors);
/* ret: int err, unsupported alg, inv token */

static int p15_mk_cert_obj(p15_ctx_t* ctx, P15_Certificates_t* certs, 
	p15_certificate_t* cer, int* warns, p15_info_t** errors);
/* ret: int err */

static int p15_process_cert_object(p15_ctx_t* ctx, void* cobj, int trusted, 
	int* warns, int* errs, p15_info_t** errors);
/* ret: int err */

static int p15_mk_crl_obj(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, p15_crl_t* crl, 
	p15_info_t **errors);
/* ret: int err */

static int p15_process_crl_object(p15_ctx_t* ctx, P15_Label_t* label, uint8_t flags, 
	ANY_t* oany, P15_EnvelopedData_t* edata, int* warns, int* errs, 
		p15_info_t** errors);
/* ret: int err */

static int p15_process_ftsver_object(p15_ctx_t* ctx, uint8_t flags, 
	ANY_t* oany, int* warns, int* errs, p15_info_t** errors);
/* ret: int err */

static int p15_process_oiddo_object(p15_ctx_t* ctx, P15_Label_t* label, 
	uint8_t flags, OBJECT_IDENTIFIER_t* id, ANY_t* oany,
	P15_EnvelopedData_t* edata, uint8_t* pwkey, uint8_t* pwkmsk,
	p15_id_t* kekid, int* warns, int* errs, p15_info_t** errors);
/* ret: int err, no auth */

static int p15_get_mask_uz(int key_params);
/* ret: -1 if err */

static int p15_pack_bitstring(BIT_STRING_t* bstr);
/* ret: int err */

static int p15_unpack_bitstring(BIT_STRING_t* bstr, int size, int ubits);
/* ret: int err, inv bit str*/

static int p15_pkey_is_2012(int keypar);
/* ret: true/false */

static int p15_ck_format(p15_ctx_t* ctx, int is2012, p15_info_t** errors);
/* ret: 0, inv_cont_format, int_err */

static int p15_remask_setup_prkey(p15_ctx_t* ctx, p15_private_key_t* prkey, 
	int to_remask, int mask_count, p15_info_t** errors);
static int p15_remask_setup_seckey(p15_ctx_t* ctx, p15_secret_key_t* seckey, 
	int to_remask, int mask_count, p15_info_t** errors);



char p15_rc_dsc[P15_RC_MAX + 1][500] = {
/* P15_RC_OK */
	"Success",
/* P15_RC_WARNINGS */
	"Warnings exist",

	"Abstract error",
/* P15_RC_INTERNAL_ERROR */
	"Internal error",

	"Reserved 4",
	"Reserved 5",
	"Reserved 6",
	"Reserved 7",
	"Reserved 8",
	"Reserved 9",

/* P15_RC_WRONG_PASSWORD */
	"Authentication failed. Maybe wrong password. Encrypted content skipped",
/* P15_RC_INVALID_ARGS */
	"Invalid arguments to a function",
/* P15_RC_CERT_DECODING_FAILED */
	"Parsing of X.509 certificate failed",
/* P15_RC_INVALID_ISSUER_RDN */
	"Parsing of Issuer distinguished name failed",
/* P15_RC_CERT_NOT_FOUND */
	"Certificate not found",
/* P15_RC_CERT_ID_IN_USE */
	"A certificate with the same ID already exists",
/* P15_RC_INVALID_SUBJECT_RDN */
	"Parsing of Subject distinguished name failed",
/* P15_RC_PRIVATE_KEY_NOT_FOUND */
	"Private key not found",
/* P15_RC_PRIVATE_KEY_ID_IN_USE */
	"A private key with the same ID already exists",
/* P15_RC_CRL_ID_IN_USE */
	"A CRL with the same ID already exists",
/* P15_RC_CRL_DECODING_FAILED */
	"Parsing of CRL failed",
/* P15_RC_CRL_NOT_FOUND */
	"CRL not found",
/* P15_RC_RNG_NOT_INITIALIZED */
	"Random generator is not initialized",
/* P15_RC_LABEL_TOO_LONG */
	"Label is too long. Must be less 256 chars",
/* P15_RC_ASN1C_ERROR */
	"ASN compiler function returned error",
/* P15_RC_INVALID_OID_STR */
	"Invalid OID string format. Must be N.N.NNNN...",
/* P15_RC_PKCS5_FAILED */
	"Failed to get key from password. PBKDF2 failed",
/* P15_RC_CRYPTOLIB_IMIT */
	"Crypto library failed for MAC(imit) generation",
/* P15_RC_BUFFER_TOO_SMALL */
	"Given buffer is too small",
/* P15_RC_DER_DECODING_FAILED */
	"Given DER cannot be decoded according given ASN",
/* P15_RC_DER_NOT_COMPLETE */
	"Given DER is not complete. More data needed",
/* P15_RC_INVALID_TOKEN */
	"Failed to parse some parts of PKCS15 software token",
/* P15_RC_IGNORED_STRUCTURE */
	"Ignoring part of the token",
/* P15_RC_URANDOM_ERROR */
	"Error reading /dev/urandom device",
/* P15_RC_INVALID_PBKDF2_PARAMS */
	"Invalid parameters for PBKDF2 algorithm",
/* P15_RC_NO_PWINFO */
	"No PasswordInfo structure found",
/* P15_RC_NO_AUTH_AND_DECRYPT */
	"Authentication/decryption failed. Token is corrupted. Some encrypted data skipped",
/* P15_RC_UNSUPPORTED_ALGORITHM */
	"Unsupported algorithm",
/* P15_RC_NO_RANDOM_CONTAINER */
	"No appropriate random init value container found in the token",
/* P15_RC_INV_OBJ_FLAGS */
	"A token object has invalid flags",
/* P15_RC_INV_FACTOR_TS_VERSION */
	"Unknown Factor-TS token version detected", 
/* P15_RC_NO_SUBJECT_FOR_PR_KEY */
	"No subject specified for private key",
/* P15_RC_NO_PRIVATE_KEY_DATA */
	"Key/mask data is absent in a private key object. This may happen because of decoding token with wrong password",
/* P15_RC_INVALID_OBJECT */
	"Invalid p15_object_t structure",
/* P15_RC_KEY_NOT_FOUND */
	"Neither private nor public key found",
/* P15_RC_PUBLIC_KEY_ID_IN_USE */
	"A public key with the same ID already exists",
/* P15_RC_PUBLIC_KEY_NOT_FOUND */
	"Public key not found",
/* P15_RC_NO_SUBJECT_FOR_PUB_KEY */
	"No subject specified for public key",
/* P15_RC_NO_OBJECT_DATA */
	"Object data is absent. This may happen because of decoding token with wrong password",
/* P15_RC_CRYPTOLIB_MASK */
	"GOST R 34.10-2001/2012 key-masking function returned error. Maybe invalid private key",
/* P15_RC_INVALID_BIT_STRING */
	"Invalid coding of BIT STRING object (length/unused_bits)",
/* P15_RC_DUPLICATE_CERT */
	"Refusing to add a duplicate certificate",
/* P15_RC_DUPLICATE_CRL	*/
	"Refusing to add a duplicate CRL",
/* P15_RC_2012_KEYS_EXIST */
	"Cannot set P15-2001 container format, because GOST R 34.10-2012 and/or GOST 28147-89 keys exist in ctx",
/* P15_RC_2012_KEY_VS_2001_FORMAT */
	"Cannot add GOST R 34.10-2012 or GOST 28147-89 key to ctx, because of P15-2001 container format forced",
/* P15_RC_CRYPTOLIB_HASH */
	"GOST R 34.11-2012 hash function returned error. Internal error",
/* P15_RC_INVALID_CONTAINER_FORMAT */
	"P15-2001(2012) constructs met in P15-2012(2001) container",
/* P15_RC_ID_IN_USE */
	"The ID is not unique. An object with this ID already exists in container",
/* P15_RC_SECRET_KEY_NOT_FOUND */
	"Secret key not found",
/* P15_RC_89UZ_Z_VS_2001_FORMAT */
	"GOST 28147-89 S-Box Z not allowed in P15-2001 containers"
};


/****************************************************************************/
#ifdef TESTOUT
void p15_dump(void* ptr, size_t len)
{
	size_t i, j, n;
	unsigned char c, *buf;
	unsigned int cc;
	char asc[100];

	j = 0;
	bzero(asc, sizeof(asc));
	buf = ptr;
	for (i = 0; i < len; i++) {
		if (j && !(j & 3)) printf(" ");
		if (j > 15) {
			asc[j] = 0;
			j = 0;
			printf("|%s|\n", asc);
			bzero(asc, sizeof(asc));
		}
		c = buf[i];
		cc = c;
		printf("%.2X ", cc);
		if (c > 31 && c < 127) asc[j] = (char)c;
		else asc[j] = '.';
		j++;
	}
	if (j) {
		n = 4 - ((j - 1) / 4);
		for (i = 0; i < n; i++) printf(" "); 
	}
	if (j && (j < 16)) {
		for (i = j; i < 16; i++) {
			printf("   ");
			asc[i] = ' ';
		}
		asc[16] = 0;
	}
	printf("|%s|\n", asc);
}
/*--------------------------------------------------------------------------*/
void p15_dump_mkey32(void* key, void* msk)
{
	uint8_t b[32];
	uint8_t* k = key;
	uint8_t* m = msk;
	int i;

	for (i = 0; i < 32; i++) b[i] = k[i] ^ m[i];
	p15_dump(b, 32);
}
/*--------------------------------------------------------------------------*/
void p15_dump_id(p15_id_t* id)
{
	OCTET_STRING_t* ostr = NULL;

	ostr = p15_malloc(sizeof(OCTET_STRING_t));
	p15_id_to_ostr(id, ostr);
	p15_dump(ostr->buf, ostr->size);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, ostr);
}
/*--------------------------------------------------------------------------*/
void p15_print_key_usage(uint16_t ku)
{
	uint16_t ku_all = P15_KU_ENCRYPT |
		P15_KU_DECRYPT |
		P15_KU_SIGN |
		P15_KU_SIGN_RECOVER |
		P15_KU_WRAP |
		P15_KU_UNWRAP |
		P15_KU_VERIFY |
		P15_KU_VERIFY_RECOVER |
		P15_KU_DERIVE |
		P15_KU_NON_REPUDIATION;

	if (ku & P15_KU_DECRYPT) printf("DECRYPT ");
	if (ku & P15_KU_SIGN) printf("SIGN ");
	if (ku & P15_KU_SIGN_RECOVER) printf("SIGN_RECOVER ");
	if (ku & P15_KU_WRAP) printf("WRAP ");
	if (ku & P15_KU_UNWRAP) printf("UNWRAP ");
	if (ku & P15_KU_VERIFY) printf("VERIFY ");
	if (ku & P15_KU_VERIFY_RECOVER) printf("VERIFY_RECOVER ");
	if (ku & P15_KU_DERIVE) printf("DERIVE ");
	if (ku & P15_KU_NON_REPUDIATION) printf("NON_REPUDIATION ");
	if (ku & ~ku_all) printf("UNKNOWN ");
}
/*--------------------------------------------------------------------------*/
void p15_print_key_access(uint8_t ka)
{
	int ka_all = P15_KA_MASK | P15_KA_PRESENT;

	if (ka == P15_KA_ABSENT) { printf("ABSENT"); return; }
	if (ka & P15_KA_PRESENT) printf("PRESENT "); 
	if (ka & P15_KA_SENSITIVE) printf("SENSITIVE ");
	if (ka & P15_KA_EXTRACTABLE) printf("EXTRACTABLE ");
	if (ka & P15_KA_ALWAYSSENSITIVE) printf("ALWAYSSENSITIVE ");
	if (ka & P15_KA_NEVEREXTRACTABLE) printf("NEVEREXTRACTABLE ");
	if (ka & P15_KA_LOCAL) printf("LOCAL ");
	if (ka & ~ka_all) printf("UNKNOWN ");
}
#endif
/****************************************************************************/
static int p15_oid_eq(OBJECT_IDENTIFIER_t* o1, OBJECT_IDENTIFIER_t* o2)
/* ret 0-eq, 1-not eq, >1-P15_RC_ const */
{
	if (!o1 || !o2) return P15_RC_INVALID_ARGS;

	if (o1->size != o2->size) return 1;
	if (memcmp(o1->buf, o2->buf, o1->size)) return 1;
	return 0;
}
/****************************************************************************/
static int p15_oid_str_eq(const char* str, OBJECT_IDENTIFIER_t* oid)
{
	OBJECT_IDENTIFIER_t* noid = NULL;
	int rc = 0;

	if (!str || !oid) return P15_RC_INVALID_ARGS;
	
	noid = p15_malloc(sizeof(OBJECT_IDENTIFIER_t));
	rc = p15_str_to_oid(str, noid);
	if (rc) goto end;

	if (oid->size != noid->size) { rc = 1; goto end; }
	if (!oid->buf || !noid->buf) { rc = 1; goto end; }
	if (memcmp(oid->buf, noid->buf, oid->size)) { rc = 1; goto end; }
end:
	ASN_STRUCT_FREE(asn_DEF_OBJECT_IDENTIFIER, noid);
	return rc;
}
/****************************************************************************/
p15_oid_t* p15_new_oid_from_str(const char* str)
{
	int l, dots, i, narcs, j;
	p15_oid_t* obj = NULL;
	int rc = 1;
	char c;
	uint32_t n;

	if (!str) return NULL;

	/* count dots */
	l = strlen(str);
	dots = 0;
	for (i = 0; i < l; i++) 
		if (str[i] == '.') dots++;
	if (dots < 1) goto end;
	narcs = dots + 1;
	obj = p15_malloc(sizeof(p15_oid_t));
	obj->arcs = p15_malloc(narcs * sizeof(uint32_t));
	obj->arcnum = narcs;
	i = 0;
	j = 0;
	n = 0;
	while (j < l) {
		c = str[j];
		j++;
		if (c == '.') {
			if (i >= narcs) goto end;
			obj->arcs[i] = n;
			n = 0;
			i++;
		} else {
			c -= 0x30;
			if (c < 0 || c > 9) goto end;
			n = n * 10 + c;
		}
	}
	if (i != narcs - 1) goto end;
	obj->arcs[i] = n;
	rc = 0;
end:
	if (rc) p15_free_oid(&obj);
	return obj;
}
/****************************************************************************/
p15_oid_t* p15_new_oid_from_arcs(uint32_t* arcs, int num)
{
	p15_oid_t* obj = NULL;
	int i;

	if (!arcs || num < 1) return NULL;

	obj = p15_malloc(sizeof(p15_oid_t));
	obj->arcnum = num;
	obj->arcs = p15_malloc(num * sizeof(uint32_t));
	for (i = 0; i < num; i++)
		obj->arcs[i] = arcs[i];
	return obj;
}
/****************************************************************************/
static p15_oid_t* p15_new_oid_from_asn(OBJECT_IDENTIFIER_t* aoid)
{
	int num;
	uint32_t tmp;
	p15_oid_t* oid;

	if (!aoid) return NULL;
	num = OBJECT_IDENTIFIER_get_arcs(aoid, &tmp, 4, 1);
	if (num < 1) return NULL;
	oid = p15_malloc(sizeof(p15_oid_t));
	oid->arcnum = num;
	oid->arcs = p15_malloc(num * sizeof(uint32_t));
	if (OBJECT_IDENTIFIER_get_arcs(aoid, oid->arcs, 4, num) != num) {
		p15_free_oid(&oid);
		return NULL;
	}
	return oid;
}
/****************************************************************************/
p15_oid_t* p15_copy_oid(p15_oid_t* oid)
{
	if (!oid) return NULL;
	return p15_new_oid_from_arcs(oid->arcs, oid->arcnum);
}
/****************************************************************************/
void p15_free_oid(p15_oid_t** poid)
{
	p15_oid_t* oid;

	if (!poid) return;
	oid = *poid;
	if (!oid) return;
	p15_free(oid->arcs);
	p15_free(oid);
	*poid = NULL;
}
/****************************************************************************/
int p15_oid_cmp(p15_oid_t* oid1, p15_oid_t* oid2)
{
	int i;

	if (!oid1 || !oid2) return P15_RC_INVALID_ARGS;
	if (oid1->arcnum < 1 || oid2->arcnum < 1 || !oid1->arcs ||
		!oid2->arcs) return P15_RC_INVALID_ARGS;
	if (oid1->arcnum != oid2->arcnum) return 1;
	for (i = 0; i < oid1->arcnum; i++) {
		if (oid1->arcs[i] != oid2->arcs[i]) return 1;
	}
	return 0;
}
/****************************************************************************/
int p15_oid_to_str(p15_oid_t* oid, char* str, int maxlen)
{
	int i;
	char part[20];

	if (!oid || !str || maxlen < 1) return P15_RC_INVALID_ARGS;
	if (!oid->arcs || oid->arcnum < 1) return P15_RC_INVALID_ARGS;
	str[0] = 0;
	for (i = 0; i < oid->arcnum; i++) {
		sprintf(part, "%s%u", i ? "." : "", oid->arcs[i]);
		if (strlen(str) + strlen(part) + 1 > (unsigned int)maxlen) {
			str[0] = 0;
			return P15_RC_BUFFER_TOO_SMALL;
		}
		strcat(str, part);
	}
	return 0;
}
/****************************************************************************/
static int p15_encuz_to_oid(int uz, OBJECT_IDENTIFIER_t* oid)
{
	int rc;

	if (!oid) return P15_RC_INVALID_ARGS;

	switch (uz) {
	case P15_89UZ_CPRO_A:
		rc = p15_str_to_oid(P15_OID_ENCR_CPRO_A, oid);
		break;

	case P15_89UZ_CPRO_B:
		rc = p15_str_to_oid(P15_OID_ENCR_CPRO_B, oid);
		break;

	case P15_89UZ_CPRO_C:
		rc = p15_str_to_oid(P15_OID_ENCR_CPRO_C, oid);
		break;

	case P15_89UZ_CPRO_D: 
		rc = p15_str_to_oid(P15_OID_ENCR_CPRO_D, oid);
		break;

	case P15_89UZ_CPRO_Z:
		rc = p15_str_to_oid(P15_OID_ENCR_CPRO_Z, oid);
		break;

	default: 
		return P15_RC_INVALID_ARGS;
	}
	if (rc == P15_RC_ASN1C_ERROR) return rc;
	if (rc) return P15_RC_INTERNAL_ERROR;
	return 0;
}
/****************************************************************************/
static int p15_pkipar_to_oid(int uz, OBJECT_IDENTIFIER_t* oid)
{
	int rc;

	if (!oid) return P15_RC_INVALID_ARGS;

	switch (uz) {
	case P15_PKIPAR_CPRO_A:
		rc = p15_str_to_oid(P15_OID_PKI_CPRO_A, oid);
		break;

	case P15_PKIPAR_CPRO_B:
		rc = p15_str_to_oid(P15_OID_PKI_CPRO_B, oid);
		break;

	case P15_PKIPAR_CPRO_C:
		rc = p15_str_to_oid(P15_OID_PKI_CPRO_C, oid);
		break;

	case P15_PKIPAR_CPRO_XCHA:
		rc = p15_str_to_oid(P15_OID_PKI_CPRO_XCHA, oid);
		break;

	case P15_PKIPAR_CPRO_XCHB:
		rc = p15_str_to_oid(P15_OID_PKI_CPRO_XCHB, oid);
		break;

	case P15_PKIPAR_TC26_512_TEST:
		rc = p15_str_to_oid(P15_OID_PKI_TC26_TEST, oid);
		break;

	case P15_PKIPAR_TC26_512_A:
		rc = p15_str_to_oid(P15_OID_PKI_TC26_A, oid);
		break;

	case P15_PKIPAR_TC26_512_B:
		rc = p15_str_to_oid(P15_OID_PKI_TC26_B, oid);
		break;
	
	default: 
		return P15_RC_INVALID_ARGS;
	}
	if (rc == P15_RC_ASN1C_ERROR) return rc;
	if (rc) return P15_RC_INTERNAL_ERROR;
	return 0;
}
/****************************************************************************/
static int p15_oid_to_encuz(OBJECT_IDENTIFIER_t* oid, int* uz)
/* ret 0-ok, P15_RC_UNSUPPORTED_ALGORITHM, P15_RC_INTERNAL_ERROR */
{
	int rc = 0;

	if (!oid || !uz) return P15_RC_INVALID_ARGS;

	rc = p15_oid_str_eq(P15_OID_ENCR_CPRO_A, oid);
	if (!rc) { *uz = P15_89UZ_CPRO_A; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_ENCR_CPRO_B, oid);
	if (!rc) { *uz = P15_89UZ_CPRO_B; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_ENCR_CPRO_C, oid);
	if (!rc) { *uz = P15_89UZ_CPRO_C; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_ENCR_CPRO_D, oid);
	if (!rc) { *uz = P15_89UZ_CPRO_D; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_ENCR_CPRO_Z, oid);
	if (!rc) { *uz = P15_89UZ_CPRO_Z; return 0; }

	if (rc == 1) return P15_RC_UNSUPPORTED_ALGORITHM;

	return P15_RC_INTERNAL_ERROR;
}
/****************************************************************************/
static int p15_oid_to_pkipar(OBJECT_IDENTIFIER_t* oid, int* uz, int only2001)
/* ret 0-ok, P15_RC_UNSUPPORTED_ALGORITHM, P15_RC_INTERNAL_ERROR */
{
	int rc = 0;

	if (!oid || !uz) return P15_RC_INVALID_ARGS;

	rc = p15_oid_str_eq(P15_OID_PKI_CPRO_A, oid);
	if (!rc) { *uz = P15_PKIPAR_CPRO_A; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_CPRO_B, oid);
	if (!rc) { *uz = P15_PKIPAR_CPRO_B; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_CPRO_C, oid);
	if (!rc) { *uz = P15_PKIPAR_CPRO_C; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_CPRO_XCHA, oid);
	if (!rc) { *uz = P15_PKIPAR_CPRO_XCHA; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_CPRO_XCHB, oid);
	if (!rc) { *uz = P15_PKIPAR_CPRO_XCHB; return 0; }

	if ((rc == 1) && only2001) return P15_RC_UNSUPPORTED_ALGORITHM;

	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_TC26_TEST, oid);
	if (!rc) { *uz = P15_PKIPAR_TC26_512_TEST; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_TC26_A, oid);
	if (!rc) { *uz = P15_PKIPAR_TC26_512_A; return 0; }
	if (rc == 1) rc = p15_oid_str_eq(P15_OID_PKI_TC26_B, oid);
	if (!rc) { *uz = P15_PKIPAR_TC26_512_B; return 0; }
	if (rc == 1) return P15_RC_UNSUPPORTED_ALGORITHM;
	return P15_RC_INTERNAL_ERROR;
}
/****************************************************************************/
static int p15_str_to_oid(const char* str, OBJECT_IDENTIFIER_t* oid)
{
	int rc = 0;
	p15_oid_t* obj = NULL;

	if (!str || !oid) return P15_RC_INVALID_ARGS;
	
	obj = p15_new_oid_from_str(str);
	if (!obj) return P15_RC_INVALID_OID_STR;
	if (OBJECT_IDENTIFIER_set_arcs(oid, obj->arcs, 4, obj->arcnum)) {
		rc = P15_RC_ASN1C_ERROR;
		goto end;
	}
end:
	p15_free_oid(&obj);
	return rc;
}
/****************************************************************************/
static int p15_set_node_id(p15_ctx_t* ctx, p15_id_t* nodeid, p15_id_t* id, 
	p15_id_t** newid)
{
	if (!ctx || !nodeid) return P15_RC_INVALID_ARGS;
	if (!id) { /* assigning new id */
		nodeid->next = NULL;
		nodeid->id = ctx->free_id;
		nodeid->ext_id_len = 0;
		ctx->free_id++;
		if (newid) 
			*newid = p15_new_id(nodeid->id);
		return 0;
	}
	*nodeid = *id;
	nodeid->next = NULL;
	p15_refine_id(nodeid);
	if (!nodeid->ext_id_len && (nodeid->id >= ctx->free_id)) 
		ctx->free_id = nodeid->id + 1;
	return 0;
}
/****************************************************************************/
static p15_der_t* p15_new_der_from_sbuf(p15_sbuf_t* sbuf)
{
	uint8_t* buf = NULL;
	size_t len;
	p15_der_t* der = NULL;

	if (!sbuf) return NULL;
	if (p15_get_solid_buf(sbuf, &buf, &len)) return NULL;
	der = p15_malloc(sizeof(p15_der_t));
	der->ptr = buf;
	der->len = len;
	return der;
}
/****************************************************************************/
static int p15_name_to_rdn(P15_Name_t* name, p15_der_t** rdn)
{
	int rc = 0;
	uint8_t* ebuf = NULL;
	size_t ebuflen;

	if (!name || !rdn) return P15_RC_INVALID_ARGS;

	rc = p15_asn_encode(&asn_DEF_P15_Name, name, &ebuf, &ebuflen, NULL);
	if (rc) goto end;
	*rdn = p15_new_der(ebuf, ebuflen);
	if (!*rdn) {
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
end:
	p15_free(ebuf);
	return rc;
}
/****************************************************************************/
static void p15_add_error_type(p15_info_t* info, asn_enc_rval_t erv)
{
	int len;

	if (!info) return;
	if (!erv.failed_type) return;
	if (!erv.failed_type->name) return;
	len = strlen(erv.failed_type->name) + 1;
	info->asn_type_name = p15_malloc(len);
	strcpy(info->asn_type_name, erv.failed_type->name);
}
/****************************************************************************/
int p15_consume_bytes(const void *buffer, size_t size, void *appkey)
{
	p15_sbuf_t *sbuf;

	if (!buffer || !size) return 0;
	if (!appkey) return -1;
	sbuf = appkey;
	p15_add_sbuf(sbuf, buffer, size);
	return 0;
}
/****************************************************************************/
p15_sbuf_t* p15_add_sbuf(p15_sbuf_t* root, const uint8_t* ptr, size_t len)
{
	p15_sbuf_t *buf, *next;

	buf = p15_malloc(sizeof(p15_sbuf_t));
	if (ptr) {
		buf->ptr = p15_malloc(len);
		memcpy(buf->ptr, ptr, len);
		buf->len = len;
	}
	if (!root) return buf;
	next = root;
	while (next->next) next = next->next;
	next->next = buf;
	return root;
}
/****************************************************************************/
void p15_free_sbuf(p15_sbuf_t** pbuf)
{
	p15_sbuf_t *buf, *next;

	if (!pbuf) return;
	buf = *pbuf;
	while (buf) {
		next = buf->next;
		if (buf->ptr) {
			p15_free(buf->ptr);
		}
		p15_free(buf);
		buf = next;
	}
	*pbuf = NULL;
}
/****************************************************************************/
int p15_get_solid_buf(p15_sbuf_t* inbuf, uint8_t** outbuf, size_t* outlen)
{
	size_t len;
	p15_sbuf_t* next;
	uint8_t* buf;

	if (!inbuf || !outbuf || !outlen) return P15_RC_INVALID_ARGS;
	len = 0;
	next = inbuf;
	while (next) {
		len += next->len;
		next = next->next;
	}
	*outlen = len;
	*outbuf = p15_malloc(len);
	buf = *outbuf;
	next = inbuf;
	while (next) {
		if (next->ptr) {
			memcpy(buf, next->ptr, next->len);
			buf += next->len;
		}
		next = next->next;
	}
	return 0;
}
/****************************************************************************/
static p15_info_t* p15_new_info(int type, int rc, char* text)
{
	p15_info_t* info;
	int len;

	info = p15_malloc(sizeof(p15_info_t));
	info->type = type;
	info->code = rc;
	if (text) {
		len = strlen(text);
		info->text = p15_malloc(len + 1);
		strcpy(info->text, text);
	}
	return info;
}
/****************************************************************************/
static p15_info_t* p15_add_error(p15_info_t **errors, int type, 
	int rc, char* text)
{
	p15_info_t *info, *lasterr;

	if (!errors) return NULL;
	info = p15_new_info(type, rc, text);
	if (!info) return NULL;
	lasterr = *errors;
	if (!lasterr) {
		*errors = info;
		return info;
	}
	while (lasterr->next) lasterr = lasterr->next;
	lasterr->next = info;
	return info;
}
/****************************************************************************/
p15_der_t* p15_new_rdn(uint8_t* rdn_der, int len)
{
	P15_Name_t* decoded = NULL;
	asn_dec_rval_t arv;
	int rlen;

	if (!rdn_der || len < 2) return NULL;

	/* ck if normal rdn */
	arv = ber_decode(0, &asn_DEF_P15_Name, (void **)&decoded, rdn_der, len);
	ASN_STRUCT_FREE(asn_DEF_P15_Name, decoded);
	if (arv.code != RC_OK) return NULL;
	rlen = arv.consumed;

	return p15_new_der(rdn_der, rlen);
}
/****************************************************************************/
p15_der_t* p15_new_der(uint8_t* der, int len)
{
	p15_der_t* obj;

	if (!der || len < 2) return NULL;

	obj = p15_malloc(sizeof(p15_der_t));
	obj->len = len;
	obj->ptr = p15_malloc(len);
	memcpy(obj->ptr, der, len);

	return obj;
}
/****************************************************************************/
void p15_free_der(p15_der_t** pder)
{
	p15_der_t* der;

	if (!pder) return;
	der = *pder;
	if (!der) return;
	
	if (der->ptr) {
		p15_free(der->ptr);
	}
	p15_free(der);
	*pder = NULL;
}
/****************************************************************************/
int p15_der_cmp(p15_der_t* der1, p15_der_t* der2)
{
	if (!der1 || !der2) return P15_RC_INVALID_ARGS;

	if (der1->len != der2->len) return 1;
	if (memcmp(der1->ptr, der2->ptr, der2->len)) return 1;
	return 0;
}
/****************************************************************************/
void p15_free_cert(p15_certificate_t** pcert)
{
	p15_certificate_t* cert;

	if (!pcert) return;
	cert = *pcert;
	if (!cert) return;
	if (cert->ptr) {
		p15_free(cert->ptr);
	}
	p15_free_der(&cert->subject);
	p15_free_der(&cert->issuer);
	p15_free_der(&cert->serial_number);

	p15_free(cert);
	*pcert = NULL;
}
/****************************************************************************/
void p15_free_private_key(p15_private_key_t** pkey)
{
	p15_private_key_t* key;
	uint8_t** pkm;
	size_t i;

	if (!pkey) return;
	key = *pkey;
	if (!key) return;
	
	if (key->key_masks) {
		pkm = key->key_masks;
		for (i = 0; i < key->mask_count; i++) {
			p15_free(pkm[i]);
		}
		p15_free(pkm);
	}
	p15_free_time(&key->start_date);
	p15_free_time(&key->end_date);
	p15_free_der(&key->subject);

	p15_free(key);
	*pkey = NULL;
}
/****************************************************************************/
void p15_free_public_key(p15_public_key_t** pkey)
{
	p15_public_key_t* key;

	if (!pkey) return;
	key = *pkey;
	if (!key) return;
	p15_free_time(&key->start_date);
	p15_free_time(&key->end_date);
	p15_free_der(&key->subject);

	p15_free(key);
	*pkey = NULL;
}
/****************************************************************************/
void p15_free_crl(p15_crl_t** pcrl)
{
	p15_crl_t* crl;

	if (!pcrl) return;
	crl = *pcrl;
	if (!crl) return;

	if (crl->ptr) {
		p15_free(crl->ptr);
	}
	p15_free_der(&crl->issuer);
	p15_free_der(&crl->this_update);
	p15_free_der(&crl->next_update);

	p15_free(crl);
	*pcrl = NULL;
}
/****************************************************************************/
void p15_free_oiddo(p15_oiddo_t** poiddo)
{
	p15_oiddo_t* oiddo;
	if (!poiddo) return;
	oiddo = *poiddo;
	if (!oiddo) return;
	p15_free_oid(&oiddo->oid);
	p15_free(oiddo->ptr);
	p15_free(oiddo);
	*poiddo = NULL;
}
/****************************************************************************/
p15_certificate_t* p15_copy_cert(p15_certificate_t* oc)
{
	p15_certificate_t* nc = NULL;

	if (!oc) return NULL;
	nc = p15_malloc(sizeof(p15_certificate_t));
	if (oc->label[0]) 
		strcpy(nc->label, oc->label);
	nc->id = oc->id;
	if (!oc->ptr || !oc->len) goto err;
	nc->ptr = p15_malloc(oc->len);
	nc->len = oc->len;
	memcpy(nc->ptr, oc->ptr, oc->len);
	nc->ca = oc->ca;
	if (!oc->subject || !oc->issuer || !oc->serial_number) goto err;
	nc->subject = p15_new_der(oc->subject->ptr, oc->subject->len);
	nc->issuer = p15_new_der(oc->issuer->ptr, oc->issuer->len);
	nc->serial_number = p15_new_der(oc->serial_number->ptr, oc->serial_number->len);
	if (!nc->subject || !nc->issuer || !nc->serial_number) goto err;

	return nc;
err:
	p15_free_cert(&nc);
	return NULL;
}
/****************************************************************************/
p15_private_key_t* p15_copy_private_key(p15_private_key_t* ok)
{
	p15_private_key_t* nk = NULL;
	unsigned int i;

	if (!ok) return NULL;
	nk = p15_malloc(sizeof(p15_private_key_t));
	if (ok->label[0])
		strcpy(nk->label, ok->label);
	nk->id = ok->id;
	if (!ok->key_masks || !ok->mask_count) goto err;
	nk->mask_count = ok->mask_count;
	nk->to_remask = ok->to_remask;
	nk->key_masks = p15_malloc(ok->mask_count * sizeof(uint8_t*));
	for (i = 0; i < ok->mask_count; i++) {
		nk->key_masks[i] = p15_malloc(P15_PAGE);
		memcpy(nk->key_masks[i], ok->key_masks[i], 64);
	}
	nk->key_usage = ok->key_usage;
	nk->key_access_flags = ok->key_access_flags;
	nk->key_parameters = ok->key_parameters;

	if (ok->start_date) {
		nk->start_date = p15_new_time(ok->start_date->year, ok->start_date->month, 
			ok->start_date->day, ok->start_date->hours, ok->start_date->minutes, 
			ok->start_date->seconds);
		if (!nk->start_date) goto err;
	}
	if (ok->end_date) {
		nk->end_date = p15_new_time(ok->end_date->year, ok->end_date->month, 
			ok->end_date->day, ok->end_date->hours, ok->end_date->minutes, 
			ok->end_date->seconds);
		if (!nk->end_date) goto err;
	}
	if (ok->subject) {
		nk->subject = p15_new_der(ok->subject->ptr, ok->subject->len);
		if (!nk->subject) goto err;
	}

	return nk;

err:
	p15_free_private_key(&nk);
	return NULL;	
}
/****************************************************************************/
p15_public_key_t* p15_copy_public_key(p15_public_key_t* ok)
{
	p15_public_key_t* nk = NULL;

	if (!ok) return NULL;
	nk = p15_malloc(sizeof(p15_public_key_t));
	if (ok->label[0])
		strcpy(nk->label, ok->label);
	nk->id = ok->id;
	memcpy(nk->key, ok->key, 128);
	nk->key_usage = ok->key_usage;
	nk->key_parameters = ok->key_parameters;

	if (ok->start_date) {
		nk->start_date = p15_new_time(ok->start_date->year, ok->start_date->month, 
			ok->start_date->day, ok->start_date->hours, ok->start_date->minutes, 
			ok->start_date->seconds);
		if (!nk->start_date) goto err;
	}
	if (ok->end_date) {
		nk->end_date = p15_new_time(ok->end_date->year, ok->end_date->month, 
			ok->end_date->day, ok->end_date->hours, ok->end_date->minutes, 
			ok->end_date->seconds);
		if (!nk->end_date) goto err;
	}
	if (ok->subject) {
		nk->subject = p15_new_der(ok->subject->ptr, ok->subject->len);
		if (!nk->subject) goto err;
	}

	return nk;
err:
	p15_free_public_key(&nk);
	return NULL;
}
/****************************************************************************/
p15_crl_t* p15_copy_crl(p15_crl_t* oc)
{
	p15_crl_t* nc = NULL;
	
	if (!oc) return NULL;
	nc = p15_malloc(sizeof(p15_crl_t));
	if (oc->label[0])
		strcpy(nc->label, oc->label);
	nc->id = oc->id;
	if (!oc->ptr || !oc->len) goto err;
	nc->len = oc->len;
	nc->ptr = p15_malloc(nc->len);
	memcpy(nc->ptr, oc->ptr, oc->len);
	if (!oc->issuer || !oc->this_update) goto err;
	nc->issuer = p15_new_der(oc->issuer->ptr, oc->issuer->len);
	nc->this_update = p15_new_der(oc->this_update->ptr, oc->this_update->len);
	if (!nc->issuer || !nc->this_update) goto err;
	if (oc->next_update) {
		nc->next_update = p15_new_der(oc->next_update->ptr, oc->next_update->len);
		if (!nc->next_update) goto err;
	}

	return nc;
err:
	p15_free_crl(&nc);
	return NULL;
}
/****************************************************************************/
p15_oiddo_t* p15_copy_oiddo(p15_oiddo_t* oo)
{
	p15_oiddo_t* no = NULL;

	if (!oo) return NULL;
	no = p15_malloc(sizeof(p15_oiddo_t));
	if (oo->label[0])
		strcpy(no->label, oo->label);

	if (!oo->oid || !oo->ptr || !oo->len) goto err;
	no->oid = p15_copy_oid(oo->oid);
	if (!no->oid) goto err;
	no->ptr = p15_malloc(oo->len);
	no->len = oo->len;
	memcpy(no->ptr, oo->ptr, oo->len);
	no->to_encrypt = oo->to_encrypt;

	return no;
err:
	p15_free_oiddo(&no);
	return NULL;
}
/****************************************************************************/
static int p15_add_object(p15_ctx_t* ctx, int type, void* obj)
{
	p15_object_t* o = NULL;
	p15_object_t* next;

	if (!ctx || !obj) return P15_RC_INVALID_ARGS;

	o = p15_malloc(sizeof(p15_object_t));	
	o->type = type;
	o->obj = obj;
	next = ctx->objects;
	if (next) {
		while (next->next) next = next->next;
		next->next = o;
		o->prev = next;
	} else 
		ctx->objects = o;
	return 0;
}
/****************************************************************************/
p15_ctx_t* p15_new_ctx(void)
{
	p15_ctx_t* ctx;

	ctx = p15_malloc(sizeof(p15_ctx_t));

	ctx->enc_uz = P15_89UZ_CPRO_A;
	ctx->free_id = 1;
/*	ctx->sk_mask_count = 2; */
	ctx->p5_c = P15_P5_DEFAULT_C;
	ctx->p5_salt_len =  P15_P5_DEFAULT_SALT_LEN;

	return ctx;
}
/****************************************************************************/
static void p15_free_ctx_content(p15_ctx_t* ctx)
/* RNG data not freed */
{
	if (!ctx) return;
	while (ctx->objects) 
		p15_rm_object(ctx, ctx->objects);

	p15_free(ctx->rnd_extra);
	ctx->rnd_extra = NULL;
	ctx->rnd_extra_len = 0;
	p15_free(ctx->rnd_new_fix_val);
	ctx->rnd_new_fix_val = NULL;

	ctx->password_hint[0] = 0;
	ctx->free_id = 1;
	ctx->fts_ver_detected = 0;
	ctx->version_major = 0;
	ctx->version_minor = 0;
}
/****************************************************************************/
void p15_free_ctx(p15_ctx_t** pctx)
{
	p15_ctx_t* ctx;

	if (!pctx) return;
	ctx = *pctx;
	if (!ctx) return;

	p15_free_ctx_content(ctx);

	p15_free(ctx->rnd_val);
	p15_free(ctx->rnd_msk);

	p15_free(ctx);
	*pctx = NULL;
}
/****************************************************************************/
int p15_rm_object(p15_ctx_t* ctx, p15_object_t* obj)
{
	if (!ctx || !obj) return P15_RC_INVALID_ARGS;
	
	/* rm from list */
	if (!obj->prev) {
		if (obj != ctx->objects) return P15_RC_INVALID_OBJECT;
		ctx->objects = obj->next;
		if (ctx->objects) ctx->objects->prev = NULL;
	} else {
		obj->prev->next = obj->next;
		if (obj->next) {
			obj->next->prev = obj->prev;
		}
	}

	switch (obj->type) {
	case P15_OT_PRIVKEY:
		p15_free_private_key((p15_private_key_t**)&obj->obj);
		break;
	case P15_OT_CERT:
	case P15_OT_TRUSTED_CERT:
		p15_free_cert((p15_certificate_t**)&obj->obj);
		break;
	case P15_OT_CRL:
		p15_free_crl((p15_crl_t**)&obj->obj);
		break;
	case P15_OT_PUBKEY:
		p15_free_public_key((p15_public_key_t**)&obj->obj);
		break;
	case P15_OT_OIDDO:
		p15_free_oiddo((p15_oiddo_t**)&obj->obj);
		break;
	case P15_OT_SECKEY:
		p15_free_secret_key((p15_secret_key_t**)&obj->obj);
		break;
	default:
		return P15_RC_INVALID_OBJECT;
	}

	p15_free(obj);
	
	return 0;
}
/****************************************************************************/
void p15_free_infos(p15_info_t** info_list)
{
	p15_info_t *info, *next;

	if (!info_list) return;
	info = *info_list;
	while (info) {
		next = info->next;
		p15_free(info->text);
		p15_free(info->asn_type_name);
		p15_free(info);
		info = next;
	}
	*info_list = NULL;
}
/****************************************************************************/
int p15_set_password_hint(p15_ctx_t* ctx, char* pwhint)
{
	if (!ctx) return P15_RC_INVALID_ARGS;
	
	if (!pwhint) {
		ctx->password_hint[0] = 0;
		return 0;
	}
	if (strnlen(pwhint, P15_LABEL_MAX + 1) > P15_LABEL_MAX) {
		return P15_RC_LABEL_TOO_LONG;
	}
	strcpy(ctx->password_hint, pwhint);
	return 0;
}
/****************************************************************************/
char* p15_get_password_hint(p15_ctx_t* ctx)
{
	if (!ctx) return NULL;
	if (!ctx->password_hint[0]) return NULL;
	return ctx->password_hint;
}
/****************************************************************************/
int p15_set_pbkdf2_params(p15_ctx_t* ctx, uint16_t c, uint16_t saltlen)
{
	if (!ctx || c < 1 || saltlen < 1) return P15_RC_INVALID_ARGS;
	ctx->p5_c = c;
	ctx->p5_salt_len = saltlen;
	return 0;
}
/****************************************************************************/
int p15_init_random(p15_ctx_t* ctx, uint8_t* rnd, uint8_t* msk)
{
#ifdef TESTOUT
	uint8_t fixed_iv[8] = { 0xF1, 0xE2, 0xD3, 0xC4, 0xB5, 0xA6, 0x97, 0x88 };
#else
	time_t t;
#endif

	if (!ctx || !rnd || !msk) return P15_RC_INVALID_ARGS;

#ifdef TESTOUT
	printf("p15_init_random:\n");
	printf("rnd:\n");
	p15_dump(rnd, 32);
	printf("msk:\n");
	p15_dump(msk, 32);
	printf("iv:\n");
	p15_dump(fixed_iv, 8);
	printf("\n");
#endif
	if (!ctx->rnd_val) ctx->rnd_val = p15_malloc(P15_PAGE);
	if (!ctx->rnd_msk) ctx->rnd_msk = p15_malloc(P15_PAGE);
	memcpy(ctx->rnd_val, rnd, 32);
	memcpy(ctx->rnd_msk, msk, 32);
#ifdef TESTOUT
	memcpy(ctx->rnd_iv, fixed_iv, 8);
#else
	t = time(NULL);
	memcpy(ctx->rnd_iv, &t, 4);
	memcpy(ctx->rnd_iv + 4, &t, 4);
#endif
	ctx->rnd_init = 1;
	return 0;
}
/****************************************************************************/
static int p15_get_safe_random(p15_ctx_t* ctx, uint8_t* rnd, int len)
{
	int f = 0;
	uint8_t i;

	if (!ctx || !rnd || len < 1) return P15_RC_INVALID_ARGS;

	if (ctx->rnd_init) return p15_get_random(ctx, rnd, len);
#if defined(WIN32)||defined(WIN64)
	f = _open(URANDOM, O_RDONLY);
	if (f < 0) goto no_urandom;
	if (_read(f, rnd, len) != len) {
		_close(f);
		goto no_urandom;
	}
	_close(f);

#else
	f = open(URANDOM, O_RDONLY);
	if (f < 0) goto no_urandom;
	if (read(f, rnd, len) != len) {
		close(f);
		goto no_urandom;
	}
	close(f);
#endif
	return 0;

no_urandom: /* simply return the counter :) */
	for (i = 0; i < len; i++) 
		rnd[i] = i + 1;
	return 0;
}
/****************************************************************************/
int p15_get_random(p15_ctx_t* ctx, uint8_t* out_buf, int len)
{
	int len8, i, l;
	uint8_t niv[8];
	uint8_t* bf;

	if (!ctx || !out_buf || !len) return P15_RC_INVALID_ARGS;
	if (!ctx->rnd_init) return P15_RC_RNG_NOT_INITIALIZED;
	len8 = len >> 3;
	bf = out_buf;
	for (i = 0; i < len8; i++) {
		gostc_sym_ECB((void*)ctx->rnd_iv, (void*)niv, 
			(void*)ctx->rnd_val, (void*)ctx->rnd_msk,
			&gostc_sym_uz[1][0]);
		memcpy(ctx->rnd_iv, niv, 8);
		memcpy(bf, niv, 8);
		bf += 8;
	}
	l = len - (len8 << 3);
	if (l) {
		gostc_sym_ECB((void*)ctx->rnd_iv, (void*)niv, 
			(void*)ctx->rnd_val, (void*)ctx->rnd_msk,
			&gostc_sym_uz[1][0]);
		memcpy(ctx->rnd_iv, niv, 8);
	}
	for (i = 0; i < l; i++) { *bf = niv[i]; bf++; }

	gostc_cleanse(niv, 8);
	return 0;
}
/****************************************************************************/
int p15_get_extra_random_len(p15_ctx_t* ctx)
{
	if (!ctx) return 0;
	return ctx->rnd_extra_len;
}
/****************************************************************************/
int p15_get_extra_random(p15_ctx_t* ctx, uint8_t* buf, int* len)
{
	int mlen;

	if (!ctx || !buf || !len) return P15_RC_INVALID_ARGS;
	if (*len < 1) return P15_RC_INVALID_ARGS;
	mlen = *len;
	if (mlen > ctx->rnd_extra_len) mlen = ctx->rnd_extra_len;
	*len = mlen;
	if (!mlen) return 0;
	memcpy(buf, ctx->rnd_extra, mlen);
	return 0;
}
/****************************************************************************/
int p15_random_setup(p15_ctx_t* ctx, uint8_t* fixed_value, int len)
{
	if (!ctx || len < 64 || len > P15_MORE_RANDOM_MAX)
		return P15_RC_INVALID_ARGS;
	p15_free(ctx->rnd_new_fix_val);
	ctx->rnd_new_fix_val = NULL;
	ctx->rnd_new_extra_len = len - 64;
	if (fixed_value) {
		ctx->rnd_new_fix_val = p15_malloc(len);
		memcpy(ctx->rnd_new_fix_val, fixed_value, len);
	}
	return 0;
}
/****************************************************************************/
int p15_set_default_enc_uz(p15_ctx_t* ctx, int uz)
{
	if (!ctx) return P15_RC_INVALID_ARGS;
	switch (uz) {
		
		case P15_89UZ_CPRO_A:
		case P15_89UZ_CPRO_B:
		case P15_89UZ_CPRO_C:
		case P15_89UZ_CPRO_D:
			ctx->enc_uz = uz;
			break;

		case P15_89UZ_CPRO_Z:
			if (ctx->p15_format == P15_FORMAT_2001) 
				return P15_RC_89UZ_Z_VS_2001_FORMAT;
			ctx->enc_uz = uz;
			ctx->p15_format = P15_FORMAT_2012; /* forcing format */
			break;
			
		default:
			return P15_RC_INVALID_ARGS;
	}
	return 0;
}
/****************************************************************************/
/*
int p15_set_skeys_mask_count(p15_ctx_t* ctx, int count)
{
	if (!ctx || count < 1 || count > P15_KEY_MASKS_MAX) 
		return P15_RC_INVALID_ARGS;
	
	ctx->sk_mask_count = count;
	return 0;
}
*/
/****************************************************************************/
p15_id_t* p15_new_id(uint32_t id_int)
{
	p15_id_t* id;

	id = p15_malloc(sizeof(p15_id_t));
	id->id = id_int;
	return id;
}
/****************************************************************************/
p15_id_t* p15_new_ext_id(uint8_t* eid, int len)
{
	p15_id_t* id;

	if (!eid || !len || len > P15_EXT_ID_MAX) return NULL;
	id = p15_malloc(sizeof(p15_id_t));
	id->ext_id_len = len;
	memcpy(id->ext_id, eid, len);
	return id;
}
/****************************************************************************/
void p15_free_id(p15_id_t** pid)
{
	p15_id_t *id, *next;

	if (!pid) return;
	id = *pid;
	while (id) {
		next = id->next;
		p15_free(id);
		id = next;
	}
	*pid = NULL;
}
/****************************************************************************/
static void p15_refine_id(p15_id_t* id)
{
	if (!id) return;
	if (id->ext_id_len != 4) return;
	id->ext_id_len = 0;
	id->id = id->ext_id[0];
	id->id <<= 8;
	id->id |= id->ext_id[1];
	id->id <<= 8;
	id->id |= id->ext_id[2];
	id->id <<= 8;
	id->id |= id->ext_id[3];
}
/****************************************************************************/
int p15_id_cmp(p15_id_t* id1, p15_id_t* id2)
{
	if (!id1 || !id2) return P15_RC_INVALID_ARGS;
	p15_refine_id(id1);
	p15_refine_id(id2);
	if (id1->ext_id_len && id2->ext_id_len) {
		if (id1->ext_id_len != id2->ext_id_len) return 1;
		if (memcmp(id1->ext_id, id2->ext_id, id1->ext_id_len)) return 1;
		return 0;
	} else if (!id1->ext_id_len && !id2->ext_id_len) {
		if (id1->id == id2->id) return 0;
		return 1;
	}
	return 1;
}
/****************************************************************************/
int p15_add_cert(p15_ctx_t* ctx, int trusted, int ca, p15_id_t* id, p15_id_t** newid, 
	char* label, uint8_t* cert, size_t certlen, p15_info_t** errors)
{
	P15_Certificate_t* certdec = NULL;
	asn_dec_rval_t drv;
	asn_enc_rval_t erv;	
	int clen;
	int rc = 0;
	int rc2 = 0;
	p15_info_t* info;
	p15_sbuf_t* sbuf = NULL;
	p15_der_t* issuer = NULL;
	p15_der_t* subject = NULL;
	p15_der_t* serial_number = NULL;
	p15_certificate_t* node = NULL;
	p15_certificate_t* next;
	int l;
	p15_object_t* obj = NULL;
	p15_secret_key_t* skey = NULL;

	if (!ctx || !cert || !certlen) {
		p15_add_error(errors, P15_ERROR, 
			P15_RC_INVALID_ARGS, NULL);
		return P15_RC_INVALID_ARGS;
	}

#ifdef TESTOUT
	printf("p15_add_cert: Adding certificate '%s':\n", label ? label : "");
	printf("trusted: %d\n", trusted);
	printf("ca: %d\n", ca);
	printf("certificate (%d bytes):\n", (int)certlen);
	p15_dump(cert, certlen);
	printf("\n");
#endif
	/* Checking id */
	if (id) {
		next = p15_get_cert(ctx, id);
		if (next) {
			rc = P15_RC_CERT_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
		"p15_add_cert: Certificate with given ID is already present in DB");
			goto end;
		}
		skey = p15_get_secret_key(ctx, id);
		if (skey) {
			rc = P15_RC_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
				"p15_add_cert: Given ID is already in use by secret key");
			goto end;
		}
	}

	drv = ber_decode(0, &asn_DEF_P15_Certificate, (void **)&certdec, 
		cert, certlen);
	if (drv.code != RC_OK) {
		rc = P15_RC_CERT_DECODING_FAILED;
		info = p15_add_error(errors, P15_ERROR, rc,
			"p15_add_cert: Certificate decoding error.");
		if (info) info->consumed = drv.consumed;
		goto end;
	}
	clen = drv.consumed;

	/* Encoding subject back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Name, &certdec->tbsCertificate.subject, 
		p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INVALID_SUBJECT_RDN;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_cert: Subject encoding error.");
		p15_add_error_type(info, erv);
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, NULL);
		goto end;
	}
	subject = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!subject) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_cert: int err 1");
		goto end;
	}

	/* Encoding issuer back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Name, &certdec->tbsCertificate.issuer, 
		p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INVALID_ISSUER_RDN;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_cert: Issuer encoding error.");
		p15_add_error_type(info, erv);
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, NULL);
		goto end;
	}
	issuer = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!issuer) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_cert: int err 2");
		goto end;
	}

	/* Encoding serialNumber back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_CertificateSerialNumber, 
		&certdec->tbsCertificate.serialNumber, p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_cert: serialNumber encoding error.");
		p15_add_error_type(info, erv);
		goto end;
	}
	serial_number = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!serial_number) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_add_cert: int err 3");
		goto end;
	}

	/* Searching for duplicate certs */
	for (obj = ctx->objects; obj; obj = obj->next) {
		if (obj->type != P15_OT_CERT && 
			obj->type != P15_OT_TRUSTED_CERT) continue;
		next = (p15_certificate_t*)obj->obj;
		if (!next) {
			rc = P15_RC_INTERNAL_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_add_cert: int err 4");
			goto end;
		}
		rc2 = p15_der_cmp(next->issuer, issuer);
		if (!rc2) rc2 = p15_der_cmp(next->serial_number, serial_number);
		if (!rc2) {
			rc = P15_RC_DUPLICATE_CERT;
			p15_add_error(errors, P15_ERROR, rc, 
"p15_add_cert: Certificate with the same issuer and serial number already exists in the ctx.");
			goto end;
		}
		if (rc2 > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc2, 
				"p15_add_cert: p15_der_cmp returned error");
			rc = P15_RC_INTERNAL_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, NULL);
			goto end;
		}
	}

	/* new structure */
	node = p15_malloc(sizeof(p15_certificate_t));

	node->subject = subject;
	node->issuer = issuer;
	node->serial_number = serial_number;
	subject = NULL;
	issuer = NULL;
	serial_number = NULL;

	node->ca = ca;

	if (label) {
		l = strnlen(label, P15_LABEL_MAX + 1);
		if (l > P15_LABEL_MAX) {
			rc = P15_RC_LABEL_TOO_LONG;
			p15_add_error(errors, P15_ERROR, rc, NULL);
			goto end;
		}
		strcpy(node->label, label);
	}

	node->ptr = p15_malloc(clen);
	node->len = clen;
	memcpy(node->ptr, cert, clen);

	/* id handling */
	if (p15_set_node_id(ctx, &node->id, id, newid)) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_cert: error in p15_set_node_id");
		goto end;
	}

	/* inserting cert */
	p15_add_object(ctx, trusted ? P15_OT_TRUSTED_CERT : P15_OT_CERT, node);
	node = NULL;
end:
	if (certdec) {
		ASN_STRUCT_FREE(asn_DEF_P15_Certificate, certdec);
	}
	p15_free_sbuf(&sbuf);
	p15_free_der(&issuer);
	p15_free_der(&subject);
	p15_free_der(&serial_number);
	p15_free_cert(&node);
	return rc;
}
/****************************************************************************/
int p15_rm_cert(p15_ctx_t* ctx, p15_id_t* id)
{
	int found = 0;
	p15_object_t *o, *o2;
	p15_certificate_t* cert;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_CERT || o->type == P15_OT_TRUSTED_CERT) {
			cert = (p15_certificate_t*)o->obj;
			if (!p15_id_cmp(&cert->id, id)) {
				found = 1;
				o2 = o->next;
				p15_rm_object(ctx, o);
				o = o2;
				continue;
			}
		}
		o = o->next;
	}

	if (found) return 0;
	return P15_RC_CERT_NOT_FOUND;
}
/****************************************************************************/
int p15_find_certs_by_subject(p15_ctx_t* ctx, p15_der_t* subject, 
	p15_id_t** found)
{
	p15_id_t* last = NULL;
	p15_id_t* id;
	p15_certificate_t* cert;
	p15_object_t* o;

	if (!ctx || !subject || !found) return P15_RC_INVALID_ARGS;

	*found = NULL;
	o = ctx->objects;
	
	while (o) {
		if (o->type == P15_OT_CERT || o->type == P15_OT_TRUSTED_CERT) {
			cert = (p15_certificate_t*)o->obj;
			if (!p15_der_cmp(cert->subject, subject)) {
				id = p15_new_id(1);
				*id = cert->id;
				id->next = NULL;
				if (!last) 
					*found = last = id;
				else {
					last->next = id;
					last = id;
				}
			}
		}
		o = o->next;
	}
	return 0;
}
/****************************************************************************/
p15_certificate_t* p15_get_cert(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_certificate_t* cert;
	p15_object_t* o;

	if (!ctx || !id) return NULL;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_CERT || o->type == P15_OT_TRUSTED_CERT) {
			cert = (p15_certificate_t*)o->obj;
			if (!p15_id_cmp(id, &cert->id)) return cert;
		}
		o = o->next;
	}
	return NULL;
}
/****************************************************************************/
p15_certificate_t* p15_get_next_cert(p15_ctx_t* ctx, p15_object_t** it, 
	int trusted)
{
	p15_object_t* o;

	if (!ctx || !it) return NULL;
	o = *it;
	if (!o) 
		o = ctx->objects;
	else
		o = o->next;
	while (o) {
		if (o->type == (trusted ? P15_OT_TRUSTED_CERT : P15_OT_CERT)) 
			break;
		o = o->next;
	}
	*it = o;
	if (o) return (p15_certificate_t*)o->obj;
	return NULL;
}
/****************************************************************************/
p15_utctime_t* p15_new_time(int year, int month, int day, 
	int hours, int minutes, int seconds)
{
	p15_utctime_t* t;

	if (year < 1980 || year > 2500) return NULL; /* :)))))))) */
	switch (month) {
	case 1: case 3: case 5: case 7: case 8: case 10: case 12: 
		if (day < 1 || day > 31) return NULL;
		break;
		
	case 2:
		if (day < 1 || day > ((year & 3) ? 28 : 29)) return NULL;
		break;
			
	case 4:	case 6: case 9: case 11:
		if (day < 1 || day > 30) return NULL;
		break;
		
	default:
		return NULL;
	}
	if (hours < 0 || hours > 23) return NULL;
	if (minutes < 0 || minutes > 59) return NULL;
	if (seconds < 0 || seconds > 59) return NULL;

	t = p15_malloc(sizeof(p15_utctime_t));
	t->year = year;
	t->month = month;
	t->day = day;
	t->hours = hours;
	t->minutes = minutes;
	t->seconds = seconds;

	return t;
}
/****************************************************************************/
void p15_free_time(p15_utctime_t** t)
{
	p15_utctime_t* tm;

	if (!t) return;
	tm = *t;
	p15_free(tm);
	*t = NULL;
}
/****************************************************************************/
static int p15_utctime_to_gt(GeneralizedTime_t* gt, p15_utctime_t* in)
{
	/* 4+2+2 +2+2+2 +1 = 15; */
	char buf[100];

	if (!in || !gt) return P15_RC_INVALID_ARGS;

	sprintf(buf, "%04d%02d%02d%02d%02d%02dZ", in->year, in->month, in->day,
		in->hours, in->minutes, in->seconds);
	if (strlen(buf) != 15) {
		return P15_RC_INTERNAL_ERROR;
	}
	if (OCTET_STRING_fromBuf(gt, (void*)buf, 15)) {
		return P15_RC_INTERNAL_ERROR;
	}

	return 0;
}
/****************************************************************************/
static int p15_gt_to_utctime(GeneralizedTime_t* gt, p15_utctime_t** out)
/* ret: int err, inv token */
{
	int st = 0; 	/* 0-3-yr, 4,5-month, 6,7-day, 8,9-hr, 10,11-min, 12,13-sec 
			 * 14-point or Z, 15-frac or Z, 16-eos */
	int ct = 0;	/* 0-digit, 1-point or Z, 2-digit or Z */
	int d = 0;	/* 0-9-digit, 10-point, 11-Z */
	int l, i;
	uint8_t c;
	int year = 0;
	int month = 0;
	int day = 0;
	int hr = 0;
	int min = 0;
	int sec = 0;

	if (!gt || !out) return P15_RC_INVALID_ARGS;
	l = gt->size;
	if (l < 15 || l > 30) return P15_RC_INVALID_TOKEN;
	for (i = 0; i < l; i++) {
		c = gt->buf[i];
		switch (c) {
		case '0': case '1': case '2': case '3': case '4': 
		case '5': case '6': case '7': case '8': case '9':
			if (ct == 1) return P15_RC_INVALID_TOKEN;
			d = c - '0';
			break;

		case '.':
			if (ct != 1) return P15_RC_INVALID_TOKEN;
			d = 10;
			break;

		case 'Z':
			if (ct == 0) return P15_RC_INVALID_TOKEN;
			d = 11;
			break;

		default:
			return P15_RC_INVALID_TOKEN;
		}

		switch (st) {
		case 0: case 1: case 2: case 3:
			year *= 10;
			year += d;
			st++;
			break;

		case 4: case 5:
			month *= 10;
			month += d;
			st++;
			break;

		case 6: case 7:
			day *= 10;
			day += d;
			st++;
			break;
		
		case 8: case 9:
			hr *= 10;
			hr += d;
			st++;
			break;

		case 10: case 11:
			min *= 10;
			min += d;
			st++;
			break;

		case 12: 
			sec = d * 10;
			st++;
			break;
		
		case 13:
			sec += d;
			st++;
			ct = 1;
			break;

		case 14:
			if (d == 11) { st = 16; break; } /* Z */
			if (d == 10) { st = 15; ct = 3; break; } /* point */
			return P15_RC_INTERNAL_ERROR;

		case 15:
			if (d == 11) { st = 16; break; } /* Z */
			if (d > 9) return P15_RC_INTERNAL_ERROR;
			break;

		default: 
			return P15_RC_INVALID_TOKEN;
		}
	}

	if (st != 16) return P15_RC_INVALID_TOKEN;
	
	*out = p15_new_time(year, month, day, hr, min, sec);
	if (!*out) return P15_RC_INVALID_TOKEN;

/* "19920722132100.30Z" */

	return 0;
}
/****************************************************************************/
int p15_add_private_key(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	p15_info_t** errors,
	int mask_count, uint8_t* key, ...)
{
	va_list masks;
	uint8_t** mask_ar = NULL;
	int i;
	int rc = 0;
	uint8_t* mask;

	if (!ctx || key_params < P15_PKIPAR_CPRO_A || !key ||
			key_params > P15_PKIPAR_TC26_512_B ||
			mask_count < 1 || mask_count > P15_KEY_MASKS_MAX) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

	if (ctx->p15_format == P15_FORMAT_2001 && p15_pkey_is_2012(key_params)) {
		rc = P15_RC_2012_KEY_VS_2001_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_add_private_key: Unable to add a GOST R 34.10-2012 private key to P15-2001 container");
		goto end;
	}

	mask_ar = p15_malloc(mask_count * sizeof(uint8_t*));
	mask_ar[0] = key;
	if (mask_count > 1) {
		va_start(masks, key);
		for (i = 1; i < mask_count; i++) {
			mask = va_arg(masks, uint8_t*);
			if (!mask) {
				va_end(masks);
				rc = P15_RC_INVALID_ARGS;
				p15_add_error(errors, P15_ERROR, rc, 
"p15_add_private_key: One of the masks is NULL pointer");
				goto end;
			}
			mask_ar[i] = mask;
		}
		va_end(masks);
		mask = NULL;
	}
	rc = p15_add_private_key2(ctx, id, newid, label, key_usage, key_access_flags, 
		key_params, start_date, end_date, mask_count, mask_ar, errors);
end:
	p15_free(mask_ar);
	return rc;	
}
/****************************************************************************/
int p15_add_private_key2(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	int mask_count, uint8_t** key_masks,
	p15_info_t** errors)
{
	int rc = 0;
	p15_private_key_t* pkey = NULL;
	p15_private_key_t* next;
	int l, i;
	p15_utctime_t* tm;
	int is512 = 0;
	p15_secret_key_t* skey = NULL;

	if (!ctx || key_params < P15_PKIPAR_CPRO_A || !key_masks ||
			key_params > P15_PKIPAR_TC26_512_B ||
			mask_count < 1 || mask_count > P15_KEY_MASKS_MAX) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

	if (ctx->p15_format == P15_FORMAT_2001 && p15_pkey_is_2012(key_params)) {
		rc = P15_RC_2012_KEY_VS_2001_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_add_private_key: Unable to add a GOST R 34.10-2012 private key to P15-2001 container");
		goto end;
	}

	is512 = p15_pkey_is_2012(key_params);

#ifdef TESTOUT
	printf("p15_add_private_key: Adding key '%s':\n", label ? label: "");
	printf("key usage: ");
	p15_print_key_usage(key_usage);
	printf("\nkey access: ");
	p15_print_key_access(key_access_flags);
	printf("\n");
	printf("key parameters (1-8 - cproA,B,C,XchA,XchB,512test,512A,512B): %d\n", 
		key_params);
	if (start_date) {
		printf("start date: %d-%.2d-%.2d %d:%.2d:%.2d\n", 
			start_date->year,
			start_date->month,
			start_date->day,
			start_date->hours,
			start_date->minutes,
			start_date->seconds);
	} else printf("start date: absent\n");
	if (end_date) {
		printf("end date: %d-%.2d-%.2d %d:%.2d:%.2d\n", 
			end_date->year,
			end_date->month,
			end_date->day,
			end_date->hours,
			end_date->minutes,
			end_date->seconds);
	} else printf("end date: absent\n");
	if (mask_count > 1) printf("key: masked\n");
	else {
		printf("key (little-endian):\n");
		p15_dump(key_masks[0], is512 ? 64 : 32);
	}
	printf("\n");
#endif

	/* Checking ID */
	if (id) {
		next = p15_get_private_key(ctx, id);
		if (next) {
			rc = P15_RC_PRIVATE_KEY_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
		"p15_add_private_key: Private key with given ID already exists in DB");
			goto end;
		}
		skey = p15_get_secret_key(ctx, id);
		if (skey) {
			rc = P15_RC_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
			"p15_add_private_key: Given ID is already in use by secret key");
			goto end;
		}
	}

	pkey = p15_malloc(sizeof(p15_private_key_t));

	if (label) {
		l = strnlen(label, P15_LABEL_MAX + 1);
		if (l > P15_LABEL_MAX) {
			rc = P15_RC_LABEL_TOO_LONG;
			p15_add_error(errors, P15_ERROR, rc, NULL);
			goto end;
		}
		strcpy(pkey->label, label);
	}

	pkey->mask_count = mask_count;
	l = sizeof(uint8_t*) * mask_count;
	pkey->key_masks = p15_malloc(l);
	for (i = 0; i < mask_count; i++) 
		pkey->key_masks[i] = p15_malloc(P15_PAGE);
	for (i = 0; i < mask_count; i++) 
		memcpy(pkey->key_masks[i], key_masks[i], is512 ? 64: 32);

	pkey->key_usage = key_usage;
	pkey->key_parameters = key_params;
	key_access_flags &= P15_KA_MASK;
	if (key_access_flags > 1) 
		key_access_flags &= 0xFE;
	pkey->key_access_flags = key_access_flags;

	if (start_date) {
		tm = start_date;
		pkey->start_date = p15_new_time(tm->year, tm->month, tm->day,
			tm->hours, tm->minutes, tm->seconds);
	}

	if (end_date) {
		tm = end_date;
		pkey->end_date = p15_new_time(tm->year, tm->month, tm->day,
			tm->hours, tm->minutes, tm->seconds);
	}

	/* Handling IDs */
	if (p15_set_node_id(ctx, &pkey->id, id, newid)) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_private_key: error in p15_set_node_id");
		goto end;
	}

	/* Inserting */
	p15_add_object(ctx, P15_OT_PRIVKEY, pkey);
	pkey = NULL;

	/* Forcing 2012 format if 512 key */
	if (is512) ctx->p15_format = P15_FORMAT_2012;
end:
	p15_free_private_key(&pkey);
	return rc;
}
/****************************************************************************/
static int p15_remask_setup_prkey(p15_ctx_t* ctx, p15_private_key_t* prkey, 
	int to_remask, int mask_count, p15_info_t** errors)
{
	int rc = 0;
	int interr = 0;
	char st[4096];

	prkey->to_remask = to_remask ? 1 : 0;
	if (mask_count > 0) {
		rc = p15_remask_private_key(ctx, prkey, mask_count);
		if (!(	rc == P15_RC_RNG_NOT_INITIALIZED ||
			rc == P15_RC_NO_PRIVATE_KEY_DATA ||
			rc == P15_RC_CRYPTOLIB_MASK)) interr = 1;
		if (rc) {
			sprintf(st, 
"p15_remask_setup: p15_remask_private_key returned error for private key '%s'",
				prkey->label);
			p15_add_error(errors, interr ? P15_INTERNAL_ERROR : P15_ERROR, rc, st);
			if (interr) rc = P15_RC_INTERNAL_ERROR;
		}
	}
	return rc;
}
/****************************************************************************/
static int p15_remask_setup_seckey(p15_ctx_t* ctx, p15_secret_key_t* seckey, 
	int to_remask, int mask_count, p15_info_t** errors)
{
	int rc = 0;
	int interr = 0;
	char st[4096];

	seckey->to_remask = to_remask ? 1 : 0;
	if (mask_count > 0) {
		rc = p15_remask_secret_key(ctx, seckey, mask_count);
		if (!(	rc == P15_RC_RNG_NOT_INITIALIZED ||
			rc == P15_RC_NO_PRIVATE_KEY_DATA)) interr = 1;
		if (rc) {
			sprintf(st, 
"p15_remask_setup: p15_remask_secret_key returned error for secret key '%s'",
				seckey->label);
			p15_add_error(errors, interr ? P15_INTERNAL_ERROR : P15_ERROR, rc, st);
			if (interr) rc = P15_RC_INTERNAL_ERROR;
		}
	}
	return rc;
}
/****************************************************************************/
int p15_remask_setup(p15_ctx_t* ctx, p15_private_key_t* prkey, p15_secret_key_t* seckey,
	int to_remask, int mask_count, p15_info_t** errors)
{
	int rc = 0;
	p15_object_t* oit = NULL;
	p15_private_key_t* k = NULL;
	p15_secret_key_t* sk = NULL;

	if (!ctx || mask_count < 0) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, "p15_remask_setup");
		goto end;
	}
	
	if (prkey) {
		rc = p15_remask_setup_prkey(ctx, prkey, to_remask, mask_count, errors);
		if (rc) goto end;
	}

	if (seckey) {
		rc = p15_remask_setup_seckey(ctx, seckey, to_remask, mask_count, errors);
		if (rc) goto end;
	}

	if (prkey || seckey) goto end;

	oit = NULL;
	while ((k = p15_get_next_private_key(ctx, &oit))) {
		rc = p15_remask_setup_prkey(ctx, k, to_remask, mask_count, errors);
		if (rc) goto end; 
	}

	oit = NULL;
	while ((sk = p15_get_next_secret_key(ctx, &oit))) {
		rc = p15_remask_setup_seckey(ctx, sk, to_remask, mask_count, errors);
		if (rc) goto end;
	}
end:
	return rc;
}
/****************************************************************************/
static int p15_get_mask_uz(int key_params)
/* ret: -1 if error */
{
	switch (key_params) {
		case P15_PKIPAR_CPRO_A: 
			return 1;
		case P15_PKIPAR_CPRO_B:
			return 2;
		case P15_PKIPAR_CPRO_C:
			return 3;
		case P15_PKIPAR_CPRO_XCHA:
			return 1;
		case P15_PKIPAR_CPRO_XCHB:
			return 3;
		case P15_PKIPAR_TC26_512_TEST:
			return 4;
		case P15_PKIPAR_TC26_512_A:
			return 5;
		case P15_PKIPAR_TC26_512_B:
			return 6;
		default:
			return -1;
	}
	return -1;
}
/****************************************************************************/
int p15_remask_private_key(p15_ctx_t* ctx, p15_private_key_t* key, int mask_count)
/* ret: P15_RC_RNG_NOT_INITIALIZED, P15_RC_NO_PRIVATE_KEY_DATA,	P15_RC_CRYPTOLIB_MASK2001 */
{
	int uz;
	uint8_t kbuf[64];
	int rc = 0;
	int i;
	uint8_t **nmasks = NULL;
	uint8_t zero[64];
	int is512 = 0;

	if (!ctx || !key || mask_count < 1) 
		return P15_RC_INVALID_ARGS;
	if (!ctx->rnd_init) 
		return P15_RC_RNG_NOT_INITIALIZED;
	if (key->mask_count < 1 || !key->key_masks)
		return P15_RC_NO_PRIVATE_KEY_DATA;
	bzero(zero, sizeof(zero));
	is512 = p15_pkey_is_2012(key->key_parameters);
	uz = p15_get_mask_uz(key->key_parameters);
	if (uz < 0) 
		return P15_RC_INTERNAL_ERROR;
	
	/* unmasking */
	memcpy(kbuf, key->key_masks[0], is512 ? 64 : 32);
	for (i = 1; i < (int)key->mask_count; i++) {
		if (!gostc_asym_unmask_key(kbuf, key->key_masks[i], uz)) {
			rc = P15_RC_CRYPTOLIB_MASK;
			goto end;
		}
	}
	nmasks = p15_malloc(sizeof(uint8_t*) * mask_count);
	for (i = 0; i < mask_count; i++)
		nmasks[i] = p15_malloc(P15_PAGE);
#ifdef TESTOUT
	printf("Remasking private key '%s':\n", key->label);
	printf("Unmasked key (little-endian):\n");
	p15_dump(kbuf, is512 ? 64 : 32);
#endif
	
	/* masking */
	for (i = 1; i < mask_count; i++) {
		while (!memcmp(nmasks[i], zero, is512 ? 64 : 32)) { /* avoid zero mask */
			if (p15_get_random(ctx, nmasks[i], is512 ? 64 : 32)) {
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
		}
		if (!gostc_asym_mask_key(kbuf, nmasks[i], uz)) {
			rc = P15_RC_CRYPTOLIB_MASK;
			goto end;
		}
	}
	memcpy(nmasks[0], kbuf, is512 ? 64 : 32);

	/* mask replacing */
	for (i = 0; i < (int)key->mask_count; i++) 
		p15_free(key->key_masks[i]);
	p15_free(key->key_masks);
	key->key_masks = nmasks;
	key->mask_count = mask_count;
	nmasks = NULL;
#ifdef TESTOUT
	if (key->mask_count == 1) {
		printf("The key is left unmasked.\n");
	} else {
		for (i = 0; i < (int)key->mask_count; i++) {
			if (!i) printf("Masked key (little-endian):\n");
			else printf("Mask %d:\n", i);
			p15_dump(key->key_masks[i], is512 ? 64 : 32);
		}
	}
	printf("\n");
#endif
end:
	if (nmasks) {
		for (i = 0; i < mask_count; i++)
			p15_free(nmasks[i]);
		p15_free(nmasks);
	}
	gostc_cleanse(kbuf, is512 ? 64 : 32);
	return rc;	
}
/****************************************************************************/
int p15_remask_secret_key(p15_ctx_t* ctx, p15_secret_key_t* key, int mask_count)
/* ret: P15_RC_RNG_NOT_INITIALIZED, P15_RC_NO_SECRET_KEY_DATA */
{
	uint8_t kbuf[32];
	int rc = 0;
	int i, j;
	uint8_t **nmasks = NULL;
	uint8_t zero[32];

	if (!ctx || !key || mask_count < 1) 
		return P15_RC_INVALID_ARGS;
	if (!ctx->rnd_init) 
		return P15_RC_RNG_NOT_INITIALIZED;
	if (key->mask_count < 1 || !key->key_masks)
		return P15_RC_NO_SECRET_KEY_DATA;
	bzero(zero, sizeof(zero));
	
	/* unmasking */
	memcpy(kbuf, key->key_masks[0], 32);
	for (i = 1; i < (int)key->mask_count; i++) {
		if (!key->key_masks[i]) {
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		for (j = 0; j < 32; j++) kbuf[j] ^= key->key_masks[i][j];
	}
	nmasks = p15_malloc(sizeof(uint8_t*) * mask_count);
	for (i = 0; i < mask_count; i++)
		nmasks[i] = p15_malloc(P15_PAGE);
#ifdef TESTOUT
	printf("Remasking secret key '%s':\n", key->label);
	printf("Unmasked key (little-endian):\n");
	p15_dump(kbuf, 32);
#endif
	
	/* masking */
	for (i = 1; i < mask_count; i++) {
		while (!memcmp(nmasks[i], zero, 32)) { /* avoid zero mask */
			if (p15_get_random(ctx, nmasks[i], 32)) {
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
		}
		for (j = 0; j < 32; j++) kbuf[j] ^= nmasks[i][j];
	}
	memcpy(nmasks[0], kbuf, 32);

	/* mask replacing */
	for (i = 0; i < (int)key->mask_count; i++) 
		p15_free(key->key_masks[i]);
	p15_free(key->key_masks);
	key->key_masks = nmasks;
	key->mask_count = mask_count;
	nmasks = NULL;
#ifdef TESTOUT
	if (key->mask_count == 1) {
		printf("The key is left unmasked.\n");
	} else {
		for (i = 0; i < (int)key->mask_count; i++) {
			if (!i) printf("Masked key (little-endian):\n");
			else printf("Mask %d:\n", i);
			p15_dump(key->key_masks[i], 32);
		}
	}
	printf("\n");
#endif
end:
	if (nmasks) {
		for (i = 0; i < mask_count; i++)
			p15_free(nmasks[i]);
		p15_free(nmasks);
	}
	gostc_cleanse(kbuf, 32);
	return rc;	
}
/****************************************************************************/
int p15_set_pkey_subject(p15_ctx_t* ctx, p15_id_t* id, 
	p15_der_t* rdn)
{
	p15_private_key_t* privkey;
	p15_public_key_t* pubkey;
	p15_der_t* rdn2;
	p15_certificate_t* cert;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;

	privkey = p15_get_private_key(ctx, id);
	pubkey = p15_get_public_key(ctx, id);
	if (!privkey && !pubkey) return P15_RC_KEY_NOT_FOUND;

	if (!rdn) { /* subject from cert */
		cert = p15_get_cert(ctx, id);
		if (!cert) return P15_RC_CERT_NOT_FOUND;
		if (!cert->subject) return P15_RC_INTERNAL_ERROR;
		rdn = cert->subject;
	}

	rdn2 = p15_new_der(rdn->ptr, rdn->len);
	if (!rdn2) return P15_RC_INVALID_SUBJECT_RDN;
	if (privkey) {
		p15_free_der(&privkey->subject);
		privkey->subject = rdn2;
		rdn2 = NULL;
	}
	if (pubkey) {
		if (!rdn2) 
			rdn2 = p15_new_der(rdn->ptr, rdn->len);
		p15_free_der(&pubkey->subject);
		pubkey->subject = rdn2;
		rdn2 = NULL;
	}

	return 0;
}
/****************************************************************************/
int p15_rm_private_key(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_private_key_t *key;
	int found = 0;
	p15_object_t *o, *o2;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_PRIVKEY) {
			key = (p15_private_key_t*)o->obj;
			if (!p15_id_cmp(&key->id, id)) {
				found = 1;
				o2 = o->next;
				p15_rm_object(ctx, o);
				o = o2;
				continue;
			}
		}
		o = o->next;
	}
	if (found) return 0;
	return P15_RC_PRIVATE_KEY_NOT_FOUND;
}
/****************************************************************************/
int p15_find_private_keys_by_subject(p15_ctx_t* ctx, p15_der_t* subject, 
	p15_id_t** found)
{
	p15_id_t* last = NULL;
	p15_id_t* id;
	p15_object_t* o;
	p15_private_key_t* key;

	if (!ctx || !subject || !found) return P15_RC_INVALID_ARGS;

	*found = NULL;
	o = ctx->objects;

	while (o) {
		if (o->type == P15_OT_PRIVKEY) {
			key = (p15_private_key_t*)o->obj;
			if (!p15_der_cmp(key->subject, subject)) {
				id = p15_new_id(1);
				*id = key->id;
				id->next = NULL;
				if (!last) 
					*found = last = id;
				else {
					last->next = id;
					last = id;
				}
			}
		}
		o = o->next;
	}
	return 0;
}
/****************************************************************************/
p15_private_key_t* p15_get_private_key(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_private_key_t* key;
	p15_object_t* o;

	if (!ctx || !id) return NULL;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_PRIVKEY) {
			key = (p15_private_key_t*)o->obj;
			if (!p15_id_cmp(id, &key->id)) return key;
		}
		o = o->next;
	}
	return NULL;
}
/****************************************************************************/
p15_private_key_t* p15_get_next_private_key(p15_ctx_t* ctx, p15_object_t** it) 
{
	p15_object_t* o;

	if (!ctx || !it) return NULL;
	o = *it;
	if (!o) 
		o = ctx->objects;
	else
		o = o->next;
	while (o) {
		if (o->type == P15_OT_PRIVKEY) 
			break;
		o = o->next;
	}
	*it = o;
	if (o) return (p15_private_key_t*)o->obj;
	return NULL;
}
/****************************************************************************/
int p15_add_public_key(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint8_t* key, uint16_t key_usage, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	p15_info_t** errors)
{
	int rc = 0;
	p15_public_key_t* pkey = NULL;
	p15_public_key_t* next;
	int l;
	p15_utctime_t* tm;
	int is1024;
	p15_secret_key_t* skey = NULL;

	if (!ctx || key_params < P15_PKIPAR_CPRO_A || !key ||
			key_params > P15_PKIPAR_TC26_512_B) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

	is1024 = p15_pkey_is_2012(key_params);

	if (ctx->p15_format == P15_FORMAT_2001 && is1024) {
		rc = P15_RC_2012_KEY_VS_2001_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_add_public_key: Unable to add a GOST R 34.10-2012 public key to P15-2001 container");
		goto end;
	}

#ifdef TESTOUT
	printf("p15_add_public_key: Adding key '%s'\n", label ? label : "");
	printf("key usage: ");
	p15_print_key_usage(key_usage);
	printf("\n");
	printf("key parameters (1-8 - cproA,B,C,XchA,XchB,512Test,512A,512B): %d\n", 
		key_params);
	if (start_date) {
		printf("start date: %d-%.2d-%.2d %d:%.2d:%.2d\n", 
			start_date->year,
			start_date->month,
			start_date->day,
			start_date->hours,
			start_date->minutes,
			start_date->seconds);
	} else printf("start date: absent\n");
	if (end_date) {
		printf("end date: %d-%.2d-%.2d %d:%.2d:%.2d\n", 
			end_date->year,
			end_date->month,
			end_date->day,
			end_date->hours,
			end_date->minutes,
			end_date->seconds);
	} else printf("end date: absent\n");
	printf("key (little-endian):\n");
	p15_dump(key, is1024 ? 128 : 64);
	printf("\n");
#endif
	/* Checking ID */
	if (id) {
		next = p15_get_public_key(ctx, id);
		if (next) {
			rc = P15_RC_PUBLIC_KEY_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
		"p15_add_public_key: Public key with given ID already exists in DB");
			goto end;
		}
		skey = p15_get_secret_key(ctx, id);
		if (skey) {
			rc = P15_RC_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
			"p15_add_public_key: Given ID is already in use by secret key");
			goto end;
		}
	}

	pkey = p15_malloc(sizeof(p15_public_key_t));

	if (label) {
		l = strnlen(label, P15_LABEL_MAX + 1);
		if (l > P15_LABEL_MAX) {
			rc = P15_RC_LABEL_TOO_LONG;
			p15_add_error(errors, P15_ERROR, rc, NULL);
			goto end;
		}
		strcpy(pkey->label, label);
	}

	memcpy(pkey->key, key, is1024 ? 128 : 64);

	pkey->key_usage = key_usage;
	pkey->key_parameters = key_params;

	if (start_date) {
		tm = start_date;
		pkey->start_date = p15_new_time(tm->year, tm->month, tm->day,
			tm->hours, tm->minutes, tm->seconds);
	}

	if (end_date) {
		tm = end_date;
		pkey->end_date = p15_new_time(tm->year, tm->month, tm->day,
			tm->hours, tm->minutes, tm->seconds);
	}

	/* Handling IDs */
	if (p15_set_node_id(ctx, &pkey->id, id, newid)) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_public_key: error in p15_set_node_id");
		goto end;
	}

	/* Inserting */
	p15_add_object(ctx, P15_OT_PUBKEY, pkey);
	pkey = NULL;

	/* Forcing 2012 format if 1024 key */
	if (is1024) ctx->p15_format = P15_FORMAT_2012;
end:
	p15_free_public_key(&pkey);
	return rc;
}
/****************************************************************************/
int p15_rm_public_key(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_public_key_t *key;
	int found = 0;
	p15_object_t *o, *o2;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_PUBKEY) {
			key = (p15_public_key_t*)o->obj;
			if (!p15_id_cmp(&key->id, id)) {
				found = 1;
				o2 = o->next;
				p15_rm_object(ctx, o);
				o = o2;
				continue;
			}
		}
		o = o->next;
	}
	if (found) return 0;
	return P15_RC_PUBLIC_KEY_NOT_FOUND;
}
/****************************************************************************/
int p15_find_public_keys_by_subject(p15_ctx_t* ctx, p15_der_t* subject, 
	p15_id_t** found)
{
	p15_id_t* last = NULL;
	p15_id_t* id;
	p15_object_t* o;
	p15_public_key_t* key;

	if (!ctx || !subject || !found) return P15_RC_INVALID_ARGS;

	*found = NULL;
	o = ctx->objects;

	while (o) {
		if (o->type == P15_OT_PUBKEY) {
			key = (p15_public_key_t*)o->obj;
			if (!p15_der_cmp(key->subject, subject)) {
				id = p15_new_id(1);
				*id = key->id;
				id->next = NULL;
				if (!last) 
					*found = last = id;
				else {
					last->next = id;
					last = id;
				}
			}
		}
		o = o->next;
	}
	return 0;
}
/****************************************************************************/
p15_public_key_t* p15_get_public_key(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_public_key_t* key;
	p15_object_t* o;

	if (!ctx || !id) return NULL;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_PUBKEY) {
			key = (p15_public_key_t*)o->obj;
			if (!p15_id_cmp(id, &key->id)) return key;
		}
		o = o->next;
	}
	return NULL;
}
/****************************************************************************/
p15_public_key_t* p15_get_next_public_key(p15_ctx_t* ctx, p15_object_t** it) 
{
	p15_object_t* o;

	if (!ctx || !it) return NULL;
	o = *it;
	if (!o) 
		o = ctx->objects;
	else
		o = o->next;
	while (o) {
		if (o->type == P15_OT_PUBKEY) 
			break;
		o = o->next;
	}
	*it = o;
	if (o) return (p15_public_key_t*)o->obj;
	return NULL;
}
/****************************************************************************/
int p15_add_crl(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid, 
	char* label, uint8_t* crl, int crllen, int mult_crls_per_issuer, 
	p15_info_t** errors)
{
	int rc = 0;
	int rc2;
	p15_crl_t* node = NULL;
	p15_crl_t* next = NULL;
	asn_dec_rval_t drv;
	asn_enc_rval_t erv;
	P15_CertificateList_t* crldec = NULL;
	p15_info_t* info;
	p15_sbuf_t* sbuf = NULL;
	p15_der_t* issuer = NULL;
	p15_der_t* this_update = NULL;
	p15_der_t* next_update = NULL;
	int clen;
	int l;
	p15_object_t* obj = NULL;
	p15_secret_key_t* skey = NULL;

	if (!ctx || !crl || !crllen) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

#ifdef TESTOUT
	printf("p15_add_crl: Adding CRL '%s':\n", label ? label : "");
	printf("crl (%d bytes):\n", crllen);
	p15_dump(crl, crllen);
	printf("\n");
#endif
	/* Checking id */
	if (id) {
		next = p15_get_crl(ctx, id);
		if (next) {
			rc = P15_RC_CRL_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
			"p15_add_crl: CRL with given ID already exists in DB");
			goto end;
		}
		skey = p15_get_secret_key(ctx, id);
		if (skey) {
			rc = P15_RC_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
				"p15_add_crl: Given ID is already in use by secret key");
			goto end;
		}
	}

	drv = ber_decode(0, &asn_DEF_P15_CertificateList, (void **)&crldec, 
		crl, crllen);
	if (drv.code != RC_OK) {
		rc = P15_RC_CRL_DECODING_FAILED;
		info = p15_add_error(errors, P15_ERROR, rc,
			"p15_add_crl: CRL decoding error.");
		if (info) info->consumed = drv.consumed;
		goto end;
	}
	clen = drv.consumed;

	/* Encoding issuer back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Name, &crldec->tbsCertList.issuer, 
		p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INVALID_ISSUER_RDN;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: Issuer encoding error.");
		p15_add_error_type(info, erv);
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, NULL);
		goto end;
	}
	issuer = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!issuer) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: int err 2");
		goto end;
	}

	/* Encoding thisUpdate back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Time, &crldec->tbsCertList.thisUpdate,
		p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: thisUpdate encoding error.");
		p15_add_error_type(info, erv);
		goto end;
	}
	this_update = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!this_update) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: int err 3");
		goto end;
	}
	
	/* Encoding nextUpdate back */
	if (!crldec->tbsCertList.nextUpdate) goto skip_next_update;
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Time, crldec->tbsCertList.nextUpdate,
		p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: nextUpdate encoding error.");
		p15_add_error_type(info, erv);
		goto end;
	}
	next_update = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!next_update) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: int err 4");
		goto end;
	}
skip_next_update: 

	/* Searching for duplicate CRLs */
	for (obj = ctx->objects; obj; obj = obj->next) {
		if (obj->type != P15_OT_CRL) continue;
		next = (p15_crl_t*)obj->obj;
		if (!next) {
			rc = P15_RC_INTERNAL_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_add_crl: int err 5");
			goto end;
		}
		rc2 = p15_der_cmp(next->issuer, issuer);
		if (!rc2 && mult_crls_per_issuer) 
			rc2 = p15_der_cmp(next->this_update, this_update);
		if (!rc2 && mult_crls_per_issuer) {
			if ((next->next_update && !next_update) ||
				(!next->next_update && next_update)) rc2 = 1;
		}
		if (!rc2 && mult_crls_per_issuer) 
			rc2 = p15_der_cmp(next->next_update, next_update);
		if (!rc2) {
			rc = P15_RC_DUPLICATE_CRL;
			p15_add_error(errors, P15_ERROR, rc, 
				mult_crls_per_issuer ?
"p15_add_crl: CRL with the same issuer and this/nextUpdate already exists in the ctx." : 
"p15_add_crl: CRL with the same issuer already exists in the ctx.");
			goto end;
		}
		if (rc2 > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc2, 
				"p15_add_crl: p15_der_cmp returned error");
			rc = P15_RC_INTERNAL_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, NULL);
			goto end;
		}
	}

	/* new structure */
	node = p15_malloc(sizeof(p15_crl_t));

	node->issuer = issuer;
	node->this_update = this_update;
	node->next_update = next_update;
	issuer = NULL;
	this_update = NULL;
	next_update = NULL;

	if (label) {
		l = strnlen(label, P15_LABEL_MAX + 1);
		if (l > P15_LABEL_MAX) {
			rc = P15_RC_LABEL_TOO_LONG;
			p15_add_error(errors, P15_ERROR, rc, NULL);
			goto end;
		}
		strcpy(node->label, label);
	}

	node->ptr = p15_malloc(clen);
	node->len = clen;
	memcpy(node->ptr, crl, clen);

	/* id handling */
	if (p15_set_node_id(ctx, &node->id, id, newid)) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_crl: error in p15_set_node_id");
		goto end;
	}

	/* inserting crl */
	p15_add_object(ctx, P15_OT_CRL, node);
	node = NULL;

end:
	if (crldec) {
		ASN_STRUCT_FREE(asn_DEF_P15_CertificateList, crldec);
	}
	p15_free_sbuf(&sbuf);
	p15_free_der(&issuer);
	p15_free_der(&this_update);
	p15_free_der(&next_update);
	p15_free_crl(&node);

	return rc;
}
/****************************************************************************/
int p15_rm_crl(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_crl_t *crl;
	int found = 0;
	p15_object_t *o, *o2;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_CRL) {
			crl = (p15_crl_t*)o->obj;
			if (!p15_id_cmp(id, &crl->id)) {
				found = 1;
				o2 = o->next;
				p15_rm_object(ctx, o);
				o = o2;
				continue;
			}
		}
		o = o->next;
	}
	if (found) return 0;
	return P15_RC_CRL_NOT_FOUND;
}
/****************************************************************************/
int p15_find_crl_by_issuer(p15_ctx_t* ctx, p15_der_t* issuer,
	p15_id_t** found)
{
	p15_id_t* last = NULL;
	p15_id_t* id;
	p15_crl_t* crl;
	p15_object_t* o;

	if (!ctx || !issuer || !found) return P15_RC_INVALID_ARGS;
	
	*found = NULL;
	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_CRL) {
			crl = (p15_crl_t*)o->obj;
			if (!p15_der_cmp(crl->issuer, issuer)) {
				id = p15_new_id(1);
				*id = crl->id;
				id->next = NULL;
				if (!last) 
					*found = last = id;
				else {
					last->next = id;
					last = id;
				}
			}
		}
		o = o->next;
	}
	return 0;
}
/****************************************************************************/
p15_crl_t* p15_get_crl(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_crl_t* crl;
	p15_object_t* o;

	if (!ctx || !id) return NULL;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_CRL) {
			crl = (p15_crl_t*)o->obj;
			if (!p15_id_cmp(id, &crl->id)) return crl;
		}
		o = o->next;
	}
	return NULL;
}
/****************************************************************************/
p15_crl_t* p15_get_next_crl(p15_ctx_t* ctx, p15_object_t** it) 
{
	p15_object_t* o;

	if (!ctx || !it) return NULL;
	o = *it;
	if (!o) 
		o = ctx->objects;
	else
		o = o->next;
	while (o) {
		if (o->type == P15_OT_CRL) 
			break;
		o = o->next;
	}
	*it = o;
	if (o) return (p15_crl_t*)o->obj;
	return NULL;
}
/****************************************************************************/
int p15_add_oiddo(p15_ctx_t* ctx, char* label, const char* oidst, 
	p15_oid_t* oid, uint8_t* data, size_t len, int to_encrypt, 
	p15_info_t** errors)
{
	int rc = 0;
	p15_oiddo_t* oiddo = NULL;

	if (!ctx || (oid && oidst) || (!oid && !oidst) || 
		(data && !len) || (!data && len)) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, "p15_add_oiddo");
		return rc;
	}

#ifdef TESTOUT
	printf("p15_add_oiddo: Adding abstract data object '%s':\n", label ? label: "");
	if (oidst) printf("oid: %s\n", oidst);
	else {
		char oid1[201];
		oid1[0] = 0;
		p15_oid_to_str(oid, oid1, 200);
		printf("oid: %s\n", oid1);
	}
	printf("to_encrypt: %d\n", to_encrypt);
	printf("data (%d bytes):\n", (int)len);
	p15_dump(data, len);
	printf("\n");
#endif

	oiddo = p15_malloc(sizeof(p15_oiddo_t));
	oiddo->to_encrypt = to_encrypt;
	if (len) {
		oiddo->len = len;
		oiddo->ptr = p15_malloc(len);
		memcpy(oiddo->ptr, data, len);
	}
	if (oidst) 
		oiddo->oid = p15_new_oid_from_str(oidst);
	else
		oiddo->oid = p15_copy_oid(oid);
	if (!oiddo->oid) {
		rc = P15_RC_INVALID_OID_STR;
		p15_add_error(errors, P15_ERROR, rc, "p15_add_oiddo");
		goto end;
	}
	if (label) {
		if (strnlen(label, P15_LABEL_MAX + 1) > P15_LABEL_MAX) {
			rc = P15_RC_LABEL_TOO_LONG;
			p15_add_error(errors, P15_ERROR, rc, NULL);
			goto end;
		}
		strcpy(oiddo->label, label);
	}

	p15_add_object(ctx, P15_OT_OIDDO, oiddo);
	oiddo = NULL;
end:
	p15_free_oiddo(&oiddo);
	return rc;
}
/****************************************************************************/
p15_oiddo_t* p15_get_next_oiddo(p15_ctx_t* ctx, p15_object_t** it, 
	p15_oid_t* oid)
{
	p15_object_t* o;
	p15_oiddo_t* oiddo = NULL;

	if (!ctx || !it) return NULL;
	o = *it;
	if (!o) 
		o = ctx->objects;
	else
		o = o->next;
	while (o) {
		if (o->type != P15_OT_OIDDO) { o = o->next; continue; }
		if (oid) {
			oiddo = (p15_oiddo_t*)o->obj;
			if (!oiddo) { *it = NULL; return NULL;}
			if (p15_oid_cmp(oiddo->oid, oid)) {
				o = o->next;
				continue;
			}
		}
		break;
	}
	*it = o;
	if (o) return (p15_oiddo_t*)o->obj;
	*it = NULL;
	return NULL;
}
/****************************************************************************/
static int p15_id_to_ostr(p15_id_t* id, OCTET_STRING_t* ostr)
{
	uint8_t eid[P15_EXT_ID_MAX];
	int sz = 0;

	if (!id || !ostr) return P15_RC_INVALID_ARGS;
	if (!id->ext_id_len) {
		sz = 4;
		eid[0] = id->id >> 24;
		eid[1] = id->id >> 16;
		eid[2] = id->id >> 8;
		eid[3] = id->id;
	} else {
		sz = id->ext_id_len;
		memcpy(eid, id->ext_id, sz);
	}

	if (OCTET_STRING_fromBuf(ostr, (void*)eid, sz)) 
		return P15_RC_ASN1C_ERROR;

	return 0;
}
/****************************************************************************/
static int p15_process_obj_id(p15_ctx_t* ctx, P15_Identifier_t* asnid, int objtype, 
	p15_id_t* objid)
/* ret: inv args, cert_id/crl_id/key_id/id in use, inv token, int err 
 * objtype - P15_OT_ const
 */
{
	int rc = 0;
	p15_id_t* id = NULL;
	int rc2;

	if (!ctx || !asnid || !objid) return P15_RC_INVALID_ARGS;
	if (!asnid->buf || asnid->size < 1 || asnid->size > P15_EXT_ID_MAX) 
		return P15_RC_INVALID_TOKEN;
	id = p15_new_ext_id(asnid->buf, asnid->size);
	if (!id) return P15_RC_INTERNAL_ERROR;
	p15_refine_id(id);
	
	switch (objtype) {
	case P15_OT_CERT:
	case P15_OT_TRUSTED_CERT:
		if (p15_get_cert(ctx, id)) {
			rc = P15_RC_CERT_ID_IN_USE;
			goto end;
		}
		if (p15_get_secret_key(ctx, id)) {
			rc = P15_RC_ID_IN_USE;
			goto end;
		}
		break;

	case P15_OT_PRIVKEY:
		if (p15_get_private_key(ctx, id)) {
			rc = P15_RC_PRIVATE_KEY_ID_IN_USE;
			goto end;
		}
		if (p15_get_secret_key(ctx, id)) {
			rc = P15_RC_ID_IN_USE;
			goto end;
		}
		break;

	case P15_OT_CRL:
		if (p15_get_crl(ctx, id)) {
			rc = P15_RC_CRL_ID_IN_USE;
			goto end;
		}
		if (p15_get_secret_key(ctx, id)) {
			rc = P15_RC_ID_IN_USE;
			goto end;
		}
		break;

	case P15_OT_PUBKEY:
		if (p15_get_public_key(ctx, id)) {
			rc = P15_RC_PUBLIC_KEY_ID_IN_USE;
			goto end;
		}
		if (p15_get_secret_key(ctx, id)) {
			rc = P15_RC_ID_IN_USE;
			goto end;
		}
		break;

	case P15_OT_SECKEY:
		rc2 = p15_ck_id_exist(ctx, id);
		if (rc2 > 1) {
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc2) {
			rc = P15_RC_ID_IN_USE;
			goto end;
		}
		break;

	default: 
		rc = P15_RC_INVALID_ARGS;
		goto end;
	}
	if (!id->ext_id_len && id->id >= ctx->free_id) 
		ctx->free_id = id->id + 1;

	*objid = *id;
end:
	p15_free_id(&id);
	return rc;
}
/****************************************************************************/
static int p15_mk_kekri(p15_ctx_t* ctx, P15_RecipientInfo_t** pri, 
	uint8_t **pkey, uint8_t **pkmsk, uint8_t* pwkey, uint8_t* pwmsk, 
	p15_info_t** errors)
{/* only internal errors allowed */
	uint8_t *key;
	uint8_t *kmsk;
	int rc = 0;
	P15_RecipientInfo_t* ri = NULL;
	P15_KEKRecipientInfo_t* kekri;
	p15_id_t* kekid = NULL;
	P15_Gost28147_89_KeyWrapParameters_t* aparams = NULL;
	P15_Gost28147_89_EncryptedKey_t* aekey= NULL;
	uint8_t ukm[8];
	uint8_t okey[32];
	uint8_t pwmkey[32];
	uint8_t mac[4];
	uint8_t ekey[32];
	uint8_t* eokey = NULL;
	size_t eokeylen;
	int i;
	char st[500];

	/* Generate new key */
	key = p15_malloc(P15_PAGE);
	*pkey = key; 
	kmsk = p15_malloc(P15_PAGE);
	*pkmsk = kmsk;
	rc = p15_get_random(ctx, key, 32);
	if (!rc) rc = p15_get_random(ctx, kmsk, 32);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri: RNG failed");
		return rc;
	}

	/* Filling the RecipientInfo */
	ri = p15_malloc(sizeof(P15_RecipientInfo_t));
	*pri = ri;
	ri->present = P15_RecipientInfo_PR_kekri;
	kekri = &ri->choice.kekri;

	/* kekri */
	if (asn_long2INTEGER(&kekri->version, 4)) { /* version = 4 */
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_kekri:asn_long2INTEGER:version");
		return rc;
	}

	/* kekri.kekid.keyIdentifier */
	kekid = p15_new_id(ctx->free_id);
#ifdef TESTOUT
	printf("kekri.kekid.keyIdentifier:\n");
	p15_dump_id(kekid);
#endif
	rc = p15_id_to_ostr(kekid, &kekri->kekid.keyIdentifier);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:p15_id_to_ostr: kekid");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* kekri->keyEncryptionAlgorithm */

	/* kekri->keyEncryptionAlgorithm.algorithm */

#ifdef TESTOUT
	printf("kekri.keyEncryptionAlgorithm.algorithm: %s\n", P15_OID_GOST_KEY_WRAP);
#endif
	rc = p15_str_to_oid(P15_OID_GOST_KEY_WRAP, 
		&kekri->keyEncryptionAlgorithm.algorithm);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:p15_str_to_oid: kekri encr_alg");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* kekri->keyEncryptionAlgorithm.parameters */

	aparams = p15_malloc(sizeof(P15_Gost28147_89_KeyWrapParameters_t));

	/* aparams->encryptionParamSet */
	
#ifdef TESTOUT
	printf(
"kekri.keyEncryptionAlgorithm.parameters.encryptionParamSet (1-4,5 = cproA-D,Z): %d\n",
		ctx->enc_uz);
#endif

	rc = p15_encuz_to_oid(ctx->enc_uz, &aparams->encryptionParamSet);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:p15_encuz_to_oid: aparams->encryptionParamSet");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* aparams->ukm */
	rc = p15_get_random(ctx, ukm, 8);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:p15_get_random: ukm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("kekri.keyEncryptionAlgorithm.parameters.ukm:\n");
	p15_dump(ukm, 8);
#endif
	aparams->ukm = p15_malloc(sizeof(OCTET_STRING_t));
	if (OCTET_STRING_fromBuf(aparams->ukm, (void*)ukm, 8)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_kekri:OCTET_STRING_fromBuf:aparams->ukm");
		goto end;
	}

	/* set ANY */

	kekri->keyEncryptionAlgorithm.parameters = ANY_new_fromType(
		&asn_DEF_P15_Gost28147_89_KeyWrapParameters, aparams);
	if (!kekri->keyEncryptionAlgorithm.parameters) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_kekri:ANY_new_fromType failed for Gost28147_89_KeyWrapParameters");
		goto end;
	}

	/* kekri->encryptedKey */
	for (i = 0; i < 32; i++) {
		pwmkey[i] = pwkey[i] ^ pwmsk[i];
		okey[i] = key[i] ^ kmsk[i];
	}
	rc = gostc_sym_keywrap(okey, pwmkey, ukm, ctx->enc_uz, ekey, mac);
#ifdef TESTOUT
	printf("Key wrap:\n");
	printf("Pw key:\n");
	p15_dump(pwmkey, 32);
	printf("Session key:\n");
	p15_dump(okey, 32);
	printf("Wrapped key:\n");
	p15_dump(ekey, 32);
	printf("MAC of key:\n");
	p15_dump(mac, 4);
#endif
	gostc_cleanse(okey, 32);
	gostc_cleanse(pwmkey, 32);
	if (rc) {
		sprintf(st, "p15_mk_kekri:p15_keywrap: rc=%d", rc);
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}

	aekey = p15_malloc(sizeof(P15_Gost28147_89_EncryptedKey_t));
	
	/* aekey->encryptedKey */
	rc = OCTET_STRING_fromBuf(&aekey->encryptedKey, (void*)ekey, 32);
	/* aekey->macKey */
	if (!rc) rc = OCTET_STRING_fromBuf(&aekey->macKey, (void*)mac, 4);
	if (rc) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:OCTET_STRING_fromBuf failed for aekey");
		goto end;
	}

	/* Encode aekey */

	rc = p15_asn_encode(&asn_DEF_P15_Gost28147_89_EncryptedKey, aekey, 
		&eokey, &eokeylen, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:p15_asn_encode failed for aekey");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* Set kekri->encryptedKey */
	if (OCTET_STRING_fromBuf(&kekri->encryptedKey, (void*)eokey, eokeylen)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_kekri:OCTET_STRING_fromBuf failed for eokey");
		goto end;
	}
#ifdef TESTOUT
	printf("kekri.encryptedKey content:\n");
	p15_dump(eokey, eokeylen);
	printf("\n");
#endif
end:
	p15_free_id(&kekid);
	p15_free(eokey);
	ASN_STRUCT_FREE(asn_DEF_P15_Gost28147_89_KeyWrapParameters, aparams);
	ASN_STRUCT_FREE(asn_DEF_P15_Gost28147_89_EncryptedKey, aekey);
	return rc;
}
/****************************************************************************/
static int p15_get_pwkey(p15_ctx_t* ctx, uint8_t* pw, int pwlen, 
	uint8_t** pwkey, uint8_t** pwkmsk, uint8_t** salt)
{
	int rc = 0;
	int i;

	if (!ctx || !salt || !pwkey || !pwkmsk) 
		return P15_RC_INVALID_ARGS;

	/* Container format should be determined */
	if (ctx->p15_format != P15_FORMAT_2001 && ctx->p15_format != P15_FORMAT_2012) 
		return P15_RC_INTERNAL_ERROR;

	*pwkey = p15_malloc(P15_PAGE);
	*pwkmsk = p15_malloc(P15_PAGE);
	if (!*salt) {
		*salt = p15_malloc(ctx->p5_salt_len);
		rc = p15_get_random(ctx, *salt, ctx->p5_salt_len);
		if (rc) return rc;
	}

	rc = p15_get_safe_random(ctx, *pwkmsk, 32);
	if (rc) return rc;

	if (!pw) pwlen = 0;
	if (ctx->p15_format == P15_FORMAT_2012) 
		rc = gostc_GetPBKDF2_2012(pw, pwlen, *salt, ctx->p5_salt_len, ctx->p5_c,
			*pwkey, 32);
	else
		rc = gostc_GetPBKDF2(pw, pwlen, *salt, ctx->p5_salt_len, ctx->p5_c, *pwkey);
	if (rc) {
		gostc_cleanse(*pwkey, 32);
		return P15_RC_PKCS5_FAILED;
	}

	for (i = 0; i < 32; i++) (*pwkey)[i] ^= (*pwkmsk)[i];
#ifdef TESTOUT
	printf(
"p15_get_pwkey: Generating the password key from the password using PBKDF2 (%s).\n",
		(ctx->p15_format == P15_FORMAT_2001) ? "HMAC-GOST3411-94" :
			"HMAC-GOST3411-2012");
	if (pwlen) {
		printf("Input password:\n");
		p15_dump(pw, pwlen);
	} else {
		printf("Input password: <empty>\n");
	}
	printf("Iteration count: %d\n", ctx->p5_c);
	printf("Salt:\n");
	p15_dump(*salt, ctx->p5_salt_len);
	printf("Generated password key:\n");
	p15_dump_mkey32(*pwkey, *pwkmsk);
	printf("\n");
#endif
	return 0;
}
/****************************************************************************/
static int p15_mk_enveloped_data(p15_ctx_t* ctx, P15_EnvelopedData_t* edata,
	uint8_t* data, int len, uint8_t* pwkey, uint8_t* pwmsk, 
	p15_info_t** errors)
{
	int rc = 0;
	P15_RecipientInfo_t* ri = NULL;
	uint8_t *key = NULL;
	uint8_t *kmsk = NULL;
	P15_GostEncryptionImitAlgorithmIdentifier_t* ealg = NULL;
	P15_Gost28147_89_Parameters_t* gpar = NULL;
	uint8_t iv[8];
	uint8_t* encdata = NULL;
	void* mh;
	OCTET_STRING_t* eostr = NULL;

	if (!ctx || !edata || !data || !len || !pwkey || !pwmsk) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_enveloped_data");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}

	/* edata->version */
	if (asn_long2INTEGER(&edata->version, 2)) { /* version = 2 */
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_enveloped_data:asn_long2INTEGER for edata->version");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* edata->recipientInfos */
#ifdef TESTOUT
	printf("p15_mk_enveloped_data: Making EnvelopedData.recipientInfos (1 kekri):\n");
#endif
	rc = p15_mk_kekri(ctx, &ri, &key, &kmsk, pwkey, pwmsk, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_enveloped_data:p15_mk_kekri returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (asn_set_add(&edata->recipientInfos.list, ri)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_enveloped_data:asn_set_add for edata->recipientInfos");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	ri = NULL;

	/* edata->encryptedContentInfo */
	/* edata->encryptedContentInfo.contentType */

#ifdef TESTOUT
	printf("EnvelopedData.encryptedContentInfo.contentType: %s\n", P15_OID_ENC_CTYPE);
#endif
	rc = p15_str_to_oid(P15_OID_ENC_CTYPE, 
		&edata->encryptedContentInfo.contentType);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_enveloped_data:p15_str_to_oid: error for encryptedContentInfo.contentType");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* edata->encryptedContentInfo.contentEncryptionAlgorithm */
	ealg = &edata->encryptedContentInfo.contentEncryptionAlgorithm;

#ifdef TESTOUT
	printf(
"EnvelopedData.encryptedContentInfo.contentEncryptionAlgorithm.algorithm: %s\n",
		P15_OID_GOST_ENCR_IMIT);
	printf("Making EnvelopedData.encryptedContentInfo.contentEncryptionAlgorithm.parameters (Gost28147_89_Parameters):\n");
#endif
	/* ealg->algorithm */
	rc = p15_str_to_oid(P15_OID_GOST_ENCR_IMIT, &ealg->algorithm);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_enveloped_data:p15_str_to_oid: error for ealg->algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* ealg->parameters */
	gpar = p15_malloc(sizeof(asn_DEF_P15_Gost28147_89_Parameters)); 
	
	/* gpar->iv */
	rc = p15_get_random(ctx, iv, 8);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_enveloped_data:p15_get_random for iv");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("par.iv\n");
	p15_dump(iv, 8);
	printf("par.encryptionParamSet (1-4,5 - cproA-D,Z): %d\n", ctx->enc_uz);
#endif
	if (OCTET_STRING_fromBuf(&gpar->iv, (void*)iv, 8)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_enveloped_data:OCTET_STRING_fromBuf for iv");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
	/* gpar->encryptionParamSet */
	rc = p15_encuz_to_oid(ctx->enc_uz, &gpar->encryptionParamSet);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_enveloped_data:p15_encuz_to_oid for iv");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	ealg->parameters = ANY_new_fromType(&asn_DEF_P15_Gost28147_89_Parameters, 
		gpar);
	if (!ealg->parameters) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_enveloped_data:ANY_new_fromType failed for gpar");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Encoded Gost28147-89-Parameters:\n");
	p15_dump(ealg->parameters->buf, ealg->parameters->size);
#endif
	/* *edata->encryptedContentInfo.encryptedContent */

	/* encrypt */
	encdata = p15_malloc(len + 4);
	gostc_sym_EncryptCFB(data, encdata, iv, len, (void*)key, (void*)kmsk, 
		&gostc_sym_uz[ctx->enc_uz][0]);

	/* mac */
	mh = gostc_sym_MacOpen();
	gostc_sym_DoMac(mh, data, len, (void*)key, (void*)kmsk, 
		&gostc_sym_uz[ctx->enc_uz][0]);
	gostc_sym_MacClose(mh, encdata + len, (void*)key, (void*)kmsk, 
		&gostc_sym_uz[ctx->enc_uz][0]);

#ifdef TESTOUT
	printf("Making EnvelopedData.encryptedContentInfo.encryptedContent (enc data + imit) :\n");
	printf("EnvelopedData.encryptedContentInfo.encryptedContent (w/o ostr header):\n");
	p15_dump(encdata, len + 4);
	printf("\n");
#endif
	/* set *edata->encryptedContentInfo.encryptedContent */

	eostr = p15_malloc(sizeof(OCTET_STRING_t));
	if (OCTET_STRING_fromBuf(eostr, (void*)encdata, len + 4)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_enveloped_data:OCTET_STRING_fromBuf: eostr");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	edata->encryptedContentInfo.encryptedContent = eostr;
	eostr = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_RecipientInfo, ri);
	ASN_STRUCT_FREE(asn_DEF_P15_Gost28147_89_Parameters, gpar);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, eostr);
	p15_free(key);
	p15_free(kmsk);
	p15_free(encdata);
	return rc;
}
/****************************************************************************/
static int p15_mk_rng_do(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, 
	uint8_t* pwkey, uint8_t* pwkmsk, p15_info_t** errors)
{
	int rc = 0;
	P15_DataType_t* dt = NULL;
	P15_PKCS15ObjectOidDO_t* odo = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	char label[] = "Random Init Value";
	P15_EnvelopedData_t* edata = NULL;
	P15_RandomInitValue_t* rndinit = NULL;
	uint8_t* rndbuf = NULL;
	OCTET_STRING_t* morernd = NULL;
	uint8_t* ebuf = NULL;
	size_t ebuflen;

	if (!ctx || !dobjs || !pwkey || !pwkmsk) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_mk_rng_do");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}
	dt = p15_malloc(sizeof(P15_DataType_t));
	dt->present = P15_DataType_PR_oidDO;
	odo = &dt->choice.oidDO;

	/* odo->commonObjectAttributes	CommonObjectAttributes, */
	coa = &odo->commonObjectAttributes;

	/* *coa->label */
	coa->label = p15_malloc(sizeof(UTF8String_t));
	if (OCTET_STRING_fromBuf(coa->label, label, strlen(label))) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_rng_do: OCTET_STRING_fromBuf error for label");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Making RandomInitValue DataObject:\n");
	printf("Choosing DataType.oidDO choice.\n");
	printf("oidDO.commonObjectAttributes.label:\n");
	p15_dump(coa->label->buf, coa->label->size);
	printf("oidDO.commonObjectAttributes.flags: 0xC0, len=1, unused_bits=6\n");
#endif
	/* *coa->flags */
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0xC0; /* private, modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_rng_do:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;

	}

	/* odo->classAttributes CommonDataObjectAttributes,*/
	odo->classAttributes.applicationOID = p15_malloc(sizeof(OBJECT_IDENTIFIER_t));
	rc = p15_str_to_oid(P15_OID_RNG, odo->classAttributes.applicationOID);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
	"p15_mk_rng_do:p15_str_to_oid: error for odo->classAttributes.applicationOID");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->typeAttributes [1] OidDO */
	rc = p15_str_to_oid(P15_OID_RNG, &odo->typeAttributes.id);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_rng_do:p15_str_to_oid: error for odo->typeAttributes.id");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	odo->typeAttributes.value.present = P15_ObjectValueAny_PR_direct_protected;
	edata = &odo->typeAttributes.value.choice.direct_protected;
#ifdef TESTOUT
	printf("oidDO.classAttributes.applicationOID: %s\n", P15_OID_RNG);
	printf("oidDO.typeAttributes.id: %s\n", P15_OID_RNG);
	printf("Choosing oidDO.typeAttributes.value.direct-protected.\n");
	printf("Making RandomInitValue:\n");
#endif
	/* RandomInitValue */
	rndinit = p15_malloc(sizeof(P15_RandomInitValue_t));

	/* rndinit->randomInit */
	rndbuf = p15_malloc(64);
	if (ctx->rnd_new_fix_val) { /* from fixed value */
		memcpy(rndbuf, ctx->rnd_new_fix_val, 64);
	} else { /* random init from RNG */
		rc = p15_get_random(ctx, rndbuf, 64);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_rng_do:p15_get_random: error for rndbuf 64");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}
	if (OCTET_STRING_fromBuf(&rndinit->randomInit, (void*)rndbuf, 64)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_rng_do: OCTET_STRING_fromBuf error for rndinit");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("RandomInitValue.randomInit:\n");
	p15_dump(rndbuf, 64);
#endif	
	/* *rndinit->moreRandom */
	if (ctx->rnd_new_extra_len) {
		p15_free(rndbuf);
		rndbuf = p15_malloc(ctx->rnd_new_extra_len);
		if (ctx->rnd_new_fix_val) {
			memcpy(rndbuf, ctx->rnd_new_fix_val + 64, ctx->rnd_new_extra_len);
		} else {
			rc = p15_get_random(ctx, rndbuf, ctx->rnd_new_extra_len);
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_rng_do:p15_get_random: error for rndbuf extra");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
		}
		morernd = p15_malloc(sizeof(OCTET_STRING_t));
		if (OCTET_STRING_fromBuf(morernd, (void*)rndbuf, 
			ctx->rnd_new_extra_len)) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_rng_do: OCTET_STRING_fromBuf error for morernd");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		rndinit->moreRandom = morernd;
		morernd = NULL;
#ifdef TESTOUT
		printf("RandomInitValue.moreRandom:\n");
		p15_dump(rndbuf, ctx->rnd_new_extra_len);
#endif
	}

	/* encode rndinit */
	rc = p15_asn_encode(&asn_DEF_P15_RandomInitValue, rndinit, &ebuf, &ebuflen, 
		errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_rng_do:p15_asn_encode failed for rndinit");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Encoded RandomInitValue:\n");
	p15_dump(ebuf, ebuflen);
	printf("Making EnvelopedData of RandomInitValue:\n");
#endif
	rc = p15_mk_enveloped_data(ctx, edata, ebuf, ebuflen, pwkey, pwkmsk, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_rng_do:p15_mk_enveloped_data failed");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* inserting the data object */
	if (asn_sequence_add(&dobjs->choice.objectsd.list, dt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_rng_do:asn_sequence_add failed for dt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	dt = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_DataType, dt);
	ASN_STRUCT_FREE(asn_DEF_P15_RandomInitValue, rndinit);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, morernd);
	p15_free(rndbuf);
	p15_free(ebuf);
	return rc;
}
/****************************************************************************/
static int p15_mk_factor_version(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, 
	p15_info_t** errors)
{
	int rc = 0;
	P15_DataType_t* dt = NULL;
	P15_PKCS15ObjectOidDO_t* odo = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	char label[] = "Factor-TS version";
	P15_FactorTSVersion_t* ftsver = NULL;
	int ver_maj = 0;
	int ver_min = 0;


	if (!ctx || !dobjs) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_factor_version");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}
	if (ctx->p15_format == P15_FORMAT_2001) {
		ver_maj = P15_2001_FTS_MAJOR_VERSION;
		ver_min = P15_2001_FTS_MINOR_VERSION;
	} else {
		ver_maj = P15_2012_FTS_MAJOR_VERSION;
		ver_min = P15_2012_FTS_MINOR_VERSION;
	}

#ifdef TESTOUT
	printf("Making Factor-TS version DataObject:\n");
	printf("Setting DataType.oidDO choice.\n");
#endif
	dt = p15_malloc(sizeof(P15_DataType_t));
	dt->present = P15_DataType_PR_oidDO;
	odo = &dt->choice.oidDO;

	/* odo->commonObjectAttributes	CommonObjectAttributes, */
	coa = &odo->commonObjectAttributes;

	/* *coa->label */
	coa->label = p15_malloc(sizeof(UTF8String_t));
	if (OCTET_STRING_fromBuf(coa->label, label, strlen(label))) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
	"p15_mk_factor_version: OCTET_STRING_fromBuf error for label");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("oidDO.commonObjectAttributes.label:\n");
	p15_dump(coa->label->buf, coa->label->size);
	printf("oidDO.commonObjectAttributes.flags: 0x00, size=1, unused_bits=6\n");
#endif

	/* *coa->flags */
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0; /* non-private, non-modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_factor_version:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->classAttributes CommonDataObjectAttributes, */
	odo->classAttributes.applicationOID = 
		p15_malloc(sizeof(OBJECT_IDENTIFIER_t));
	rc = p15_str_to_oid(P15_OID_FACTOR_VER,	
		odo->classAttributes.applicationOID);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_factor_version:p15_str_to_oid: error for odo->classAttributes.applicationOID");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->typeAttributes [1] OidDO */
	rc = p15_str_to_oid(P15_OID_FACTOR_VER, &odo->typeAttributes.id);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_factor_version:p15_str_to_oid: error for odo->typeAttributes.id");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("oidDO.classAttributes.applicationOID: %s\n", P15_OID_FACTOR_VER);
	printf("oidDO.typeAttributes.id: %s\n", P15_OID_FACTOR_VER);
	printf("Choosing oidDO.typeAttributes.value.direct choice.\n");
	printf("FactorTSVersion.majorVersion: %d\n", ver_maj);
	printf("FactorTSVersion.minorVersion: %d\n", ver_min);
#endif
	odo->typeAttributes.value.present = P15_ObjectValueAny_PR_direct;

	/* FactorTSVersion */
	ftsver = p15_malloc(sizeof(P15_FactorTSVersion_t));

	/* ftsver->majorVersion */
	if (asn_long2INTEGER(&ftsver->majorVersion, ver_maj)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_factor_version:asn_long2INTEGER:version");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* ftsver->minorVersion */
	ftsver->minorVersion = p15_malloc(sizeof(INTEGER_t));
	if (asn_long2INTEGER(ftsver->minorVersion, ver_min)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_factor_version:asn_long2INTEGER:version (minor)");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* &odo->typeAttributes.value.choice.direct */
	rc = ANY_fromType(&odo->typeAttributes.value.choice.direct, 
		&asn_DEF_P15_FactorTSVersion, ftsver);
	if (rc) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_factor_version:ANY_fromType failed for ftsver");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Encoded oidDO.typeAttributes.value.direct:\n");
	p15_dump(odo->typeAttributes.value.choice.direct.buf, 
		odo->typeAttributes.value.choice.direct.size);
	printf("\n");
#endif
	/* insert dt */
	if (asn_sequence_add(&dobjs->choice.objectsd.list, dt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_factor_version:asn_sequence_add failed for dt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	dt = NULL;

end:
	ASN_STRUCT_FREE(asn_DEF_P15_FactorTSVersion, ftsver);
	ASN_STRUCT_FREE(asn_DEF_P15_DataType, dt);
	return rc;
}
/****************************************************************************/
static int p15_mk_prkey_obj(p15_ctx_t* ctx, P15_PrivateKeys_t* prkeys, 
	p15_private_key_t* pk, uint8_t* pwkey, uint8_t* pwkmsk, int* warns, 
	p15_info_t** errors)
/* ret int err, P15_RC_NO_PRIVATE_KEY_DATA, P15_RC_CRYPTOLIB_MASK2001 */
{
	int rc = 0;
	P15_PrivateKeyType_t* pkt = NULL;
	P15_PKCS15ObjectGost01PrivateKey_t* pko2001 = NULL;
	P15_PKCS15ObjectGost12PrivateKey_t* pko2012 = NULL;
	P15_CommonObjectAttributes_t* coa;
	P15_CommonKeyAttributes_t* cka;
	P15_CommonPrivateKeyAttributes_t* cpka;
	P15_Name_t* sbj = NULL;
	P15_ObjectValueGostPrivateKey_t* ovgpk = NULL;
	P15_KeyInfoPrivateGost01_t* ki2001 = NULL;
	P15_KeyInfoPrivateGost12_t* ki2012 = NULL;
	P15_EnvelopedData_t* edata = NULL;
	P15_GostPrivateKey_t* gpk = NULL;
	char st[4096];
	int i; /*j;*/
	uint8_t* ebuf = NULL;
	size_t ebuflen = 0;
	P15_Gost01KeyParameters_t* pars2001 = NULL;
	P15_Gost12KeyParameters_t* pars2012 = NULL;
	uint8_t ka = 0;
	uint8_t* keysbuf = NULL;
	int is2012;
	int is512;

	if (!ctx || !prkeys || !pk || !pwkey || !pwkmsk || !warns) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_mk_prkey_obj");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}
	is2012 = (ctx->p15_format == P15_FORMAT_2001) ? 0 : 1;
	is512 = p15_pkey_is_2012(pk->key_parameters);

	if (!pk->key_masks || pk->mask_count < 1) rc = 1;
	if (!rc) { if (!pk->key_masks[0]) rc = 1; }
	if (rc)	{
		rc = P15_RC_NO_PRIVATE_KEY_DATA;
		strcpy(st, "p15_mk_prkey_obj: No key/mask material set for private key '");
		strcat(st, pk->label[0] ? pk->label : "NO_LABEL");
		strcat(st, "'");
		p15_add_error(errors, P15_ERROR, rc, st);
		return rc;
	}

	/* remasking */
	if (pk->to_remask) {
		rc = p15_remask_private_key(ctx, pk, pk->mask_count);
		if (rc) {
			int ierr = 0;
			if (rc != P15_RC_CRYPTOLIB_MASK) ierr = 1;
			sprintf(st, 
"p15_mk_prkey_obj: p15_remask_private_key returned error for private key '%s'", 
				pk->label);
			p15_add_error(errors, ierr ? P15_INTERNAL_ERROR: P15_ERROR, rc, st);
			return rc;
		}
	}

#ifdef TESTOUT
	printf("p15_mk_prkey_obj: Making private key '%s' object:\n", pk->label);
	printf("Choosing PrivateKeyType.privateGostR3410_%sKey choice.\n", 
		is2012 ? "2012" : "2001");
#endif
	pkt = p15_malloc(sizeof(P15_PrivateKeyType_t));
	if (is2012) {
		pkt->present = P15_PrivateKeyType_PR_privateGostR3410_2012Key;
		pko2012 = &pkt->choice.privateGostR3410_2012Key;
	} else {
		pkt->present = P15_PrivateKeyType_PR_privateGostR3410_2001Key;
		pko2001 = &pkt->choice.privateGostR3410_2001Key;
	}

	/* pko->commonObjectAttributes */
	coa = is2012 ? &pko2012->commonObjectAttributes : &pko2001->commonObjectAttributes;

	/* *coa->label */
	if (pk->label[0]) {
		coa->label = p15_malloc(sizeof(P15_Label_t));
		if (OCTET_STRING_fromBuf(coa->label, pk->label, strlen(pk->label))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_prkey_obj: OCTET_STRING_fromBuf error for label");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("key.commonObjectAttributes.label:\n");
		p15_dump(coa->label->buf, coa->label->size);
#endif
	}
#ifdef TESTOUT
	printf("key.commonObjectAttributes.flags: 0x80, len=1, unused_bits=6\n");
#endif

	/* *coa->flags */
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0x80; /* private, non-modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_prkey_obj:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* pko->classAttributes		CommonKeyAttributes, */
	cka = is2012 ? &pko2012->classAttributes : &pko2001->classAttributes;

	/* cka->iD */
	rc = p15_id_to_ostr(&pk->id, &cka->iD);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_prkey_obj:p15_id_to_ostr returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* cka->usage */
	cka->usage.buf = p15_malloc(2);
	cka->usage.size = 2;
	cka->usage.bits_unused = 6;
	cka->usage.buf[0] = pk->key_usage >> 8;
	cka->usage.buf[1] = (uint8_t)pk->key_usage;
	rc = p15_pack_bitstring(&cka->usage);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_prkey_obj:p15_pack_bitstring: error for cka->usage");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("key.classAttributes.iD:\n");
	p15_dump(cka->iD.buf, cka->iD.size);
	printf("key.classAttributes.usage: len=2, unused_bits=6:\n");
	p15_dump(cka->usage.buf, cka->usage.size);
#endif
	/* *cka->accessFlags */
	if (!pk->key_access_flags) goto start_date;
	ka = pk->key_access_flags;
	ka &= P15_KA_MASK & 0xFE;
	cka->accessFlags = p15_malloc(sizeof(P15_KeyAccessFlags_t));
	cka->accessFlags->buf = p15_malloc(1);
	cka->accessFlags->size = 1;
	cka->accessFlags->bits_unused = 3;
	cka->accessFlags->buf[0] = ka;
	rc = p15_pack_bitstring(cka->accessFlags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_prkey_obj:p15_pack_bitstring: error for cka->accessFlags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("key.classAttributes.accessFlags: len=1, unused_bits=3: 0x%.2X\n", ka);
#endif

	/* *cka->startDate */
start_date:
	if (pk->start_date) {
		cka->startDate = p15_malloc(sizeof(GeneralizedTime_t));
		rc = p15_utctime_to_gt(cka->startDate, pk->start_date);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_prkey_obj:p15_utctime_to_gt returned error for pk->start_date");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("key.classAttributes.startDate:\n");
		p15_dump(cka->startDate->buf, cka->startDate->size);
#endif
	}

	/* *cka->endDate */
	if (pk->end_date) {
		cka->endDate = p15_malloc(sizeof(GeneralizedTime_t));
		rc = p15_utctime_to_gt(cka->endDate, pk->end_date);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_prkey_obj:p15_utctime_to_gt returned error for pk->end_date");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("key.classAttributes.endDate:\n");
		p15_dump(cka->endDate->buf, cka->endDate->size);
#endif
	}

	/* *pko->subClassAttributes */
	/* *cpka->subjectName */
	if (pk->subject) {
		cpka = p15_malloc(sizeof(P15_CommonPrivateKeyAttributes_t));
		if (is2012) pko2012->subClassAttributes = cpka;
		else pko2001->subClassAttributes = cpka;
		rc = p15_asn_decode(pk->subject->ptr, pk->subject->len, &asn_DEF_P15_Name, 
			(void**)&sbj, errors);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_prkey_obj:p15_asn_decode returned error for subject");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		cpka->subjectName = sbj;
		sbj = NULL;
#ifdef TESTOUT
		printf("key.subClassAttributes.subjectName:\n");
		p15_dump(pk->subject->ptr, pk->subject->len);
#endif
	} else {
		*warns = 1;
		strcpy(st, "p15_mk_prkey_obj: No subject specified for private key '");
		strcat(st, pk->label[0] ? pk->label : "NO_LABEL");
		strcat(st, "'");
		p15_add_error(errors, P15_WARNING, P15_RC_NO_SUBJECT_FOR_PR_KEY, st);
	}
	
	/* pko->typeAttributes		[1] PrivateGost(01|12)KeyAttributes */
	if (is2012) {
		ovgpk = &pko2012->typeAttributes.value;
		ki2012 = p15_malloc(sizeof(P15_KeyInfoPrivateGost12_t));
		pko2012->typeAttributes.keyInfo = ki2012;
	} else {
		ovgpk = &pko2001->typeAttributes.value;
		ki2001 = p15_malloc(sizeof(P15_KeyInfoPrivateGost01_t));
		pko2001->typeAttributes.keyInfo = ki2001;
	}

	/* ovgpk */
	ovgpk->present = P15_ObjectValueGostPrivateKey_PR_direct_protected;
	edata = &ovgpk->choice.direct_protected;
	
	/* GostPrivateKey */
	gpk = p15_malloc(sizeof(P15_GostPrivateKey_t));
	keysbuf = p15_malloc(pk->mask_count << (is512 ? 6 : 5));

	/* Remaining in Little-Endian */
	for (i = 0; i < (int)pk->mask_count; i++) 
		memcpy(keysbuf + (i << (is512 ? 6 : 5)), &pk->key_masks[i][0], 
			is512 ? 64 : 32 );
	if (OCTET_STRING_fromBuf(gpk, (char*)keysbuf, pk->mask_count << (is512 ? 6 : 5))) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_prkey_obj: OCTET_STRING_fromBuf failed for keysbuf");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* Encoding */
	rc = p15_asn_encode(&asn_DEF_P15_GostPrivateKey, gpk, &ebuf, &ebuflen, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_prkey_obj:p15_asn_encode returned error for gpk");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Encoded GostPrivateKey (%d bytes):\n", (int)ebuflen);
	p15_dump(ebuf, ebuflen);
	printf(
"Wrapping GostPrivateKey to EnvelopedData (key.typeAttributes.value.direct-protected:\n");
#endif

	/* mk envdata */
	rc = p15_mk_enveloped_data(ctx, edata, ebuf, ebuflen, pwkey, pwkmsk, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_prkey_obj: p15_mk_enveloped_data failed for gpk");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* ki */
#ifdef TESTOUT
	printf("Making key.typeAttributes.keyInfo.paramsAndOps.parameters:\n");
	if (is2012) {
		printf(
"pars.gostR3410-2012ParamSet (1-8 - cproA,B,C,XchA,XchB,512Test,512A,512B): %d\n",
			pk->key_parameters);
	} else {
		printf("pars.cryptoProParamSet (1-5 - cproA,B,C,XchA,XchB): %d\n",
			pk->key_parameters);
	}
	printf("\n");
#endif
	if (is2012) {
		ki2012->present = P15_KeyInfoPrivateGost12_PR_paramsAndOps12;
		pars2012 = &ki2012->choice.paramsAndOps12.parameters;
		pars2012->present = P15_Gost12KeyParameters_PR_gostR3410_2012ParamSet;
	} else {
		ki2001->present = P15_KeyInfoPrivateGost01_PR_paramsAndOps01;
		pars2001 = &ki2001->choice.paramsAndOps01.parameters;
		pars2001->present = P15_Gost01KeyParameters_PR_cryptoProParamSet;
	}

	/* pars->publicKeyParamSet */
	rc = p15_pkipar_to_oid(pk->key_parameters, is2012 ? 
		&pars2012->choice.gostR3410_2012ParamSet : 
		&pars2001->choice.cryptoProParamSet);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_prkey_obj:p15_pkipar_to_oid failed for pars->cryptoProParamSet/gostR3410_2012ParamSet");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	if (asn_sequence_add(&prkeys->choice.objectsk.list, pkt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_prkey_obj: asn_sequence_add failed for pkt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	pkt = NULL;
end:
	p15_free(ebuf);
	p15_free(keysbuf);
	ASN_STRUCT_FREE(asn_DEF_P15_PrivateKeyType, pkt);
	ASN_STRUCT_FREE(asn_DEF_P15_Name, sbj);
	ASN_STRUCT_FREE(asn_DEF_P15_GostPrivateKey, gpk);
	return rc;
}
/****************************************************************************/
static int p15_mk_pubkey_obj(p15_ctx_t* ctx, P15_PublicKeys_t* pubkeys, 
	p15_public_key_t* pbk, int* warns, p15_info_t** errors)
{
	P15_PublicKeyType_t* pkt = NULL;
	P15_PKCS15ObjectGost01PublicKey_t* pko2001 = NULL;
	P15_PKCS15ObjectGost12PublicKey_t* pko2012 = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CommonKeyAttributes_t* cka = NULL;
	P15_CommonPublicKeyAttributes_t* cpka = NULL;
	P15_ObjectValueGostPublicKeyChoice_t* ovgpk = NULL;
	P15_KeyInfoPublicGost01_t* ki2001 = NULL;
	P15_KeyInfoPublicGost12_t* ki2012 = NULL;
	P15_GostPublicKeyChoice_t* gpkc = NULL;
	OCTET_STRING_t* rgpk = NULL;
	P15_Gost01KeyParameters_t* pars2001 = NULL;
	P15_Gost12KeyParameters_t* pars2012 = NULL;
	int rc = 0;
	P15_Name_t* sbj = NULL;
	char st[4096];
	uint8_t keybuf[128];
	int is1024;
	int is2012;
	/* int i; */

	if (!ctx || !pubkeys || !pbk || !warns) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_mk_pubkey_obj");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}
	is2012 = (ctx->p15_format == P15_FORMAT_2001) ? 0 : 1;
	is1024 = p15_pkey_is_2012(pbk->key_parameters);

#ifdef TESTOUT
	printf("p15_mk_pubkey_obj: Making public key '%s' object:\n", pbk->label);
	printf("Choosing PublicKeyType.publicGostR3410_20%sKey choice.\n", 
		is2012 ? "12" : "01");
#endif

	pkt = p15_malloc(sizeof(P15_PublicKeyType_t));
	pkt->present = is2012 ? P15_PublicKeyType_PR_publicGostR3410_2012Key :
		P15_PublicKeyType_PR_publicGostR3410_2001Key;
	if (is2012) pko2012 = &pkt->choice.publicGostR3410_2012Key;
	else pko2001 = &pkt->choice.publicGostR3410_2001Key;

	/* pko->commonObjectAttributes */
	coa = is2012 ? &pko2012->commonObjectAttributes : &pko2001->commonObjectAttributes;

	/* *coa->label */
	if (pbk->label[0]) {
		coa->label = p15_malloc(sizeof(P15_Label_t));
		if (OCTET_STRING_fromBuf(coa->label, pbk->label, strlen(pbk->label))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_pubkey_obj: OCTET_STRING_fromBuf error for label");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("key.commonObjectAttributes.label:\n");
		p15_dump(coa->label->buf, coa->label->size);
#endif
	}
#ifdef TESTOUT
	printf("key.commonObjectAttributes.flags: 0x00, len=1, unused_bits=6\n");
#endif

	/* *coa->flags */
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0x00; /* non-private, non-modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_pubkey_obj:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* pko->classAttributes		CommonKeyAttributes, */
	cka = is2012 ? &pko2012->classAttributes : &pko2001->classAttributes;

	/* cka->iD */
	rc = p15_id_to_ostr(&pbk->id, &cka->iD);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_pubkey_obj:p15_id_to_ostr returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* cka->usage */
	cka->usage.buf = p15_malloc(2);
	cka->usage.size = 2;
	cka->usage.bits_unused = 6;
	cka->usage.buf[0] = pbk->key_usage >> 8;
	cka->usage.buf[1] = (uint8_t)pbk->key_usage;

	rc = p15_pack_bitstring(&cka->usage);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_pubkey_obj:p15_pack_bitstring: error for cka->usage");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}


#ifdef TESTOUT
	printf("key.classAttributes.iD:\n");
	p15_dump(cka->iD.buf, cka->iD.size);
	printf("key.classAttributes.usage: len=2, unused_bits=6:\n");
	p15_dump(cka->usage.buf, cka->usage.size);
#endif

	/* *cka->startDate */
	if (pbk->start_date) {
		cka->startDate = p15_malloc(sizeof(GeneralizedTime_t));
		rc = p15_utctime_to_gt(cka->startDate, pbk->start_date);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_pubkey_obj:p15_utctime_to_gt returned error for pbk->start_date");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("key.classAttributes.startDate:\n");
		p15_dump(cka->startDate->buf, cka->startDate->size);
#endif
	}

	/* *cka->endDate */
	if (pbk->end_date) {
		cka->endDate = p15_malloc(sizeof(GeneralizedTime_t));
		rc = p15_utctime_to_gt(cka->endDate, pbk->end_date);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_pubkey_obj:p15_utctime_to_gt returned error for pk->end_date");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("key.classAttributes.endDate:\n");
		p15_dump(cka->endDate->buf, cka->endDate->size);
#endif
	}

	/* *pko->subClassAttributes */
	/* *cpka->subjectName */
	if (pbk->subject) {
		cpka = p15_malloc(sizeof(P15_CommonPublicKeyAttributes_t));
		if (is2012) pko2012->subClassAttributes = cpka;
		else pko2001->subClassAttributes = cpka;

		rc = p15_asn_decode(pbk->subject->ptr, pbk->subject->len, &asn_DEF_P15_Name, 
			(void**)&sbj, errors);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_pubkey_obj:p15_asn_decode returned error for subject");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		cpka->subjectName = sbj;
		sbj = NULL;
#ifdef TESTOUT
		printf("key.subClassAttributes.subjectName:\n");
		p15_dump(pbk->subject->ptr, pbk->subject->len);
#endif
	} else {
		*warns = 1;
		strcpy(st, "p15_mk_pubkey_obj: No subject specified for public key '");
		strcat(st, pbk->label[0] ? pbk->label : "NO_LABEL");
		strcat(st, "'");
		p15_add_error(errors, P15_WARNING, P15_RC_NO_SUBJECT_FOR_PUB_KEY, st);
	}
	
	/* pko->typeAttributes */
	ovgpk = is2012 ? &pko2012->typeAttributes.value : &pko2001->typeAttributes.value;
	if (is2012) {
		ki2012 = pko2012->typeAttributes.keyInfo = 
			p15_malloc(sizeof(P15_KeyInfoPublicGost12_t));
	} else {
		ki2001 = pko2001->typeAttributes.keyInfo = 
			p15_malloc(sizeof(P15_KeyInfoPublicGost01_t));
	}

	/* ovgpk */
	ovgpk->present = P15_ObjectValueGostPublicKeyChoice_PR_direct;
	gpkc = &ovgpk->choice.direct;

	/* GostPublicKeyChoice - choosing raw */
	gpkc->present = P15_GostPublicKeyChoice_PR_raw;
	rgpk = &gpkc->choice.raw;

	/* remaining in little-endian */
	memcpy(keybuf, pbk->key, is1024 ? 128 : 64);
	if (OCTET_STRING_fromBuf(rgpk, (char*)keybuf, is1024 ? 128 : 64)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_mk_pubkey_obj: OCTET_STRING_fromBuf failed for keybuf");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Making key.typeAttributes.value:\n");
	printf("Choosing key.typeAttributes.value.direct.raw choice.\n");
	printf("GostR3410Point (w/o ostr header):\n");
	p15_dump(pbk->key, is1024 ? 128 : 64);
#endif

	/* ki */
#ifdef TESTOUT
	printf("Making key.typeAttributes.keyInfo.paramsAndOps.parameters:\n");
	if (is2012) printf(
"pars.gostR3410-2012ParamSet (1-8 - cproA,B,C,XchA,XchB,512Test,512A,512B): %d\n", 
			pbk->key_parameters);
	else printf("pars.cryptoProParamSet (1-5 - cproA,B,C,XchA,XchB): %d\n",
			pbk->key_parameters);
	printf("\n");
#endif
	if (is2012) {
		ki2012->present = P15_KeyInfoPublicGost12_PR_paramsAndOpsPub12;
		pars2012 = &ki2012->choice.paramsAndOpsPub12.parameters;
		pars2012->present = P15_Gost12KeyParameters_PR_gostR3410_2012ParamSet;
	} else {
		ki2001->present = P15_KeyInfoPublicGost01_PR_paramsAndOpsPub01;
		pars2001 = &ki2001->choice.paramsAndOpsPub01.parameters;
		pars2001->present = P15_Gost01KeyParameters_PR_cryptoProParamSet;
	}

	/* pars->cryptoProParamSet */
	rc = p15_pkipar_to_oid(pbk->key_parameters, 
		is2012 ? &pars2012->choice.gostR3410_2012ParamSet :
		&pars2001->choice.cryptoProParamSet);
	if (rc) {
		sprintf(st, "p15_mk_pubkey_obj:p15_pkipar_to_oid failed for pars->%sParamSet",
			is2012 ? "gostR3410-2012" : "cryptoPro");
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (asn_sequence_add(&pubkeys->choice.objectspk.list, pkt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_pubkey_obj: asn_sequence_add failed for pkt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	pkt = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_PublicKeyType, pkt);
	ASN_STRUCT_FREE(asn_DEF_P15_Name, sbj);
	return rc;
}
/****************************************************************************/
static int p15_mk_cert_obj(p15_ctx_t* ctx, P15_Certificates_t* certs, 
	p15_certificate_t* cer, int* warns, p15_info_t** errors)
/* ret: int err */
{
	int rc = 0;
	P15_CertificateType_t* ct = NULL;
	P15_PKCS15ObjectX509Certificate_t* x509 = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CommonCertificateAttributes_t* cca = NULL;
	P15_X509CertificateAttributes_t* xca = NULL;
	P15_Certificate_t* direct = NULL;
	P15_Certificate_t* tmpcer = NULL;
	P15_Name_t* tmprdn = NULL;
	char st[4096];

	if (!ctx || !certs || !cer || !warns) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_mk_cert_obj");
		return P15_RC_INTERNAL_ERROR;
	}
#ifdef TESTOUT
	printf("Making object for certificate '%s':\n", cer->label);
	printf("Choosing CertificateType.x509Certificate choice.\n");
#endif
	ct = p15_malloc(sizeof(P15_CertificateType_t));
	ct->present = P15_CertificateType_PR_x509Certificate;
	x509 = &ct->choice.x509Certificate;

	/* x509->commonObjectAttributes */
	coa = &x509->commonObjectAttributes;

	/* coa->label */
	if (cer->label[0]) {
		coa->label = p15_malloc(sizeof(P15_Label_t));
		if (OCTET_STRING_fromBuf(coa->label, cer->label, strlen(cer->label))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_cert_obj: OCTET_STRING_fromBuf error for label");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("x509.commonObjectAttributes.label:\n");
		p15_dump(coa->label->buf, coa->label->size); 
#endif
	}

	/* *coa->flags */
#ifdef TESTOUT
	printf("x509.commonObjectAttributes.flags: 0x00, len=1, unused_bits=6\n");
#endif
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0x00; /* non-private, non-modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_cert_obj:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* x509->classAttributes */
	cca = &x509->classAttributes;

	/* cca->iD */
	rc = p15_id_to_ostr(&cer->id, &cca->iD);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_cert_obj:p15_id_to_ostr returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("x509.classAttributes.iD:\n");
	p15_dump(cca->iD.buf, cca->iD.size);
#endif

	/* *cca->authority */
	if (cer->ca) {
		cca->authority = p15_malloc(sizeof(BOOLEAN_t));
		*cca->authority = 1;
#ifdef TESTOUT
		printf("x509.classAttributes.authority: true\n");
#endif
	}

	/* x509->typeAttributes */
	xca = &x509->typeAttributes;
	
	/* xca->value */
	xca->value.present = P15_ObjectValueCertificate_PR_direct;
	direct = &xca->value.choice.direct;
#ifdef TESTOUT
	printf("Choosing x509.typeAttributes.value.direct choice.\n");
	printf("Certificate (%d bytes):\n", (int)cer->len);
	p15_dump(cer->ptr, cer->len);
#endif

	/* decoding cert */
	rc = p15_asn_decode(cer->ptr, cer->len, &asn_DEF_P15_Certificate, (void**)&tmpcer, 
		errors);
	if (rc) {
		ASN_STRUCT_FREE(asn_DEF_P15_Certificate, tmpcer);
		tmpcer = NULL;
		sprintf(st, 
"p15_mk_cert_obj: Certificate '%s': p15_asn_decode returned error for cer->ptr", 
			cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* Copying to main structure */
	*direct = *tmpcer;
	p15_free(tmpcer);
	tmpcer = NULL;

	if (!cer->issuer || !cer->subject) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st, 
"p15_mk_cert_obj: Certificate '%s': Internal error: cer->issuer or cer->subject is NULL", 
			cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}

	/* *xca->subject */
#ifdef TESTOUT
	printf("x509.typeAttributes.subject:\n");
	p15_dump(cer->subject->ptr, cer->subject->len);
#endif
	rc = p15_asn_decode(cer->subject->ptr, cer->subject->len, &asn_DEF_P15_Name,
		(void**)&tmprdn, errors);
	if (rc) {
		ASN_STRUCT_FREE(asn_DEF_P15_Name, tmprdn);
		tmprdn = NULL;
		sprintf(st, 
"p15_mk_cert_obj: Certificate '%s': p15_asn_decode returned error for cer->subject", 
			cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	xca->subject = p15_malloc(sizeof(P15_Name_t));
	*xca->subject = *tmprdn;
	p15_free(tmprdn);
	tmprdn = NULL;

	/* *xca->issuer */
#ifdef TESTOUT
	printf("x509.typeAttributes.issuer:\n");
	p15_dump(cer->issuer->ptr, cer->issuer->len);
	printf("\n");
#endif
	rc = p15_asn_decode(cer->issuer->ptr, cer->issuer->len, &asn_DEF_P15_Name,
		(void**)&tmprdn, errors);
	if (rc) {
		ASN_STRUCT_FREE(asn_DEF_P15_Name, tmprdn);
		tmprdn = NULL;
		sprintf(st, 
"p15_mk_cert_obj: Certificate '%s': p15_asn_decode returned error for cer->issuer", 
			cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	xca->issuer = p15_malloc(sizeof(P15_Name_t));
	*xca->issuer = *tmprdn;
	p15_free(tmprdn);
	tmprdn = NULL;

	/* *xca->serialNumber */
	xca->serialNumber = p15_malloc(sizeof(INTEGER_t));
	xca->serialNumber->size = direct->tbsCertificate.serialNumber.size;
	xca->serialNumber->buf = p15_malloc(xca->serialNumber->size);
	memcpy(xca->serialNumber->buf, direct->tbsCertificate.serialNumber.buf,
		xca->serialNumber->size);

	if (asn_sequence_add(&certs->choice.objectsc.list, ct)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_cert_obj: asn_sequence_add failed for ct");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	ct = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_CertificateType, ct);
	return rc;
}
/****************************************************************************/
static int p15_mk_crl_obj(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, p15_crl_t* crl, 
	p15_info_t **errors)
{
	int rc = 0;
	P15_DataType_t* dt = NULL;
	P15_PKCS15ObjectOidDO_t* odo = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CRLContainer_t* crlc = NULL;
	P15_CertificateList_t* cl = NULL;

	if (!ctx || !dobjs || !crl) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS, 
			"p15_mk_crl_obj");
		return P15_RC_INTERNAL_ERROR;
	}
#ifdef TESTOUT
	printf("Making CRL object '%s'\n", crl->label);
	printf("Choosing DataType.oidDO choice.\n");
#endif
	dt = p15_malloc(sizeof(P15_DataType_t));
	dt->present = P15_DataType_PR_oidDO;
	odo = &dt->choice.oidDO;

	/* odo->commonObjectAttributes CommonObjectAttributes, */
	coa = &odo->commonObjectAttributes;

	/* *coa->label */
	if (crl->label[0]) {
		coa->label = p15_malloc(sizeof(UTF8String_t));
		if (OCTET_STRING_fromBuf(coa->label, crl->label, 
				strlen(crl->label))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
				"p15_mk_crl_obj: OCTET_STRING_fromBuf error for label");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("oidDO.commonObjectAttributes.label:\n");
		p15_dump(coa->label->buf, coa->label->size);
#endif
	}

	/* *coa->flags */
#ifdef TESTOUT
	printf("oidDO.commonObjectAttributes.flags: 0x00, len=1, unused_bits=6\n");
#endif
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0; /* non-private, non-modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_crl_obj:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->classAttributes CommonDataObjectAttributes, */
#ifdef TESTOUT
	printf("oidDO.classAttributes.applicationOID: %s\n", P15_OID_CRL);
	printf("oidDO.typeAttributes.id: %s\n", P15_OID_CRL);
	printf("Choosing oidDO.typeAttributes.value.direct choice.\n");
#endif
	odo->classAttributes.applicationOID = 
		p15_malloc(sizeof(OBJECT_IDENTIFIER_t));
	rc = p15_str_to_oid(P15_OID_CRL, odo->classAttributes.applicationOID);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_crl_obj:p15_str_to_oid: error for odo->classAttributes.applicationOID");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->typeAttributes [1] OidDO */
	rc = p15_str_to_oid(P15_OID_CRL, &odo->typeAttributes.id);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_crl_obj:p15_str_to_oid: error for odo->typeAttributes.id");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	odo->typeAttributes.value.present = P15_ObjectValueAny_PR_direct;


	/* CRLContainer	*/

	crlc = p15_malloc(sizeof(P15_CRLContainer_t));

	/* crlc->iD */
	rc = p15_id_to_ostr(&crl->id, &crlc->iD);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_crl_obj:p15_id_to_ostr returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Making CRLContainer structure:\n");
	printf("CRLContainer.iD:\n");
	p15_dump(crlc->iD.buf, crlc->iD.size);
	printf("CRLContainer.issuer:\n");
	p15_dump(crl->issuer->ptr, crl->issuer->len);
	printf("CRLContainer.crl (%d bytes):\n", (int)crl->len);
	p15_dump(crl->ptr, crl->len);
#endif

	/* *crlc->issuer */

	rc = p15_asn_decode(crl->issuer->ptr, crl->issuer->len, &asn_DEF_P15_Name, 
			(void**)&crlc->issuer, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_crl_obj:p15_asn_decode returned error for issuer");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}


	/* crlc->crl */
	rc = p15_asn_decode(crl->ptr, crl->len, &asn_DEF_P15_CertificateList, (void**)&cl, 
		errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_crl_obj:p15_asn_decode returned error for cl");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	crlc->crl = *cl;
	p15_free(cl);
	cl = NULL;

	/* encode */
	if(ANY_fromType(&odo->typeAttributes.value.choice.direct, 
		&asn_DEF_P15_CRLContainer, crlc)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_crl_obj:ANY_fromType failed for crlc");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Encoded oidDO.typeAttributes.value.direct (CRLContainer):\n");
	p15_dump(odo->typeAttributes.value.choice.direct.buf, 
		odo->typeAttributes.value.choice.direct.size);
	printf("\n");
#endif

	/* insert dt */
	if (asn_sequence_add(&dobjs->choice.objectsd.list, dt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_crl_obj:asn_sequence_add failed for dt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	dt = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_DataType, dt);
	ASN_STRUCT_FREE(asn_DEF_P15_CRLContainer, crlc);
	ASN_STRUCT_FREE(asn_DEF_P15_CertificateList, cl);
	return rc;
}
/****************************************************************************/
static int p15_mk_oiddo_obj(p15_ctx_t* ctx, P15_DataObjects_t* dobjs, 
	p15_oiddo_t* oiddo, uint8_t* pwkey, uint8_t* pwkmsk, 
	p15_info_t** errors)
/* ret: 0, int err, no obj data */
{
	int rc = 0;
	char oidst[1024];
	P15_DataType_t* dt = NULL;
	P15_PKCS15ObjectOidDO_t* odo = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	OCTET_STRING_t* ostr = NULL;
	uint8_t* ostrdata = NULL;
	size_t ostrdatalen; 

	if (!ctx || !dobjs || !oiddo || !pwkey || !pwkmsk) {
	rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_oiddo_obj");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}

	if (!oiddo->ptr || !oiddo->len) {
		rc = P15_RC_NO_OBJECT_DATA;
		sprintf(oidst, "p15_mk_oiddo_obj: p15_oiddo_t object '%s'.",
			oiddo->label);
		p15_add_error(errors, P15_ERROR, rc, oidst);
		goto end;
	}

	rc = p15_oid_to_str(oiddo->oid, oidst, sizeof(oidst));
	if (rc && rc != P15_RC_BUFFER_TOO_SMALL) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_oiddo_obj: p15_oid_to_str returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) strcpy(oidst, "Very long OID :)");
	rc = 0;

#ifdef TESTOUT
	printf("Making abstract data object '%s' \n", oiddo->label);
	printf("Choosing DataType.oidDO choice.\n");
#endif
	dt = p15_malloc(sizeof(P15_DataType_t));
	dt->present = P15_DataType_PR_oidDO;
	odo = &dt->choice.oidDO;

	/* odo->commonObjectAttributes CommonObjectAttributes, */
	coa = &odo->commonObjectAttributes;

	/* *coa->label */
	if (oiddo->label[0]) {
		coa->label = p15_malloc(sizeof(UTF8String_t));
		if (OCTET_STRING_fromBuf(coa->label, oiddo->label, 
				strlen(oiddo->label))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_mk_oiddo_obj: OCTET_STRING_fromBuf error for label");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("oidDO.commonObjectAttributes.label:\n");
		p15_dump(coa->label->buf, coa->label->size);
#endif
	}

	/* *coa->flags */
#ifdef TESTOUT
	printf("oidDO.commonObjectAttributes.flags: %s, len=1, unused_bits=6\n",
		oiddo->to_encrypt ? "0xC0" : "0x40");
#endif
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = oiddo->to_encrypt ? 0xC0 : 0x40;
	/* (non)private, modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_oiddo_obj:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->classAttributes CommonDataObjectAttributes, */
#ifdef TESTOUT
	printf("oidDO.classAttributes.applicationOID: %s\n", P15_OID_ABSTRACT_DATA);
	printf("oidDO.typeAttributes.id: %s\n", oidst);
	printf("Choosing oidDO.typeAttributes.value.direct%s choice.\n",
		oiddo->to_encrypt ? "-protected" : "");
#endif
	odo->classAttributes.applicationOID = 
		p15_malloc(sizeof(OBJECT_IDENTIFIER_t));
	rc = p15_str_to_oid(P15_OID_ABSTRACT_DATA, odo->classAttributes.applicationOID);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_oiddo_obj:p15_str_to_oid: error for odo->classAttributes.applicationOID");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* odo->typeAttributes [1] OidDO */
	rc = p15_str_to_oid(oidst, &odo->typeAttributes.id);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_oiddo_obj:p15_str_to_oid: error for odo->typeAttributes.id");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	odo->typeAttributes.value.present = oiddo->to_encrypt ? 
		P15_ObjectValueAny_PR_direct_protected :
		P15_ObjectValueAny_PR_direct;

	/* Wrapping to octet string */
	ostr = p15_malloc(sizeof(OCTET_STRING_t));
	if (OCTET_STRING_fromBuf(ostr, (void*)oiddo->ptr, oiddo->len)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_oiddo_obj:OCTET_STRING_fromBuf: ostr");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	if (!oiddo->to_encrypt) 
		goto not_encrypted;

	/* Encrypting */
	rc = p15_asn_encode(&asn_DEF_OCTET_STRING, ostr, &ostrdata, &ostrdatalen, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_oiddo_obj:p15_asn_encode: ostr");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("Data to encrypt (wrapped in OCTET STRING): (%d bytes):\n", (int)ostrdatalen);
	p15_dump(ostrdata, ostrdatalen);
	printf(
"Wrapping the data to EnvelopedData (oidDO.typeAttributes.value.direct-protected):\n");
#endif
	rc = p15_mk_enveloped_data(ctx, 
		&odo->typeAttributes.value.choice.direct_protected,
		ostrdata, ostrdatalen, pwkey, pwkmsk, errors);	
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_oiddo_obj:p15_mk_enveloped_data returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	goto insert_dt;		

not_encrypted:
	/* odo->typeAttributes.value.choice.direct */

	rc = ANY_fromType(&odo->typeAttributes.value.choice.direct, 
		&asn_DEF_OCTET_STRING, ostr);
	if (rc) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_oiddo_obj:ANY_fromType failed for ostr");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("Data (wrapped in OCTET STRING) oidDO.typeAttributes.value.direct:\n");
	p15_dump(odo->typeAttributes.value.choice.direct.buf, 
		odo->typeAttributes.value.choice.direct.size);
	printf("\n");
#endif

insert_dt:
	/* insert dt */
	if (asn_sequence_add(&dobjs->choice.objectsd.list, dt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_oiddo_obj:asn_sequence_add failed for dt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	dt = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_DataType, dt);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, ostr);
	p15_free(ostrdata);
	return rc;
}
/****************************************************************************/
static int p15_mk_encap_token(p15_ctx_t* ctx, uint8_t* pwkey, uint8_t* pwkmsk, 
	uint8_t* salt, OCTET_STRING_t** petoken, uint8_t* tokenhash, int* warns,
	p15_info_t** errors)
/* ret: int err, P15_RC_NO_PRIVATE_KEY_DATA, P15_RC_NO_OBJECT_DATA, crypto2001remask 
 * P15_RC_NO_SECRET_KEY_DATA
 */
{
	int rc = 0;
	P15_PKCS15Token_t* token = NULL;
	P15_KeyManagementInfo_t* infos = NULL;
	P15_KeyManagementInfoElem_t* info = NULL;
	p15_id_t* id = NULL;
	P15_PasswordInfo_t* pwinfo;
	P15_Label_t* pwhint = NULL;
	P15_PBKDF2_params_t* ppar = NULL;
	P15_HmacGostR3411Algorithm_t* aprf = NULL;
	P15_PKCS15Objects_t* pobjs = NULL;
	P15_DataObjects_t* dobjs = NULL;
	OCTET_STRING_t* ostr = NULL;
	uint8_t* ebuf = NULL;
	size_t ebuflen;
	void* hh;
	gostc_hash12_ctx_t* hh12 = NULL;
	p15_object_t* object = NULL;
	char* hmac_oid = NULL;
	int is2012;
	int rc2 = 0;

	if (!ctx || !pwkey || !pwkmsk || !salt || !petoken || !tokenhash || !warns) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_mk_encap_token");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}

	is2012 = (ctx->p15_format == P15_FORMAT_2001) ? 0 : 1;

	token = p15_malloc(sizeof(P15_PKCS15Token_t));

	/* token->version */
	if (asn_long2INTEGER(&token->version, 0)) { /* version = 0 */
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_encap_token:asn_long2INTEGER:version");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* *token->keyManagementInfo*/
	/* build info */
	info = p15_malloc(sizeof(P15_KeyManagementInfoElem_t));

	/* info->keyId */
	id = p15_new_id(ctx->free_id);
	rc = p15_id_to_ostr(id, &info->keyId);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:p15_id_to_ostr: info->keyId");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("Making PKCS15Token:\n\n");
	printf("token.keyManagementInfo.keyId:\n");
	p15_dump(info->keyId.buf, info->keyId.size);
#endif
	/* info->keyinfo */
	info->keyInfo.present = keyInfo_PR_passwordInfo;
	pwinfo = &info->keyInfo.choice.passwordInfo;

	/* pwinfo->hint */
	if (ctx->password_hint[0]) {
		pwhint = p15_malloc(sizeof(P15_Label_t));
		if (OCTET_STRING_fromBuf(pwhint, ctx->password_hint, 
			strlen(ctx->password_hint))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_mk_encap_token:OCTET_STRING_fromBuf: pwhint");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("token.keyManagementInfo.keyInfo.passwordInfo.hint:\n");
		p15_dump(pwhint->buf, pwhint->size);
#endif
		pwinfo->hint = pwhint;
		pwhint = NULL;
	}

	/* pwinfo->algId.algorithm */

#ifdef TESTOUT
	printf("token.keyManagementInfo.keyInfo.passwordInfo.algId.algorithm: %s\n",
		P15_OID_PBKDF2);
#endif
	rc = p15_str_to_oid(P15_OID_PBKDF2, &pwinfo->algId.algorithm);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_encap_token:p15_str_to_oid error for pwinfo->algId.algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* pwinfo->algId.parameters */
	
	ppar = p15_malloc(sizeof(P15_PBKDF2_params_t));
	ppar->salt.present = salt_PR_specified;
#ifdef TESTOUT
	printf(
"Making token.keyManagementInfo.keyInfo.passwordInfo.algId.parameters (PBKDF2-params):\n");
	printf("par.salt.specified:\n");
	p15_dump(salt, ctx->p5_salt_len);
#endif
	if (OCTET_STRING_fromBuf(&ppar->salt.choice.specified, (void*)salt,
		ctx->p5_salt_len)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:OCTET_STRING_fromBuf: ppar->salt");
		rc = P15_INTERNAL_ERROR;
		goto end;
	}

	if (asn_long2INTEGER(&ppar->iterationCount, ctx->p5_c)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:asn_long2INTEGER: ppar->iterationCount");
		rc = P15_INTERNAL_ERROR;
		goto end;
	}
	hmac_oid = is2012 ? P15_OID_HMAC_GOST_2012 : P15_OID_HMAC_GOST_94;
#ifdef TESTOUT
	printf("par.iterationCount:\n");
	p15_dump(ppar->iterationCount.buf, ppar->iterationCount.size);
	printf("par.keyLength: 32\n");
	printf("par.prf.algorithm: %s\n", hmac_oid);
#endif
	ppar->keyLength = p15_malloc(sizeof(INTEGER_t));
	if (asn_long2INTEGER(ppar->keyLength, 32)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:asn_long2INTEGER: ppar->keyLength");
		rc = P15_INTERNAL_ERROR;
		goto end;
	}

	aprf = p15_malloc(sizeof(P15_HmacGostR3411Algorithm_t));
	rc = p15_str_to_oid(hmac_oid, &aprf->algorithm);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_mk_encap_token:p15_str_to_oid error for aprf->algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	ppar->prf = aprf;
	aprf = NULL;

	/* encode to DER */
	pwinfo->algId.parameters = ANY_new_fromType(&asn_DEF_P15_PBKDF2_params, 
		ppar);
	if (!pwinfo->algId.parameters) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:ANY_new_fromType failed for ppar");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("token.keyManagementInfo.keyInfo.passwordInfo.algId.parameters (encoded):\n");
	p15_dump(pwinfo->algId.parameters->buf, pwinfo->algId.parameters->size);
#endif

	/* insert to infos */

	infos = p15_malloc(sizeof(P15_KeyManagementInfo_t));
	if (asn_sequence_add(&infos->list, info)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:asn_sequence_add failed for infos");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	info = NULL;

	token->keyManagementInfo = infos;
	infos = NULL;

	
	/* token->pkcs15Objects */

#ifdef TESTOUT
	printf("\nMaking token.pkcs15Objects:\n\n");
#endif

	/* adding DataObjects for RNG and Factor-TS ver */
	pobjs = p15_malloc(sizeof(P15_PKCS15Objects_t));
	pobjs->present = P15_PKCS15Objects_PR_dataObjects;
	dobjs = &pobjs->choice.dataObjects;
	if (asn_sequence_add(&token->pkcs15Objects.list, pobjs)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:asn_sequence_add failed for pobjs(data)");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	pobjs = NULL;
	dobjs->present = P15_PathOrObjectsDataType_PR_objectsd;
	/* Factor-TS version */
	rc = p15_mk_factor_version(ctx, dobjs, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:p15_mk_factor_version failed");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* RNG */
	rc = p15_mk_rng_do(ctx, dobjs, pwkey, pwkmsk, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:p15_mk_rng_do failed");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* Iterating objects */
	object = ctx->objects;
	while (object) {
		rc = p15_save_object(ctx, object, token, pwkey, pwkmsk, warns, errors);
		if (rc) break;
		object = object->next;
	}
	if (rc == P15_RC_INTERNAL_ERROR) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_encap_token:p15_save_object returned error");
		goto end;
	}
	if (rc) goto end;

	/* encode to ostring */

	rc = p15_asn_encode(&asn_DEF_P15_PKCS15Token, token, &ebuf, &ebuflen, 
		errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:p15_asn_encode failed for PKCS15Token");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	ostr = p15_malloc(sizeof(OCTET_STRING_t));
	*petoken = ostr;
	if (OCTET_STRING_fromBuf(ostr, (void*)ebuf, ebuflen)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:OCTET_STRING_fromBuf failed for petoken");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* calc hash */
	if (is2012) {
		hh12 = gostc_hash12_new();
		if (!hh12) rc2 = 1;
		if (!rc2) rc2 = gostc_hash12_init(hh12, 0);
		if (!rc2) rc2 = gostc_hash12_do(hh12, ebuf, ebuflen);
		if (!rc2) rc2 = gostc_hash12_finish(hh12, tokenhash);
		if (rc2) {
			rc = P15_RC_CRYPTOLIB_HASH;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_encap_token:One of gostc_hash12_* functions returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	} else {
		hh = gostc_hash_open();
		gostc_hash_do(hh, ebuf, ebuflen);
		gostc_hash_close(hh, tokenhash);
	}
end:
	p15_free(ebuf);
	p15_free_id(&id);
	ASN_STRUCT_FREE(asn_DEF_P15_PKCS15Token, token);
	ASN_STRUCT_FREE(asn_DEF_P15_KeyManagementInfo, infos);
	ASN_STRUCT_FREE(asn_DEF_P15_KeyManagementInfoElem, info);
	ASN_STRUCT_FREE(asn_DEF_P15_Label, pwhint);
	ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, ppar);
	ASN_STRUCT_FREE(asn_DEF_P15_HmacGostR3411Algorithm, aprf);
	ASN_STRUCT_FREE(asn_DEF_P15_PKCS15Objects, pobjs);
	gostc_hash12_clear(hh12);
	gostc_hash12_free(&hh12);

	return rc;
}
/****************************************************************************/
static int p15_save_object(p15_ctx_t* ctx, p15_object_t* object, 
	P15_PKCS15Token_t* token, uint8_t* pwkey, uint8_t* pwkmsk, 
	int* warns, p15_info_t** errors)
/* ret: 0, int err, no pr key data, no sec key data, no obj data, crypto2001mask */
{
	int rc = 0;
	P15_PKCS15Objects_t* pobjs = NULL;
	P15_PrivateKeys_t* prkeys = NULL;
	P15_PublicKeys_t* pubkeys = NULL;
	P15_Certificates_t* certs = NULL;
	P15_DataObjects_t* dobjs = NULL;
	P15_SecretKeys_t* skeys = NULL;
	char st[1024];
	p15_private_key_t* pk = NULL;
	p15_public_key_t* pbk = NULL;
	p15_certificate_t* cer = NULL;
	p15_crl_t* crl = NULL;
	p15_oiddo_t* oiddo = NULL;
	p15_secret_key_t* sk = NULL;

	if (!ctx || !object || !token || !pwkey || !pwkmsk || !warns) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_save_object");
		return rc;
	}

	if (!object->obj) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_OBJECT, 
			"p15_save_object: object->obj is NULL");
		goto end;
	}
	
	pobjs = p15_malloc(sizeof(P15_PKCS15Objects_t));
	switch (object->type) {
	case P15_OT_PRIVKEY:
#ifdef TESTOUT
		printf("\nAdding token.pkcs15Objects.privateKeys element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_privateKeys;
		prkeys = &pobjs->choice.privateKeys;
		prkeys->present = P15_PathOrObjectsPrivateKeyType_PR_objectsk;
		pk = (p15_private_key_t*)object->obj;
		break;

	case P15_OT_CERT:
#ifdef TESTOUT
		printf("\nAdding token.pkcs15Objects.certificates element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_certificates;
		certs = &pobjs->choice.certificates;
		certs->present = P15_PathOrObjectsCertificateType_PR_objectsc;
		cer = (p15_certificate_t*)object->obj;
		break;

	case P15_OT_TRUSTED_CERT:
#ifdef TESTOUT
		printf("\nAdding token.pkcs15Objects.trustedCertificates element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_trustedCertificates;
		certs = &pobjs->choice.trustedCertificates;
		certs->present = P15_PathOrObjectsCertificateType_PR_objectsc;
		cer = (p15_certificate_t*)object->obj;
		break;

	case P15_OT_CRL:
#ifdef TESTOUT
		printf("\nAdding token.pkcs15Objects.dataObjects (CRL) element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_dataObjects;
		dobjs = &pobjs->choice.dataObjects;
		dobjs->present = P15_PathOrObjectsDataType_PR_objectsd;
		crl = (p15_crl_t*)object->obj;
		break;

	case P15_OT_PUBKEY:
#ifdef TESTOUT
		printf("\nAdding token.pkcs15Objects.publicKeys element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_publicKeys;
		pubkeys = &pobjs->choice.publicKeys;
		pubkeys->present = P15_PathOrObjectsPublicKeyType_PR_objectspk;
		pbk = (p15_public_key_t*)object->obj;
		break;

	case P15_OT_OIDDO:
#ifdef TESTOUT
		printf(
"\nAdding token.pkcs15Objects.dataObjects (abstract data) element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_dataObjects;
		dobjs = &pobjs->choice.dataObjects;
		dobjs->present = P15_PathOrObjectsDataType_PR_objectsd;
		oiddo = (p15_oiddo_t*)object->obj;
		break;
	
	case P15_OT_SECKEY:	
#ifdef TESTOUT
		printf("\nAdding token.pkcs15Objects.secretKeys element:\n\n");
#endif
		pobjs->present = P15_PKCS15Objects_PR_secretKeys;
		skeys = &pobjs->choice.secretKeys;
		skeys->present = P15_PathOrObjectsSecretKeyType_PR_objectss;
		sk = (p15_secret_key_t*)object->obj;
		break;

	default:
		sprintf(st, "p15_save_object: Unsupported object type (%d)", 
			object->type);
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_OBJECT, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* adding */
	if (asn_sequence_add(&token->pkcs15Objects.list, pobjs)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_object:asn_sequence_add failed for pobjs");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	pobjs = NULL;

	/* calling object method */
	switch (object->type) {
	case P15_OT_PRIVKEY:
		rc = p15_mk_prkey_obj(ctx, prkeys, pk, pwkey, pwkmsk, warns, errors);
		if (rc == P15_RC_INTERNAL_ERROR) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_save_object:p15_mk_prkey_obj returned error");
			goto end;
		}
		if (rc) goto end;
		break;

	case P15_OT_TRUSTED_CERT:
	case P15_OT_CERT:
		rc = p15_mk_cert_obj(ctx, certs, cer, warns, errors);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_object:p15_mk_cert_obj returned error");
			goto end;
		}
		break;

	case P15_OT_CRL:
		rc = p15_mk_crl_obj(ctx, dobjs, crl, errors);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_object:p15_mk_crl_obj returned error");
			goto end;
		}
		break;
	
	case P15_OT_PUBKEY:
		rc = p15_mk_pubkey_obj(ctx, pubkeys, pbk, warns, errors);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_object:p15_mk_pubkey_obj returned error");
			goto end;
		}
		break;
	
	case P15_OT_OIDDO:
		rc = p15_mk_oiddo_obj(ctx, dobjs, oiddo, pwkey, pwkmsk, errors);
		if (rc == P15_RC_INTERNAL_ERROR) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_object:p15_mk_oiddo_obj returned error");
			goto end;
		}
		if (rc) goto end;
		break;
	
	case P15_OT_SECKEY:
		rc = p15_mk_seckey_obj(ctx, skeys, sk, pwkey, pwkmsk, warns, errors);
		if (rc == P15_RC_INTERNAL_ERROR) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_save_object:p15_mk_seckey_obj returned error");
			goto end;
		}
		if (rc) goto end;
		break;

	default:
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_object: Internal error 1");
		goto end;
	}
end:
	ASN_STRUCT_FREE(asn_DEF_P15_PKCS15Objects, pobjs);
	return rc;
}
/****************************************************************************/
static int p15_asn_encode(asn_TYPE_descriptor_t* td, void* el, 
	uint8_t** ebuf, size_t* ebuflen, p15_info_t** errors)
{
	int rc = 0;
/*	char errb[500]; */
/*	size_t errlen = 500; */
	p15_sbuf_t* sbuf = NULL;
	asn_enc_rval_t erv;
	p15_info_t* info;

	if (!td || !el || !ebuf || !ebuflen) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_asn_encode");
		return rc;
	}

/*	errb[0] = 0;
	if (asn_check_constraints(td, el, errb, &errlen)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, errb);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_asn_encode:asn_check_constraints returned error");
		goto end;
	}
*/
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(td, el, p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_ASN1C_ERROR;
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_asn_encode:der_encode return error");
		p15_add_error_type(info, erv);
		goto end;
	}
	rc = p15_get_solid_buf(sbuf, ebuf, ebuflen); 
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_asn_encode:p15_get_solid_buf returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
end:
	p15_free_sbuf(&sbuf);
	return rc;
}
/****************************************************************************/
int p15_save_token(p15_ctx_t* ctx, uint8_t* token_out, size_t* token_len, 
	uint8_t* password, int passlen, p15_info_t** errors)
{
	int rc = 0;
	int rc2;
	int warns = 0;
	P15_AuthenticatedData_t* token = NULL;
	P15_RecipientInfo_t* auth_recipient_info = NULL;
	uint8_t* auth_key = NULL;
	uint8_t* auth_kmsk = NULL;
	uint8_t* pwkey = NULL;
	uint8_t* pwkmsk = NULL;
	uint8_t* salt = NULL;
	P15_GostR3411Algorithm_t* ghashalg = NULL;
	OCTET_STRING_t* encaptoken = NULL;
	uint8_t tokenhash[64];
	P15_AuthAttributes_t* aattrs = NULL;
	P15_Attribute_t* attr = NULL;
	OBJECT_IDENTIFIER_t* oid = NULL;
	P15_AttributeValue_t* aval = NULL;
	OCTET_STRING_t* ostrhash = NULL;
	uint8_t* ebuf = NULL;
	size_t ebuflen;
	uint8_t hmac[64];
	uint8_t akey[32];
	void* hh;
	int i;
	uint8_t* etoken = NULL;
	size_t etokenlen;
	char* hmac_oid = NULL;
	char* hash_oid = NULL;
	int is2012;
	gostc_hmac12_ctx_t* hh12 = NULL;

	if (!ctx || !token_len || (!token_out && *token_len )) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc; 
	}	
	
	/* Rnd */
	if (!ctx->rnd_init) {
		rc = P15_RC_RNG_NOT_INITIALIZED;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

	/* Refining container format */
	if (ctx->p15_format == P15_FORMAT_ANY) ctx->p15_format = P15_FORMAT_2012;

	is2012 = ctx->p15_format != P15_FORMAT_2001;

	/* Password keys */
	rc = p15_get_pwkey(ctx, password, passlen, &pwkey, &pwkmsk, &salt);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_get_pwkey");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* Filling AuthenticatedData */
	token = p15_malloc(sizeof(P15_AuthenticatedData_t));
	if (asn_long2INTEGER(&token->version, 0)) { /* version = 0 */
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_save_token:asn_long2INTEGER:version");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("Making KEKRecipientInfo for AuthenticatedData:\n");
#endif
	rc2 = p15_mk_kekri(ctx, &auth_recipient_info, &auth_key, &auth_kmsk, 
		pwkey, pwkmsk, errors);
	if (rc2 == 1) warns = 1;
	else if (rc2) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_save_token:p15_mk_kekri failed for AuthenticatedData");
		goto end; 
	}

	/* token->recipientInfos */

	if (asn_set_add(&token->recipientInfos.list, auth_recipient_info)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:asn_set_add failed for auth_recipient_info");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	auth_recipient_info = NULL;

	/* token->macAlgorithm.algorithm */
	hmac_oid = is2012 ? P15_OID_HMAC_GOST_2012 : P15_OID_HMAC_GOST_94;

#ifdef TESTOUT
	printf("AuthenticatedData.macAlgorithm.algorithm: %s\n", hmac_oid);
#endif
	rc = p15_str_to_oid(hmac_oid, &token->macAlgorithm.algorithm);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_str_to_oid: error with macAlgorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* token->macAlgorithm.parameters -- absent */

	/* token->digestAlgorithm */

	ghashalg = p15_malloc(sizeof(P15_GostR3411Algorithm_t));
	token->digestAlgorithm = ghashalg;

	/* ghashalg->algorithm */
	hash_oid = is2012 ? P15_OID_GOST_R3411_2012_512 : P15_OID_GOST_R3411_94;
#ifdef TESTOUT
	printf("AuthenticatedData.digestAlgorithm.algorithm: %s\n", hash_oid);
#endif
	rc = p15_str_to_oid(hash_oid, &ghashalg->algorithm);
	ghashalg = NULL;
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_str_to_oid: error with digest algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* ghashalg->parameters - absent */

	/* token->encapContentInfo.eContentType */

#ifdef TESTOUT
	printf("AuthenticatedData.encapContentInfo.eContentType: %s\n\n", 
		P15_OID_CONTENT_TYPE);
#endif
	rc = p15_str_to_oid(P15_OID_CONTENT_TYPE, &token->encapContentInfo.eContentType);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_str_to_oid: error with encapContentInfo");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* token->encapContentInfo.eContent */
	rc = p15_mk_encap_token(ctx, pwkey, pwkmsk, salt, &encaptoken, 
		tokenhash, &warns, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_mk_encap_token failed");
		goto end;
	}
	if (rc) goto end;
#ifdef TESTOUT
	printf(
"\nAuthenticatedData.encapContentInfo.eContent (w/o ostr header) (%d bytes):\n",
		encaptoken->size);
	p15_dump(encaptoken->buf, encaptoken->size);
	printf("Message digest (hash) of the eContent (ostr header not included):\n");
	p15_dump(tokenhash, is2012 ? 64 : 32);
	printf("\n");
#endif
	token->encapContentInfo.eContent = encaptoken;
	encaptoken = NULL;
	
	/* token->authAttrs */
	aattrs = p15_malloc(sizeof(P15_AuthAttributes_t));

#ifdef TESTOUT
	printf("Making AuthenticatedData.authAttrs:\n");
	printf("attr1 (id-contentType):\n");
	printf("attr1.attrType: %s\n", P15_OID_AA_ID_CTYPE);
#endif
	/* attr - id-contentType */
	attr = p15_malloc(sizeof(P15_Attribute_t));
	rc = p15_str_to_oid(P15_OID_AA_ID_CTYPE, &attr->attrType);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:p15_str_to_oid: error with P15_OID_AA_ID_CTYPE");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	oid = p15_malloc(sizeof(OBJECT_IDENTIFIER_t));
#ifdef TESTOUT
	printf("attr1.attrValue1: %s\n", P15_OID_AA_CTYPE);
#endif
	rc = p15_str_to_oid(P15_OID_AA_CTYPE, oid);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:p15_str_to_oid: error with P15_OID_AA_CTYPE");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	aval = ANY_new_fromType(&asn_DEF_OBJECT_IDENTIFIER, oid);
	if (!aval) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:ANY_new_fromType failed for oid");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (asn_set_add(&attr->attrValues.list, aval)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:asn_set_add failed for aval (1)");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	aval = NULL;
	if (asn_set_add(&aattrs->list, attr)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:asn_set_add failed for aattrs (1)");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	attr = NULL;

	/* attr - id-messageDigest */
#ifdef TESTOUT
	printf("attr2 (id-messageDigest):\n");
	printf("attr2.attrType: %s\n", P15_OID_AA_ID_MDIGEST);
	printf("attr2.attrValue1: ostr len=%s - message digest (see above)\n", 
		is2012 ? "64" : "32");
	printf("\n");
#endif

	attr = p15_malloc(sizeof(P15_Attribute_t));
	rc = p15_str_to_oid(P15_OID_AA_ID_MDIGEST, &attr->attrType);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:p15_str_to_oid: error with P15_OID_AA_ID_MDIGEST");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	ostrhash = p15_malloc(sizeof(OCTET_STRING_t));
	if (OCTET_STRING_fromBuf(ostrhash, (void*)tokenhash, is2012 ? 64 : 32)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:OCTET_STRING_fromBuf failed for tokenhash");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	aval = ANY_new_fromType(&asn_DEF_OCTET_STRING, ostrhash);
	if (!aval) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:ANY_new_fromType failed for ostrhash");
		goto end;
	}
	if (asn_set_add(&attr->attrValues.list, aval)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:asn_set_add failed for aval (2)");
		goto end;
	}
	aval = NULL;
	if (asn_set_add(&aattrs->list, attr)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:asn_set_add failed for aattrs (2)");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	attr = NULL;

	/* encode to calculate hmac */
	rc = p15_asn_encode(&asn_DEF_P15_AuthAttributes, aattrs, &ebuf, &ebuflen, 
		errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_asn_encode failed for aattrs");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* calc hmac */
	for (i = 0; i < 32; i++) akey[i] = auth_key[i] ^ auth_kmsk[i];
	if (is2012) {
		hh12 = gostc_hmac12_new();
		if (!hh12) rc2 = 1;
		if (!rc2) rc2 = gostc_hmac12_init(hh12, akey, 32);
	} else hh = gostc_hmac_open(akey, 32);
#ifdef TESTOUT
	printf("Encoded authAttrs (standalone):\n");
	p15_dump(ebuf, ebuflen);
	printf("Calculating HMAC of authAttr:\nKey:\n");
	p15_dump(akey, 32);	
#endif
	gostc_cleanse(akey, 32);
	if (is2012) {
		if (!rc2) rc2 = gostc_hmac12_do(hh12, ebuf, ebuflen);
		if (!rc2) rc2 = gostc_hmac12_finish(hh12, hmac);
		if (rc2) {
			rc = P15_RC_CRYPTOLIB_HASH;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_save_token:One of gostc_hmac12_* functions returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	} else {
		gostc_hmac_do(hh, ebuf, ebuflen);
		gostc_hmac_close(hh, hmac);
	}

#ifdef TESTOUT
	printf("authAttrs HMAC (AuthenticatedData.mac):\n");
	p15_dump(hmac, is2012 ? 64 : 32);
#endif
	/* set *token->authAttrs */
	token->authAttrs = aattrs;
	aattrs = NULL;

	/* token->mac */
	rc = OCTET_STRING_fromBuf(&token->mac, (void*)hmac, is2012 ? 64 : 32);
	if (rc) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_save_token:OCTET_STRING_fromBuf failed for mac");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* ENCODE the TOKEN */
	rc = p15_asn_encode(&asn_DEF_P15_AuthenticatedData, token, 
		&etoken, &etokenlen, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_save_token:p15_asn_encode failed for etoken");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("\n\nTHE WHOLE SOFTWARE TOKEN (%d bytes):\n", (int)etokenlen);
	p15_dump(etoken, etokenlen);
	printf("\n");
#endif
	if (etokenlen > *token_len) {
		if (*token_len) {
			rc = P15_RC_BUFFER_TOO_SMALL;
			p15_add_error(errors, P15_ERROR, rc,
				"p15_save_token: token buffer is too small");
		}
		*token_len = etokenlen;
		goto end;
	}

	*token_len = etokenlen;
	memcpy(token_out, etoken, etokenlen);
end:
	gostc_hmac12_clear(hh12);
	gostc_hmac12_free(&hh12);
	gostc_cleanse(akey, 32);
	p15_free(ebuf);
	p15_free(etoken);
	ASN_STRUCT_FREE(asn_DEF_P15_AttributeValue, aval);
	ASN_STRUCT_FREE(asn_DEF_P15_AuthenticatedData, token);
	ASN_STRUCT_FREE(asn_DEF_P15_RecipientInfo, auth_recipient_info);
	ASN_STRUCT_FREE(asn_DEF_P15_GostR3411Algorithm, ghashalg);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, encaptoken);
	ASN_STRUCT_FREE(asn_DEF_P15_AuthAttributes, aattrs);
	ASN_STRUCT_FREE(asn_DEF_P15_Attribute, attr);
	ASN_STRUCT_FREE(asn_DEF_OBJECT_IDENTIFIER, oid);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, ostrhash);

	p15_free(auth_key);
	p15_free(auth_kmsk);
	p15_free(pwkey);
	p15_free(pwkmsk);
	p15_free(salt);
	
	if (rc) return rc;
	if (warns) return 1;
	return 0;
}
/****************************************************************************/
static int p15_asn_decode(uint8_t* der, size_t len, asn_TYPE_descriptor_t* td,
	void** el, p15_info_t** errors)
/* ret INTERNAL_ERROR or other errors if not internal. */
{
	int rc = 0;
	asn_dec_rval_t dval;
	p15_info_t* info = NULL;

	if (!der || !len || !td || !el) rc = P15_RC_INVALID_ARGS;
	if (!rc) {
		if (*el) rc = P15_RC_INVALID_ARGS;
	}
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_asn_decode");
		return P15_RC_INTERNAL_ERROR;
	}

	dval = ber_decode(NULL, td, el, der, len);
	if (dval.code != RC_OK) {
		if (dval.code == RC_WMORE) 
			rc = P15_RC_DER_NOT_COMPLETE;
		else
			rc = P15_RC_DER_DECODING_FAILED;
		info = p15_add_error(errors, P15_ERROR, rc, NULL);
		if (info) info->consumed = dval.consumed;
		goto end;
	}
end:
	if (rc && el) {
		td->free_struct(td, *el, 0);
		*el = NULL;
	}
	return rc;
}
/****************************************************************************/
static int p15_ck_ANY_NULL(ANY_t* any)
/* 0-ANY=NULL, 1-not NULL, >1 - err */
{
	if (!any) return P15_RC_INVALID_ARGS;
	if (any->size != 2) return 1;
	if (!any->buf) return 1;
	if (any->buf[0] != 0x05 || any->buf[1]) return 1;
	return 0;
}
/****************************************************************************/
static int p15_load_kmi(p15_ctx_t* ctx, P15_KeyManagementInfo_t* kmis, 
	uint8_t* password, int passlen, p15_id_t** kekid,
	uint8_t** pwkey, uint8_t** pwkmsk, p15_info_t** errors)
/* ret 0, 1-warns, P15_RC_INTERNAL_ERROR, errors */
{
/* In this function P15-2001/2012 format detection is performed
 * and the concrete format is set.
 */
	int rc = 0;
	int c = 0;
	int fnd = 0;
	int ign = 0;
	int warns = 0;
	int errs = 0;
	P15_KeyManagementInfoElem_t* kmi = NULL;
	P15_PasswordInfo_t* pwi = NULL;
	OBJECT_IDENTIFIER_t* aoid = NULL;
	ANY_t* apar = NULL;
	P15_PBKDF2_params_t* apar2 = NULL;
	long n = 0;
	uint8_t* salt = NULL;
	char st[1024];
	int is2012;

	if (!ctx || !kekid || !pwkey || !pwkmsk) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_load_kmi");
		return P15_RC_INTERNAL_ERROR;
	}

	if (!kmis) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc, 
		"p15_load_kmi: PKCS15Token.keyManagementInfo is absent");
		goto end;
	}

	for (c = 0; c < kmis->list.count; c++) {
		if (fnd) { ign++; continue; } /* count rest of ignored objs */

		kmi = kmis->list.array[c];
		if (kmi->keyInfo.present != keyInfo_PR_passwordInfo) {
			ign++; continue; 
		}
		pwi = &kmi->keyInfo.choice.passwordInfo;
		aoid = &pwi->algId.algorithm;
		rc = p15_oid_str_eq(P15_OID_PBKDF2, aoid);
		if (rc > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_kmi:p15_oid_str_eq returned error for pwi->algorithm");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc) {
			ign++;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_load_kmi: Skipped PasswordInfo object with unknown algorithm OID in PKCS15Token.KeyManagementInfo");
			rc = 0;
			continue;
		}
		apar = pwi->algId.parameters;
		if (!apar) {
			errs = 1;
			ign++;
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo: PasswordInfo object does not contain PBKDF2 parameters. Skipping the object");
			continue;
		}
		
		/* decoding PBKDF2 */
		rc = ANY_to_type(apar, &asn_DEF_P15_PBKDF2_params, (void**)&apar2);
		if (rc || !apar2) {
			rc = 0;
			ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
			apar2 = NULL;
			errs = 1;
			ign++;
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo: Unable to decode PBKDF2 parameters in PasswordInfo object. Skipping the object");
			continue;
		}
		
		/* *apar2->prf */
		if (!apar2->prf) {
			ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
			apar2 = NULL;
			ign++;
			p15_add_error(errors, P15_WARNING, P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo.algId.parameters: prf id is not present. Skipping this PasswordInfo object.");
			continue;
		}
/*******************************************************/
/* Here is the FIRST detection of P15-2001/2012 format */
/*******************************************************/
		rc = p15_oid_str_eq(P15_OID_HMAC_GOST_94, &apar2->prf->algorithm);
		if (!rc) is2012 = 0;
		else if (rc == 1) {
			rc = p15_oid_str_eq(P15_OID_HMAC_GOST_2012, &apar2->prf->algorithm);
			if (!rc) is2012 = 1;
		}
		if (rc > 1) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_load_kmi:p15_oid_str_eq returned error for apar2->prf->algorithm");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc) {
			rc = 0;
			ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
			apar2 = NULL;
			ign++;
			p15_add_error(errors, P15_WARNING, P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo.algId.parameters.prf: unrecognized algorithm OID. Skipping this PasswordInfo object.");
			continue;
		}

		/* Checking p15_format constraint, setting ANY to concrete format */
		rc = p15_ck_format(ctx, is2012, errors);

		if (rc == P15_RC_INVALID_CONTAINER_FORMAT) {
			errs = 1;
			p15_add_error(errors, P15_ERROR, rc, 
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo.algId.parameters.prf: Unexpected HMAC algorithm for this container format.");
			rc = 0;
		} else if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_load_kmi: p15_ck_format returned error.");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		
		if (apar2->prf->parameters) {
			rc = p15_ck_ANY_NULL(apar2->prf->parameters);
			if (!rc) {
				ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
				apar2 = NULL;
				ign++;
				p15_add_error(errors, P15_ERROR, 
					P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo.algId.parameters.prf.parameters: Parameters are present and not NULL. Skipping this PasswordInfo object.");
				continue;
			}
			rc = 0;
		}

		/* apar2->iterationCount */

		if (asn_INTEGER2long(&apar2->iterationCount, &n)) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_load_kmi:asn_INTEGER2long returned error for apar2->iterationCount");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		ctx->p5_c = n;

		/* *apar2->keyLength */

		if (!apar2->keyLength)  rc = 1; 
		else rc = asn_INTEGER2long(apar2->keyLength, &n);
		if (!rc && n != 32) rc = 1;
		if (rc) {
			rc = 0;
			ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
			apar2 = NULL;
			errs = 1;
			ign++;
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo.algId.parameters: keyLength should exist for GOST key derivation algorithm and should be 32. Skipping the PasswordInfo object");
			continue;
		}

		/* apar2->salt */

		if (apar2->salt.present != salt_PR_specified) {
			ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
			apar2 = NULL;
			warns = 1;
			ign++;
			p15_add_error(errors, P15_WARNING, P15_RC_INVALID_PBKDF2_PARAMS,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo.algId.parameters: salt is not directly specified. This is not supported. Skipping the PasswordInfo object");
			continue;
		}
		ctx->p5_salt_len = apar2->salt.choice.specified.size;
		
		fnd = 1;

		salt = p15_malloc(ctx->p5_salt_len);
		memcpy(salt, apar2->salt.choice.specified.buf, ctx->p5_salt_len);
		
		/* *pwi->hint */
		ctx->password_hint[0] = 0;
		if (pwi->hint) {
			n = pwi->hint->size;
			if (n > P15_LABEL_MAX) {
				n = P15_LABEL_MAX;
				warns = 1;
				p15_add_error(errors, P15_WARNING, P15_RC_INVALID_TOKEN,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo.keyInfo.passwordInfo: Password hint exceeds the limit.");
			}
			memcpy(ctx->password_hint, pwi->hint->buf, n);
			ctx->password_hint[n] = 0;
		}

		/* kmi->keyId */
		*kekid = p15_new_ext_id(kmi->keyId.buf, kmi->keyId.size);
		if (!*kekid) {
			rc = P15_RC_INVALID_TOKEN;
			p15_add_error(errors, P15_ERROR, rc,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo: Invalid (too long) keyId");
			goto end;
		}
	}
	
	if (ign) {
		warns = 1;
		sprintf(st, "p15_load_kmi: In PKCS15Token.KeyManagementInfo: %d object(s) were ignored. Note that multiple PasswordInfo objects are not supported", ign);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}
	
	if (!fnd) {
		rc = P15_RC_NO_PWINFO;
		p15_add_error(errors, P15_ERROR, rc,
"p15_load_kmi: In PKCS15Token.KeyManagementInfo: Unable to find appropriate PasswordInfo object");
		goto end;
	}

	/* create pwkey */
	rc = p15_get_pwkey(ctx, password, passlen, pwkey, pwkmsk, &salt);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_kmi:p15_get_pwkey returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
	if (errs) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
"p15_load_kmi: PKSC15Token.KeyManagementInfo contains some erroneous structures");
		goto end;
	}
	
	if (warns) rc = 1;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_PBKDF2_params, apar2);
	p15_free(salt);
	return rc;
}
/****************************************************************************/
static int p15_load_kekris(p15_ctx_t* ctx, const char* upname,
	P15_RecipientInfos_t* ris, p15_id_t* kekid, 
	uint8_t* pwkey, uint8_t* pwkmsk, 
	uint8_t** akey, uint8_t** akmsk,
	int set_enc_uz,
	p15_info_t** errors)
/* ret 0, 1-warns, P15_RC_WRONG_PASSWORD, P15_RC_INTERNAL_ERROR, P15_RC_NO_AUTH_AND_DECRYPT, */
/* other not important errors */
{
	int warns = 0;
	int terr = 0;
	int rc = 0;
	int i;
	int fnd = 0;
	P15_RecipientInfo_t* ri = NULL;
	P15_KEKRecipientInfo_t* kekri = NULL;
	p15_id_t* kid = NULL;
	char st[1024];
	long v;
	P15_GostKeyWrapAlgorithmIdentifier_t* kalg = NULL;
	P15_Gost28147_89_KeyWrapParameters_t* kwpar = NULL;
	P15_Gost28147_89_EncryptedKey_t* ekeyt = NULL;
	int uz = 0;
	uint8_t ukm[8];
	uint8_t key[32];
	uint8_t pkey[32];
	uint8_t *ak = NULL;
	uint8_t *akm = NULL;

	if (!ctx || !upname || !ris || !kekid || !pwkey || !pwkmsk || !akey 
		|| !akmsk) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_kekris");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}
	for (i = 0; i < ris->list.count; i++) {
		ri = ris->list.array[i];
		if (ri->present != P15_RecipientInfo_PR_kekri) continue;
		kekri = &ri->choice.kekri;

		/* kekri->kekid.keyIdentifier */
		kid = p15_new_ext_id(kekri->kekid.keyIdentifier.buf,
			kekri->kekid.keyIdentifier.size);
		if (!kid) {
			rc = P15_RC_INVALID_TOKEN;
			sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo: Invalid (too long) keyIdentifier", upname);
			p15_add_error(errors, P15_ERROR, rc, st);
			rc = P15_RC_NO_AUTH_AND_DECRYPT;
			goto end;
		}

		/* Our kekid? */
		if (p15_id_cmp(kid, kekid)) continue;
		fnd = 1;
		break;
	}
	if (!fnd) {
		sprintf(st, 
"p15_load_kekris:In %s.recipientInfos: No appropriate keyIdentifier found corresponding to passwordInfo structure", upname);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* *kekri->kekid.date and *.other */
	if (kekri->kekid.date || kekri->kekid.other) {
		warns = 1;
		sprintf(st,
"p15_load_kekris:In %s.KEKRecipientInfo: keyId.date or keyId.other fields are ignored",
			upname);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* kekri->version -  4 */
	rc = asn_INTEGER2long(&kekri->version, &v);
	if (!rc && v != 4) rc = 1;
	if (rc) {
		rc = 0;
		terr = 1;
		sprintf(st,
		"p15_load_kekris:In %s.KEKRecipientInfo: version is not 4", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
	}
		
	/* kekri->keyEncryptionAlgorithm */
	kalg = &kekri->keyEncryptionAlgorithm;

	/* kalg->algorithm */
	rc = p15_oid_str_eq(P15_OID_GOST_KEY_WRAP, &kalg->algorithm);
	if (rc > 1) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo:p15_oid_str_eq returned error for kalg->algortithm", 
			upname);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm: Unknown OID of algorithm",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* kalg->parameters */
	if (!kalg->parameters) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm: No key-wrap algorithm parameters specified", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	rc = ANY_to_type(kalg->parameters, &asn_DEF_P15_Gost28147_89_KeyWrapParameters, 
		(void**)&kwpar);
	if (rc || !kwpar) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm: Failed to decode key-wrap algorithm parameters", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* kwpar->encryptionParamSet */
	rc = p15_oid_to_encuz(&kwpar->encryptionParamSet, &uz);
	if (rc == P15_RC_INTERNAL_ERROR) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm:p15_oid_to_encuz", upname);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	if (rc) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm.parameters: Unrecognized ParamSet OID", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	
	if (set_enc_uz) {
		if (ctx->p15_format == P15_FORMAT_2001 &&
			uz == P15_89UZ_CPRO_Z) {

			terr = 1;
			sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm.parameters: GOST 28147-89 S-Box Z detected in P15-2001 container format. Leaving ctx->enc_uz unchanged.", upname);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_CONTAINER_FORMAT, st);

		} else ctx->enc_uz = uz;
	}

	/* *kwpar->ukm */
	if (!kwpar->ukm) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm.parameters: UKM is absent",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	if (kwpar->ukm->size != 8) {
sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.keyEncryptionAlgorithm.parameters: UKM must be 8 bytes",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	memcpy(ukm, kwpar->ukm->buf, 8);

	/* kekri->encryptedKey */
	rc = p15_asn_decode(kekri->encryptedKey.buf, kekri->encryptedKey.size,
		&asn_DEF_P15_Gost28147_89_EncryptedKey, (void**)&ekeyt, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.encryptedKey: p15_asn_decode returned error", upname);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	if (rc) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.encryptedKey: Failed to decode encrypteKey structure",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	if (ekeyt->encryptedKey.size != 32 || ekeyt->macKey.size != 4) {
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo.encryptedKey: macKey or encryptedKey have incorrect size", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	for (i = 0; i < 32; i++) pkey[i] = pwkey[i] ^ pwkmsk[i];
	rc = gostc_sym_keyunwrap(ekeyt->encryptedKey.buf, pkey, ekeyt->macKey.buf, 
		ukm, uz, key);
	gostc_cleanse(pkey, 32);
	if (rc) gostc_cleanse(key, 32);
	if (rc > 1) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo: p15_keyunwrap returned error", upname);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	if (rc) {
		rc = P15_RC_WRONG_PASSWORD;
sprintf(st, 
"p15_load_kekris:In %s.KEKRecipientInfo: Authentication failed. Possibly wrong password",
			upname);
		p15_add_error(errors, P15_ERROR, rc, st);
		goto end;
	}
	
	*akey = ak = p15_malloc(P15_PAGE);
	*akmsk = akm = p15_malloc(P15_PAGE);
	rc = p15_get_safe_random(ctx, akm, 32);
	if (rc) {
		sprintf(st,
"p15_load_kekris:In %s.KEKRecipientInfo: p15_get_safe_random returned error", upname);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	for (i = 0; i < 32; i++) ak[i] = key[i] ^ akm[i];	

	if (terr) rc = P15_RC_INVALID_TOKEN;
	if (!rc && warns) rc = 1;
end:
	gostc_cleanse(key, 32);
	gostc_cleanse(pkey, 32);

	ASN_STRUCT_FREE(asn_DEF_P15_Gost28147_89_EncryptedKey, ekeyt);
	ASN_STRUCT_FREE(asn_DEF_P15_Gost28147_89_KeyWrapParameters, kwpar);
	p15_free_id(&kid);
	return rc;
}
/****************************************************************************/
static int p15_load_aattr(int hash2012, P15_Attribute_t* attr, int* oidfound, 
	int* hashfound, uint8_t* hash, int* ign, int* terr, p15_info_t** errors)
/* ret !0-internal error */
{
	int rc = 0;
	ANY_t* any = NULL;
	OBJECT_IDENTIFIER_t* oid = NULL;
	OCTET_STRING_t* mdig = NULL;
	char st[1024];

	rc = p15_oid_str_eq(P15_OID_AA_ID_CTYPE, &attr->attrType);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_aattr:p15_oid_str_eq returned error for attrType");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (!rc) { /* content type */
		if (attr->attrValues.list.count != 1) {
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattr:In AuthenticatedData.authAttrs.Attribute, attrType=id-contentType: attrValues must have 1 element. Skipping the Attribute");
			*terr = 1;
			*ign = *ign + 1;
			goto end;
		}
		any = attr->attrValues.list.array[0];
		rc = ANY_to_type(any, &asn_DEF_OBJECT_IDENTIFIER, (void**)&oid);
		if (rc || !oid) {
			rc = 0;
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattr:In AuthenticatedData.authAttrs.Attribute, attrType=id-contentType: Failed to decode attrValue as OID. Skipping the Attribute");
			*terr = 1;
			*ign = *ign + 1;
			goto end;
		}
		rc = p15_oid_str_eq(P15_OID_AA_CTYPE, oid);
		if (rc > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_aattr:p15_oid_str_eq returned error for attrValue");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc) {
			rc = 0;
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattr:In AuthenticatedData.authAttrs.Attribute, attrType=id-contentType: Unrecognized OID of content type. Skipping the Attribute");
			*terr = 1;
			*ign = *ign + 1;
			goto end;
		}
		*oidfound = 1;
		rc = 0;
		goto end;
	}
	
	rc = p15_oid_str_eq(P15_OID_AA_ID_MDIGEST, &attr->attrType);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_aattr:p15_oid_str_eq returned error for attrType (2)");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) { /* unknown attr */
		*ign = *ign + 1;
		rc = 0;
		goto end;
	}

	/* Message digest */

	if (attr->attrValues.list.count != 1) {
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattr:In AuthenticatedData.authAttrs.Attribute, attrType=id-messageDigest: attrValues must have 1 element. Skipping the Attribute");
		*terr = 1;
		*ign = *ign + 1;
		goto end;
	}
	any = attr->attrValues.list.array[0];
	rc = ANY_to_type(any, &asn_DEF_OCTET_STRING, (void**)&mdig);
	if (rc || !mdig) {
		rc = 0;
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattr:In AuthenticatedData.authAttrs.Attribute, attrType=id-messageDigest: Failed to decode attrValue as OCTET STRING. Skipping the Attribute");
		*terr = 1;
		*ign = *ign + 1;
		goto end;
	}
	if (mdig->size != (hash2012 ? 64 : 32)) {
		sprintf(st, "p15_load_aattr:In AuthenticatedData.authAttrs.Attribute, attrType=id-messageDigest: attrValue (OCTET_STRING) should be %s bytes. Skipping the Attribute", 
			hash2012 ? "64" : "32");
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		*terr = 1;
		*ign = *ign + 1;
		goto end;
	}
	memcpy(hash, mdig->buf, hash2012 ? 64 : 32);
	*hashfound = 1;
end:
	ASN_STRUCT_FREE(asn_DEF_OBJECT_IDENTIFIER, oid);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, mdig);
	return rc;
}
/****************************************************************************/
static int p15_load_aattrs(int hash2012, P15_AuthAttributes_t* aattrs, 
	uint8_t* hash, p15_info_t** errors)
/* ret 0, 1-warns, int err, no auth, other errs (terr) */
{
	int rc = 0;
	int warns = 0;
	int oidfound = 0;
	int hashfound = 0;
	int i;
	int ign = 0;
	int terr = 0;
	char st[1024];

	if (!aattrs || !hash) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_load_aattrs");
		return P15_RC_INTERNAL_ERROR;
	}
	for (i = 0; i < aattrs->list.count; i++) {
		rc = p15_load_aattr(hash2012, aattrs->list.array[i], &oidfound, 
			&hashfound, hash, &ign, &terr, errors);
		if (rc) return P15_RC_INTERNAL_ERROR;
	}
	
	if (!oidfound) {
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattrs: In AuthenticatedData.authAttrs: Failed to find content-type attribute");
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		return rc;
	}

	if (!hashfound) {
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_aattrs: In AuthenticatedData.authAttrs: Failed to find messageDigest attribute");
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		return rc;
	}

	if (terr) rc = P15_RC_INVALID_TOKEN;
	if (ign) {
		sprintf(st, 
"p15_load_aattrs: In AuthenticatedData.authAttrs: Ignored %d unrecognized attribute(s)",
			ign);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		warns = 1;
	}
	if (!rc && warns) rc = 1;

	return rc;
}
/****************************************************************************/
static p15_obj_it_t* p15_new_obj_it(int type)
{
	p15_obj_it_t* it = NULL;
	
	it = p15_malloc(sizeof(p15_obj_it_t));
	it->type = type;
	return it;	
}
/****************************************************************************/
static void p15_free_obj_it(p15_obj_it_t** pit)
{
	p15_obj_it_t* it;
	if (!pit) return;
	it = *pit;
	if (!it) return;
	p15_free(it);
	*pit = NULL;
}
/****************************************************************************/
static int p15_oit_ck_type(int type, P15_PKCS15Objects_t* os)
/* ret: 0-good, 1-bad, 2-unknown */
{
	if (!os) return 1;

	switch (type) {
	case P15_OIT_ANY:
	case P15_OIT_UNKNOWN:
		switch (os->present) {
		case P15_PKCS15Objects_PR_privateKeys:
		case P15_PKCS15Objects_PR_certificates:
		case P15_PKCS15Objects_PR_trustedCertificates:
		case P15_PKCS15Objects_PR_dataObjects:
		case P15_PKCS15Objects_PR_publicKeys: 
 		case P15_PKCS15Objects_PR_secretKeys: 
			if (type == P15_OIT_ANY) return 0;
			return 1;
/*		case PKCS15Objects_PR_trustedPublicKeys: 
 *		case PKCS15Objects_PR_usefulCertificates: 
 *		case PKCS15Objects_PR_authObjects: 
 */
		default: return 2;
		}

	case P15_OIT_CERT:
		if (os->present == P15_PKCS15Objects_PR_certificates) return 0;
		return 1;

	case P15_OIT_RCERT: 
		if (os->present == P15_PKCS15Objects_PR_trustedCertificates) return 0;
		return 1;

	case P15_OIT_PRKEY:
		if (os->present == P15_PKCS15Objects_PR_privateKeys) return 0;
		return 1;
	
	case P15_OIT_DATA:
		if (os->present == P15_PKCS15Objects_PR_dataObjects) return 0;
		return 1;
	
	case P15_OIT_PUBKEY:
		if (os->present == P15_PKCS15Objects_PR_publicKeys) return 0;
		return 1;
	
	case P15_OIT_SECKEY:
		if (os->present == P15_PKCS15Objects_PR_secretKeys) return 0;
		return 1;

	default:
		return 1;
	}

	return 1;
}
/****************************************************************************/
static void* p15_oit_get_obj(P15_PKCS15Objects_t* os, int oi, int* type,
	int* warns, p15_info_t** errors)
{
	P15_PrivateKeys_t* prkeys = NULL;
	P15_Certificates_t* certs = NULL;
	P15_DataObjects_t* datas = NULL;
	P15_PublicKeys_t* pubkeys = NULL;
	P15_SecretKeys_t* skeys = NULL;

	if (!os || !warns || !type) return NULL;
	switch (os->present) {
	case P15_PKCS15Objects_PR_privateKeys:
		*type = P15_OIT_PRKEY;
		prkeys = &os->choice.privateKeys;
		if (prkeys->present != P15_PathOrObjectsPrivateKeyType_PR_objectsk) {
			*warns = 1;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_oit_get_obj: In PKCS15Token.pkcs15Objects.privateKeys: Unsupported choice detected. Only direct 'objects' choice is supported. Skipping the structure");
			return NULL;
		}
		if (oi >= prkeys->choice.objectsk.list.count) return NULL;
		return prkeys->choice.objectsk.list.array[oi];

	case P15_PKCS15Objects_PR_certificates:
		*type = P15_OIT_CERT;
		certs = &os->choice.certificates;
		if (certs->present != P15_PathOrObjectsCertificateType_PR_objectsc) {
			*warns = 1;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_oit_get_obj: In PKCS15Token.pkcs15Objects.certificates: Unsupported choice detected. Only direct 'objects' choice is supported. Skipping the structure");
			return NULL;
		}
		if (oi >= certs->choice.objectsc.list.count) return NULL;
		return certs->choice.objectsc.list.array[oi];

	case P15_PKCS15Objects_PR_trustedCertificates:
		*type = P15_OIT_RCERT;
		certs = &os->choice.trustedCertificates;
		if (certs->present != P15_PathOrObjectsCertificateType_PR_objectsc) {
			*warns = 1;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_oit_get_obj: In PKCS15Token.pkcs15Objects.trustedCertificates: Unsupported choice detected. Only direct 'objects' choice is supported. Skipping the structure");
			return NULL;
		}
		if (oi >= certs->choice.objectsc.list.count) return NULL;
		return certs->choice.objectsc.list.array[oi];
	
	case P15_PKCS15Objects_PR_dataObjects:
		*type = P15_OIT_DATA;
		datas = &os->choice.dataObjects;
		if (datas->present != P15_PathOrObjectsDataType_PR_objectsd) {
			*warns = 1;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_oit_get_obj: In PKCS15Token.pkcs15Objects.dataObjects: Unsupported choice detected. Only direct 'objects' choice is supported. Skipping the structure");
			return NULL;
		}
		if (oi >= datas->choice.objectsd.list.count) return NULL;
		return datas->choice.objectsd.list.array[oi];
	
	case P15_PKCS15Objects_PR_publicKeys:
		*type = P15_OIT_PUBKEY;
		pubkeys = &os->choice.publicKeys;
		if (pubkeys->present != P15_PathOrObjectsPublicKeyType_PR_objectspk) {
			*warns = 1;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_oit_get_obj: In PKCS15Token.pkcs15Objects.publicKeys: Unsupported choice detected. Only direct 'objects' choice is supported. Skipping the structure");
			return NULL;
		}
		if (oi >= pubkeys->choice.objectspk.list.count) return NULL;
		return pubkeys->choice.objectspk.list.array[oi];

	case P15_PKCS15Objects_PR_secretKeys:
		*type = P15_OIT_SECKEY;
		skeys = &os->choice.secretKeys;
		if (skeys->present != P15_PathOrObjectsSecretKeyType_PR_objectss) {
			*warns = 1;
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_oit_get_obj: In PKCS15Token.pkcs15Objects.secretKeys: Unsupported choice detected. Only direct 'objects' choice is supported. Skipping the structure");
			return NULL;
		}
		if (oi >= skeys->choice.objectss.list.count) return NULL;
		return skeys->choice.objectss.list.array[oi];

	default:
		*type = P15_OIT_UNKNOWN;
		return NULL;
	}

	return NULL;
}
/****************************************************************************/
static int p15_obj_iterate(p15_obj_it_t* it, P15_PKCS15Token_t* token, void** obj, 
	int* type, int* warns, p15_info_t** errors)
/* ret: 0-ok (object returned), 1-no objects left, >1-internal error 
 * when it->type==P15_OIT_ANY --> actual type is returned to *type 
 */
{
	P15_PKCS15Objects_t* os = NULL;
	int rc = 0;

	if (!it || !obj || !token || !warns || !type) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
		"p15_obj_iterate");
		return 2;
	}
	if (it->st > 1) return 1;
	if (!it->st) {
		it->st = 1;
		it->objsi = 0;
		it->obji = 0;
	} else
		it->obji++;

	while (1) {
		if (it->objsi >= token->pkcs15Objects.list.count) {
			it->st = 2;
			return 1;
		}
		os = token->pkcs15Objects.list.array[it->objsi];
		rc = p15_oit_ck_type(it->type, os);
		if (rc > 1) { /* unknown undecoded group */
			it->objsi++;
			it->obji = 0;
			*obj = os;
			return 0;
		}

		if (rc) { /* other type - skipping */
			it->obji = 0;
			it->objsi++;
			continue;
		}

		/* type matched */
		*obj = p15_oit_get_obj(os, it->obji, type, warns, errors);
		if (*obj) return 0; /* object exists */
		
		/* object not exist - new group */
		it->obji = 0;
		it->objsi++;
	}
	return 2;
}
/****************************************************************************/
static int p15_iterate_data_object(void* obj,
	P15_Label_t** label, uint8_t* flags, 
	OBJECT_IDENTIFIER_t** appoid, OBJECT_IDENTIFIER_t** dataoid,
	ANY_t** oany, P15_EnvelopedData_t** edata, int* warns, int* errs,
	p15_info_t** errors)
/* ret: 0-found, 1-continue, int err */
/* Returned data should NOT be freed */
{
	int rc = 0;
	int rc2;
	char st[4096];
	P15_DataType_t* dt;
	P15_PKCS15ObjectOidDO_t* oiddo = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_OidDO_t* ta = NULL;
	char labelst[P15_LABEL_MAX + 1];

	if (!label || !flags || !appoid || !dataoid || !oany || !edata || !warns || !obj) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_iterate_data_object");
		return P15_RC_INTERNAL_ERROR;
	}

	*label = NULL;
	*flags = 0;
	*appoid = NULL;
	*dataoid = NULL;
	*oany = NULL;
	*edata = NULL;
	bzero(labelst, sizeof(labelst));

	dt = obj;

	if (dt->present != P15_DataType_PR_oidDO) {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_iterate_data_object: In PKCS15Token.pkcs15Objects.dataObjects.objects: Only oidDO objects supported. Skipping the data object");
		rc = 1;
		goto end;
	}
	oiddo = &dt->choice.oidDO;

	/* oiddo->commonObjectAttributes->label */
	coa = &oiddo->commonObjectAttributes;
	if (coa->label) {
		if (!coa->label->size) goto no_label;
		if (coa->label->size > P15_LABEL_MAX) {
			*errs = 1;
			p15_add_error(errors, P15_ERROR, P15_RC_LABEL_TOO_LONG,
"p15_iterate_data_object: In DataType.commonObjectAttributes.label: Label is too long. Skipping the label");
		} else {
			*label = coa->label;
			memcpy(labelst, coa->label->buf, coa->label->size);
		}
	}

no_label:
	/* oiddo->classAttributes */
	if (oiddo->classAttributes.applicationName) {
		*warns = 1;
		sprintf(st, 
"p15_iterate_data_object: For data object '%s' in DataType.oidDO.classAttributes: applicationName is not supported. Ignoring the field", labelst);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}
	if (!oiddo->classAttributes.applicationOID) {
		*warns = 1;
		sprintf(st,
"p15_iterate_data_object: For data object '%s' in DataType.oidDO.classAttributes: Only applicationOID option is supported. Skipping the object.", labelst);
		rc = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	*appoid = oiddo->classAttributes.applicationOID;

	/* oiddo->commonObjectAttributes (continued) */
	if (coa->flags) {
		rc2 = p15_unpack_bitstring(coa->flags, 1, 6);
		if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
			sprintf(st,
"p15_iterate_data_object: For data object '%s':p15_unpack_bitstring returned error for coa->flags", labelst);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc2) {
			*warns = 1;
			sprintf(st, 
"p15_iterate_data_object: For data object '%s' in DataType.commonObjectAttributes.flags: Unrecoginzed flag format. Skipping flags", labelst);
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		} else
			*flags = coa->flags->buf[0];
	}
	if (coa->authId) {
		*warns = 1;
		sprintf(st, 
"p15_iterate_data_object: For data object '%s' in DataType.oidDO.commonObjectAttributes: authId field is ignored", labelst);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}
	if (coa->userConsent) {
		*warns = 1;
		sprintf(st, 
"p15_iterate_data_object: For data object '%s' in DataType.oidDO.commonObjectAttributes: userConsent field is ignored", labelst);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}
	if (coa->accessControlRules) {
		*warns = 1;
		sprintf(st, 
"p15_iterate_data_object: For data object '%s' in DataType.oidDO.commonObjectAttributes:accessControlRules field is ignored", labelst);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	} 

	/* oiddo->typeAttributes */
	ta = &oiddo->typeAttributes;

	/* ta->id */
	*dataoid = &oiddo->typeAttributes.id;
/* - may be not equal
	if (p15_oid_eq(&oiddo->typeAttributes.id, *id)) {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_iterate_data_object: In DataType.oidDO.typeAttributes: id OID is not equal to applicationOID. This is not supported. Skipping the object");
		rc = 1;
		goto end;
	}
*/
	/* ta->value */
	if (ta->value.present == P15_ObjectValueAny_PR_direct) {
		*oany = &ta->value.choice.direct;
	} else if (ta->value.present == P15_ObjectValueAny_PR_direct_protected) {
		*edata = &ta->value.choice.direct_protected;
	} else {
		*warns = 1;
		sprintf(st, 
"p15_iterate_data_object: For data object '%s' in DataType.oidDO.typeAttributes.value: Only 'direct' or 'direct-protected' choices supported. Skipping the object", labelst);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		rc = 1;
		goto end;
	}
end:
	return rc;
}
/****************************************************************************/
static int p15_decrypt_enveloped_data(p15_ctx_t* ctx, p15_id_t* kekid, 
	const char* upname, P15_EnvelopedData_t* edata, 
	uint8_t** data,	int* len, uint8_t* pwkey, uint8_t* pwkmsk, 
	int* warns, int* errs, 	p15_info_t** errors)
/* ret: 0-ok, int err, no auth */
{
	int rc = 0;
	long v;
	char st[1024];
	uint8_t* ekey = NULL;
	uint8_t* ekmsk = NULL;
	P15_EncryptedContentInfo_t* eci = NULL;
	P15_ContentEncryptionAlgorithmIdentifier_t* ealgid = NULL;
	P15_Gost28147_89_Parameters_t* apar = NULL;
	uint8_t iv[8];
	int uz;
	uint8_t* data2 = NULL;
	int len2;
	void* mh = NULL;
	uint8_t imit[4];

	if (!edata || !data || !len || !pwkey || !pwkmsk || !warns || !errs) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_decrypt_enveloped_data");
		return P15_RC_INTERNAL_ERROR;
	}

	/* edata->version - should be 2 */
	rc = asn_INTEGER2long(&edata->version, &v);
	if (v != 2) rc = 1;
	if (rc) {
		rc = 0; 
		*errs = 1;
		sprintf(st, "p15_decrypt_enveloped_data: In %s:version should be 2",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
	}

	/* *edata->originatorInfo */
	if (edata->originatorInfo) {
		*warns = 1;
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s:originatorInfo is not supported and ignored",
			upname);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *edata->unprotectedAttrs */
	if (edata->unprotectedAttrs) {
		*warns = 1;
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s:unprotectedAttrs is not supported and ignored",
			upname);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* edata->recipientInfos */
	rc = p15_load_kekris(ctx, upname, &edata->recipientInfos, kekid, 
		pwkey, pwkmsk, &ekey, &ekmsk, 0, errors);
/* ret 0, 1-warns, P15_RC_WRONG_PASSWORD, P15_RC_INTERNAL_ERROR, P15_RC_NO_AUTH_AND_DECRYPT, */
/* other not important errors */
	switch (rc) {
	case 0: break;
	case 1:	rc = 0; *warns = 1; break;
	case P15_RC_INTERNAL_ERROR:
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_decrypt_enveloped_data:p15_load_kekris returned error");
		goto end;
	case P15_RC_WRONG_PASSWORD:
	case P15_RC_NO_AUTH_AND_DECRYPT:
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		sprintf(st, "p15_decrypt_enveloped_data: Failed to decrypt %s", upname);
		p15_add_error(errors, P15_ERROR, rc, st);
		goto end;
	default:
		*errs = 1;
	}

	/* edata->encryptedContentInfo */
	eci = &edata->encryptedContentInfo;

	/* eci->contentType */
	rc = p15_oid_str_eq(P15_OID_ENC_CTYPE, &eci->contentType);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_decrypt_enveloped_data:p15_oid_str_eq returned error for eci->contenType");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentType: Unrecognized content OID. Failed to decrypt the enveloped data", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* eci->contentEncryptionAlgorithm */
	ealgid = &eci->contentEncryptionAlgorithm;

	/* ealgid->algorithm */
	rc = p15_oid_str_eq(P15_OID_GOST_ENCR_IMIT, &ealgid->algorithm);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_decrypt_enveloped_data:p15_oid_str_eq returned error for ealgid->algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentEncryptionAlgorithm.algorithm: Unsupported encryption mechanism OID. Failed to decrypt the enveloped data", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* *ealgid->parameters */
	if (!ealgid->parameters) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentEncryptionAlgorithm.parameters: Algorithm parameters are absent, but should be present. Failed to decrypt the enveloped data",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	rc = ANY_to_type(ealgid->parameters, &asn_DEF_P15_Gost28147_89_Parameters,
		(void**)&apar);
	if (rc || !apar) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentEncryptionAlgorithm.parameters: Failed to decode GOST encryption algorithm parameters. Failed to decrypt the enveloped data",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* apar->iv */
	if (apar->iv.size != 8) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentEncryptionAlgorithm.parameters: IV should be 8 bytes long. Failed to decrypt the enveloped data",
			upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	memcpy(iv, apar->iv.buf, 8);

	/* apar->encryptionParamSet */
	rc = p15_oid_to_encuz(&apar->encryptionParamSet, &uz);
	if (rc == P15_RC_UNSUPPORTED_ALGORITHM) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentEncryptionAlgorithm.parameters: Unrecognized encryption param set. Failed to decrypt the enveloped data",
			upname);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	if (rc) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.contentEncryptionAlgorithm.parameters:p15_oid_to_encuz returned error", upname);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* *eci->encryptedContent */
	if (!eci->encryptedContent) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo: encryptedContent is absent. Failed to decrypt the enveloped data", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}

	/* decrypt */
	if (eci->encryptedContent->size < 5 || !eci->encryptedContent->buf) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.encryptedContent: Invalid encrypted data. Failed to decrypt the enveloped data", upname);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		goto end;
	}
	len2 = eci->encryptedContent->size - 4;
	data2 = p15_malloc(len2);
	gostc_sym_DecryptCFB(eci->encryptedContent->buf, data2, iv, len2, 
		(void*)ekey, (void*)ekmsk, &gostc_sym_uz[uz][0]);

	/* ck imit */
	mh = gostc_sym_MacOpen();
	gostc_sym_DoMac(mh, data2, len2, (void*)ekey, (void*)ekmsk, &gostc_sym_uz[uz][0]); 
	gostc_sym_MacClose(mh, imit, (void*)ekey, (void*)ekmsk, &gostc_sym_uz[uz][0]);
	if (memcmp(eci->encryptedContent->buf + len2, imit, 4)) {
		sprintf(st, 
"p15_decrypt_enveloped_data: In %s.encryptedContentInfo.encryptedContent: Mac (appended to the end) doesn not match with calculated. Failed to decrypt the enveloped data", upname);
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		p15_add_error(errors, P15_ERROR, rc, st);
		goto end;
	}

	/* set pointers */
	*data = data2;
	*len = len2;
end:
	p15_free(ekey);
	p15_free(ekmsk);
	if (rc) p15_free(data2);
	ASN_STRUCT_FREE(asn_DEF_P15_Gost28147_89_Parameters, apar);
	return rc;
}
/****************************************************************************/
static int p15_get_random_container(p15_ctx_t* ctx, P15_PKCS15Token_t* token, 
	p15_id_t* kekid, uint8_t* pwkey, uint8_t* pwkmsk,
	P15_RandomInitValue_t** rndiv, int* warns, int* errs, p15_info_t** errors)
/* ret: 0-ok, int err, P15_RC_NO_RANDOM_CONTAINER, P15_RC_NO_AUTH_AND_DECRYPT */
{
	int rc = 0;
	p15_obj_it_t* oit = NULL;
	P15_Label_t* label = NULL;
	uint8_t flags = 0;
	OBJECT_IDENTIFIER_t* appoid = NULL;
	OBJECT_IDENTIFIER_t* dataoid = NULL;
	ANY_t* direct = NULL;
	P15_EnvelopedData_t* encrypted = NULL;
	uint8_t* ddata = NULL;
	int len;
	P15_RandomInitValue_t* r = NULL;
	void *obj = NULL;
	int otype;
		
	if (!token || !rndiv) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_get_random_container");
		return P15_RC_INTERNAL_ERROR;
	}

	*rndiv = NULL;

	oit = p15_new_obj_it(P15_OIT_DATA);

	while (1) {
		obj = NULL;
		rc = p15_obj_iterate(oit, token, &obj, &otype, warns, errors);
		/* ret: 0-found, 1-no more data objs, int err */
		if (rc > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_get_random_container:p15_obj_iterate returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc) {
			rc = P15_RC_NO_RANDOM_CONTAINER;
			goto end;
		}
		rc = p15_iterate_data_object(obj, &label, &flags, &appoid, &dataoid,
			&direct, &encrypted, warns, errs, errors);
		/* ret: 0-found, 1-continue (unsupported obj), int err */
		if (rc > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_get_random_container: p15_iterate_data_object returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc) {
			rc = 0;
			continue;
		}
		rc = p15_oid_str_eq(P15_OID_RNG, appoid);
		if (rc > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_get_random_container: p15_oid_str_eq returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (!rc) break; /* found random container */

		rc = 0; /* continue search */
	}

	if (p15_oid_eq(appoid, dataoid)) {
		*errs = 1;
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_get_random_container: applicationOID mismatches dataOID. Ignoring random container object");
		rc = P15_RC_NO_RANDOM_CONTAINER;
		goto end;
	}
			
	if (!encrypted) {
		*errs = 1;
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_get_random_container: Random init value container is not encrypted, but it should be. Ignoring the object");
		rc = P15_RC_NO_RANDOM_CONTAINER;
		goto end;
	}

	/* ck flags */
	if ((flags & 0xC0) != 0xC0) {
		rc = 0;
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS,
"p15_get_random_container: Random init value container has incorrect flags. Should have 'private' and 'modifiable' set");
	}

	/* decrypt */
	rc = p15_decrypt_enveloped_data(ctx, kekid, "RandomContainer", 
		encrypted,
		&ddata, &len, pwkey, pwkmsk, warns, errs, errors);
	if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
		p15_add_error(errors, P15_ERROR, rc, 
"p15_get_random_container: Failed to decrypt random init value container");
		goto end;
	}
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_get_random_container:p15_decrypt_enveloped_data returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* decode */
	rc = p15_asn_decode(ddata, len, &asn_DEF_P15_RandomInitValue, (void**)rndiv, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_get_random_container:p15_asn_decode returned error for rndiv");
		goto end;
	}
	if (rc) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, 
"p15_get_random_container: Failed to decode RandomInitValue structure. Skipping the object");
		rc = P15_RC_NO_RANDOM_CONTAINER;
		goto end;
	}
	r = *rndiv;
	if (r->randomInit.size != 64) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, 
"p15_get_random_container: In RandomInitValue: randomInit must be 64 bytes long. Skipping the object");
		rc = P15_RC_NO_RANDOM_CONTAINER;
		goto end;
	}

end:
	p15_free_obj_it(&oit);
	p15_free(ddata);
	if (rc) { 
		r = *rndiv;
		ASN_STRUCT_FREE(asn_DEF_P15_RandomInitValue, r);
		*rndiv = NULL;
	}
	return rc;
}
/****************************************************************************/
static int p15_process_crl_object(p15_ctx_t* ctx, P15_Label_t* label, uint8_t flags, 
	ANY_t* oany, P15_EnvelopedData_t* edata, int* warns, int* errs, 
		p15_info_t** errors)
/* ret int err */
{
	int rc = 0;
	p15_crl_t* crl = NULL;
	char st[4096];
	P15_CRLContainer_t* crlc = NULL;
	uint8_t* ebuf = NULL;
	size_t ebuflen;
	p15_der_t* tmprdn = NULL;
	p15_der_t* this_update = NULL;
	p15_der_t* next_update = NULL;
	p15_sbuf_t* sbuf = NULL;
	p15_info_t* info = NULL;
	asn_enc_rval_t erv;

	if (!ctx || !warns || !errs || (!oany && !edata) || (oany && edata)) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_process_crl_object");
		return P15_RC_INTERNAL_ERROR;
	}

	crl = p15_malloc(sizeof(p15_crl_t));
	if (label) {
		if (label->size > 0) 
			memcpy(crl->label, label->buf, label->size);
	}

	/* flags */
	if (flags) {
		*warns = 1;
		sprintf(st,
"p15_process_crl_object: In CRL '%s':DataType.oidDO.commonObjectAttributes.flags: CRL object has incorrect flags. Should have 'non-private' and 'non-modifiable'", crl->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
	}

	if (!oany) {
		*warns = 1;
		sprintf(st,
"p15_process_crl_object: In CRL '%s':DataType.oidDO.typeAttributes.value: Only 'direct' choice is supported for CRLs. Skipping the CRL object", crl->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}

	/* decoding */

	rc = ANY_to_type(oany, &asn_DEF_P15_CRLContainer, (void**)&crlc);
	if (rc || !crlc) {
		rc = 0;
		*errs = 1;
		sprintf(st,
"p15_process_crl_object: In CRL '%s':DataType.oidDO.typeAttributes.value.direct: Failed to decode CRLContainer structure. Skipping the CRL object", crl->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto end;
	}

	/* crlc->iD */
	rc = p15_process_obj_id(ctx, &crlc->iD, P15_OT_CRL, &crl->id);
	switch (rc) {
	case 0:
		break;
	case P15_RC_INVALID_TOKEN:
		*errs = 1;
		sprintf(st,
"p15_process_crl_object: In CRL '%s':CRLContainer.iD: iD has invalid format. Skipping the CRL object", crl->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		goto end;
		
	case P15_RC_CRL_ID_IN_USE:
	case P15_RC_ID_IN_USE:
		*errs = 1;
		sprintf(st,
"p15_process_crl_object: In CRL '%s':CRLContainer.iD: iD is not unique. Skipping the CRL object", crl->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = 0;
		goto end;

	default:
		sprintf(st, 
		"p15_process_crl_object: In CRL '%s':p15_process_obj_id returned error",
			crl->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* crlc->crl */
	rc = p15_asn_encode(&asn_DEF_P15_CertificateList, &crlc->crl, &ebuf, &ebuflen, errors);
	if (rc) {
		sprintf(st, 
"p15_process_crl_object: In CRL '%s':p15_asn_encode returned error for crlc-crl",
			crl->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	crl->ptr = ebuf;
	crl->len = ebuflen;
	ebuf = NULL;

	/* crl->issuer */
	rc = p15_name_to_rdn(&crlc->crl.tbsCertList.issuer, &crl->issuer);
	if (rc) {
		sprintf(st,
"p15_process_crl_object: In CRL '%s':p15_name_to_rdn returned error for tbsCertList.issuer",
			crl->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* Encoding thisUpdate back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Time, 
		&crlc->crl.tbsCertList.thisUpdate, p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_crl_object: In CRL '%s': thisUpdate encoding error.", 
			crl->label);
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		p15_add_error_type(info, erv);
		goto end;
	}
	this_update = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!this_update) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_crl_object: In CRL '%s': thisUpdate int err 1.", 
			crl->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	crl->this_update = this_update;
	this_update = NULL;

	/* Encoding nextUpdate back */
	if (!crlc->crl.tbsCertList.nextUpdate) goto no_next_update;

	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_Time, 
		crlc->crl.tbsCertList.nextUpdate, p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_crl_object: In CRL '%s': nextUpdate encoding error.", 
			crl->label);
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		p15_add_error_type(info, erv);
		goto end;
	}
	next_update = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!next_update) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_crl_object: In CRL '%s': nextUpdate int err 1.", 
			crl->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	crl->next_update = next_update;
	next_update = NULL;

no_next_update:
	/* *crlc->issuer - ck */
	if (crlc->issuer) {
		rc = p15_name_to_rdn(crlc->issuer, &tmprdn);
		if (rc) {
			sprintf(st,
"p15_process_crl_object: In CRL '%s':p15_name_to_rdn returned error for CRLContainer.issuer",
				crl->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		rc = p15_der_cmp(crl->issuer, tmprdn);
		if (rc) {
			*errs = 1;
			rc = 0;
			sprintf(st, 
"p15_process_crl_object: In CRL '%s':CRLContainer.issuer mismatches with CRLContainer.crl.tbsCertList.issuer. Skipping the CRL object", crl->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			goto end;
		}
	}

	/* insert */
	p15_add_object(ctx, P15_OT_CRL, crl);
	crl = NULL;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_CRLContainer, crlc);
	p15_free(ebuf);
	p15_free_crl(&crl);
	p15_free_der(&tmprdn);
	p15_free_der(&this_update);
	p15_free_der(&next_update);
	p15_free_sbuf(&sbuf);
	return rc;
}
/****************************************************************************/
static int p15_process_ftsver_object(p15_ctx_t* ctx, uint8_t flags, 
	ANY_t* oany, int* warns, int* errs, p15_info_t** errors)
/* ret: int err */
{
	P15_FactorTSVersion_t* ftsver = NULL;
	long v = 0; 
	long vmin = 0;
	int rc = 0;
	int is2012;
	int rc2;
	char st[1024];

	if (!ctx || !warns || !errs) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_process_ftsver_object");
		return P15_RC_INTERNAL_ERROR;
	}

	ctx->fts_ver_detected = 1;

	if (!oany) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_process_ftsver_object: In FactorTSVersion: The object should be 'direct' (unencrypted)");
		goto end;
	}
	rc = ANY_to_type(oany, &asn_DEF_P15_FactorTSVersion, (void**)&ftsver);
	if (rc || !ftsver) {
		rc = 0;
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_process_ftsver_object: In FactorTSVersion: Failed to decode FactorTSVersion");
		goto end;
	}
	if (asn_INTEGER2long(&ftsver->majorVersion, &v)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_process_ftsver_object:asn_INTEGER2long ftsver->majorVersion");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	ctx->version_major = v;
	
	if (ftsver->minorVersion) {
		if (asn_INTEGER2long(ftsver->minorVersion, &vmin)) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_process_ftsver_object:asn_INTEGER2long ftsver->minorVersion");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		ctx->version_minor = vmin;
	}

	if (v < P15_2001_FTS_MAJOR_VERSION) {
		*errs = 1;
		rc = 0;
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_process_ftsver_object: FactorTSVersion is < 3.0. Old containers are not supported");
		goto end;
	} else if (v == P15_2001_FTS_MAJOR_VERSION) is2012 = 0;
	else if (v == P15_2012_FTS_MAJOR_VERSION) is2012 = 1;
	else {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_INV_FACTOR_TS_VERSION,
"p15_process_ftsver_object: Unsupported Factor-TS container version detected");
		rc = 0;
		goto end;
	}
	rc2 = p15_ck_format(ctx, is2012, errors);
	if (rc2 == P15_RC_INVALID_CONTAINER_FORMAT) {
		*errs = 1;
		sprintf(st, 
"p15_process_ftsver_object: The Factor-TS version (%ld.%ld) mismatches the format of container (%s).",
		v, vmin, (ctx->p15_format == P15_FORMAT_2001) ? "P15-2001" : "P15-2012");
		p15_add_error(errors, P15_ERROR, rc2, st);
	}

/* Here a minor version may be checked if needed */

	/* flags */
/*ck_flags: */
	if (flags) { /* must be 0 (non-private, non-modifiable) */
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS,
"p15_process_ftsver_object: In FactorTSVersion: Data object has incorrect flags");
	}
	rc = 0;
end:
	ASN_STRUCT_FREE(asn_DEF_P15_FactorTSVersion, ftsver);
	return rc;
}
/****************************************************************************/
static int p15_process_oiddo_object(p15_ctx_t* ctx, P15_Label_t* label, 
	uint8_t flags, OBJECT_IDENTIFIER_t* id, ANY_t* oany,
	P15_EnvelopedData_t* edata, uint8_t* pwkey, uint8_t* pwkmsk,
	p15_id_t* kekid, int* warns, int* errs, p15_info_t** errors)
/* ret: int err, no auth */
{
	int rc = 0;
	p15_oid_t* uoid = NULL;
	char labelst[P15_LABEL_MAX + 1];
	int ddata_len = 0;
	uint8_t* ddata = NULL;
	int to_encrypt = 0;
	char st[4096];
	int no_auth = 0;
	OCTET_STRING_t* ostr = NULL;

	if (!ctx || !id || !warns || !errs ||
		(pwkey && (!pwkmsk || !kekid)) ||
		(!oany && !edata) || (oany && edata)) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_process_oiddo_object");
		return P15_RC_INTERNAL_ERROR;
	}

	uoid = p15_new_oid_from_asn(id);
	if (!uoid) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_process_oiddo_object:p15_new_oid_from_asn failed");
		goto end;
	}
	bzero(labelst, sizeof(labelst));
	if (label) 
		memcpy(labelst, label->buf, label->size);
	if (!edata) goto no_encrypt;

	to_encrypt = 1;
	if (!pwkey) {
		*warns = 1;
		sprintf(st, 
"p15_process_oiddo_object: In '%s' data object: Encrypted data will not be decrypted because of previous authentication failure. Skipping the object data.",
			labelst);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
			st);
		goto no_encrypt;
	}
	sprintf(st, "data object '%s':typeAttributes.value.direct-protected", 
		labelst);

	rc = p15_decrypt_enveloped_data(ctx, kekid, st, edata, 
		&ddata, &ddata_len, pwkey, pwkmsk, warns, errs, errors);
	/* ret: 0-ok, int err, no auth */
	if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
		sprintf(st,
"p15_process_oiddo_object: In data object '%s':typeAttributes.value.direct-protected: Failed to decrypt the object data", labelst);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		no_auth = 1;
		p15_free(ddata);
		ddata = NULL;
		ddata_len = 0;
		goto ck_flags;
	}
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_process_oiddo_object:p15_decrypt_enveloped_data returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	rc = p15_asn_decode(ddata, ddata_len, &asn_DEF_OCTET_STRING, (void**)&ostr, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,	
		"p15_process_oiddo_object:p15_asn_decode returned error for ostr");
		goto end;
	}
	if (rc) {
		rc = 0;
		*errs = 1;
		sprintf(st, 
"p15_process_oiddo_object: Failed to decode decrypted content in '%s' data object. (Should be OCTET STRING). Skipping the object",
			labelst); 
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto end;
	}

no_encrypt:
	if (!oany) goto ck_flags;
	rc = ANY_to_type(oany, &asn_DEF_OCTET_STRING, (void**)&ostr);
	if (rc || !ostr) {
		rc = 0;
		*errs = 1;
		sprintf(st, 
"p15_process_oiddo_object: Failed to decode typeAttributes.value.direct in data object '%s'. (Should be OCTET STRING). Skipping the data object",
			labelst);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto end;
	}


ck_flags:
	/* ck flags */
	if ((flags & 0x80) != (to_encrypt ? 0x80 : 0x00)) {
		*warns = 1;
		sprintf(st, 
"In data object '%s'.commonObjectAttributes.flags: Flags have incorrect value (0x%.2X)", labelst, flags);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
	}

	rc = p15_add_oiddo(ctx, labelst[0] ? labelst : NULL, NULL, uoid,
		ostr ? ostr->buf : NULL, ostr ? ostr->size : 0, to_encrypt, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
		"p15_process_oiddo_object:p15_add_oiddo returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
end:
	p15_free_oid(&uoid);
	p15_free(ddata);
	ASN_STRUCT_FREE(asn_DEF_OCTET_STRING, ostr);
	if (!rc && no_auth) rc = P15_RC_NO_AUTH_AND_DECRYPT;
	return rc;	
}
/****************************************************************************/
static int p15_process_data_object(p15_ctx_t* ctx, P15_Label_t* label, 
	uint8_t flags, 
	OBJECT_IDENTIFIER_t* appoid, OBJECT_IDENTIFIER_t* dataoid,
	ANY_t* oany, P15_EnvelopedData_t* edata, 
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid,
	int* warns, int* errs, p15_info_t** errors)
/* ret: 0-ok, int_err, no_auth */
{
	int rc = 0;
	int no_auth = 0;
	char labelst[P15_LABEL_MAX + 1];
	char st[4096];

	if (!ctx || !warns || !errs || !appoid || !dataoid ||
		(edata && oany) || (!edata && !oany) ||
		(pwkey && (!pwkmsk || !kekid))) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_process_data_object");
		return P15_RC_INTERNAL_ERROR;
	}
	
	bzero(labelst, sizeof(labelst));
	if (label) memcpy(labelst, label->buf, label->size);

	/* RNG */
	rc = p15_oid_str_eq(P15_OID_RNG, appoid);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_process_data_object:p15_oid_str_eq returned error for P15_OID_RNG");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (!rc) goto end; /* skipping RNG */
	rc = 0;

	/* Abstract data */
	rc = p15_oid_str_eq(P15_OID_ABSTRACT_DATA, appoid);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_process_data_object:p15_oid_str_eq returned error for P15_OID_ABSTRACT_DATA");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) { rc = 0; goto mono_oids; }
	rc = p15_process_oiddo_object(ctx, label, flags, dataoid, oany, edata,
		pwkey, pwkmsk, kekid, warns, errs, errors);
	if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
		rc = 0;
		no_auth = 1;
	}
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
	"p15_process_data_object:p15_process_oiddo_object returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	goto end;

mono_oids: 
	if (p15_oid_eq(appoid, dataoid)) goto unknown_oid;

	/* Factor-TS version */
	rc = p15_oid_str_eq(P15_OID_FACTOR_VER, appoid);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_process_data_object:p15_oid_str_eq returned error for P15_OID_FACTOR_VER");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) { rc = 0; goto crl; }

	rc = p15_process_ftsver_object(ctx, flags, oany, warns, errs, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
	"p15_process_data_object:p15_process_ftsver_object returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	goto end;
	
crl:
	rc = p15_oid_str_eq(P15_OID_CRL, appoid);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_process_data_object:p15_oid_str_eq returned error for P15_OID_CRL");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) { rc = 0; goto unknown_oid; }
	rc = p15_process_crl_object(ctx, label, flags, oany, edata, warns, errs, 
		errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_process_data_object:p15_process_crl_object returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	goto end;

unknown_oid:
	*warns = 1;
	sprintf(st, 
"p15_process_data_object: Data object '%s' has unsupported applicationOID. Skipping the object",
		labelst);
	p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	rc = 0;
end:
	if (!rc && no_auth) rc = P15_RC_NO_AUTH_AND_DECRYPT;
	return rc;
}
/****************************************************************************/
static int p15_process_private_key_object(p15_ctx_t* ctx, void* pkobj,
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid, int* warns, int* errs, 
	p15_info_t** errors)
/*ret: int err, no auth */
{
	int rc = 0;
	int rc2;
	P15_PrivateKeyType_t* pkt = pkobj;
	p15_private_key_t* key = NULL;
	P15_PKCS15ObjectGost01PrivateKey_t* pk2001 = NULL;
	P15_PKCS15ObjectGost12PrivateKey_t* pk2012 = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CommonKeyAttributes_t* cka = NULL;
	P15_CommonPrivateKeyAttributes_t* sca = NULL;
	uint8_t* ebuf = NULL;
	int ebuflen;
	char st[4096];
	P15_ObjectValueGostPrivateKey_t* ovgpk = NULL;
	P15_KeyInfoPrivateGost01_t* ki2001 = NULL;
	P15_KeyInfoPrivateGost12_t* ki2012 = NULL;
	P15_EnvelopedData_t* edata = NULL;
	P15_Gost01KeyParameters_t* pars2001 = NULL;
	P15_Gost12KeyParameters_t* pars2012 = NULL;
	int no_auth = 0;
	P15_GostPrivateKey_t* gpk = NULL;
	int kmasks = 0;
	uint8_t** key_masks = NULL;
	int i; /*j;*/
	int is2012;
	int is512 = -1; /* -1 - not defined */

	if (!ctx || !pkobj || !warns || !errs || (!pwkey && pwkmsk) ||
		(pwkey && !pwkmsk) || (pwkey && !kekid)) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
		"p15_process_private_key_object");
		return P15_RC_INTERNAL_ERROR;
	}

	key = p15_malloc(sizeof(p15_private_key_t));

	if (pkt->present == P15_PrivateKeyType_PR_privateGostR3410_2001Key) is2012 = 0;
	else if (pkt->present == P15_PrivateKeyType_PR_privateGostR3410_2012Key) is2012 = 1;
	else {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_process_private_key_object:Unsupported private key type. Skipping the object");
		goto end;
	}
	rc2 = p15_ck_format(ctx, is2012, errors);
	if (rc2 == P15_RC_INVALID_CONTAINER_FORMAT) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, rc2, 
"p15_process_private_key_object: The choice in PrivateKeyType mismatches the container format");
	} else if (rc2) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, 
			"p15_process_private_key_object: p15_ck_format returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	if (is2012) {
		pk2012 = &pkt->choice.privateGostR3410_2012Key;
		coa = &pk2012->commonObjectAttributes;
	} else {
		pk2001 = &pkt->choice.privateGostR3410_2001Key;
		coa = &pk2001->commonObjectAttributes;
	}

	/* *coa->Label */
	if (!coa->label) goto no_label;
	if (!coa->label->buf || !coa->label->size) goto no_label;
	if (coa->label->size > P15_LABEL_MAX) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_LABEL_TOO_LONG,
"p15_process_private_key_object:Object label is too long. Skipping the label");
		goto no_label;
	}
	memcpy(key->label, coa->label->buf, coa->label->size);

no_label:
	/* *coa->flags */
	if (!coa->flags) {
		*warns = 1;
		sprintf(st, 
"p15_process_private_key_object: In private key '%s':commonObjectAttributes.flags: No object flags", 
			key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}
	rc2 = p15_unpack_bitstring(coa->flags, 1, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_private_key_object: In private key '%s':p15_unpack_bitstring returned error for coa->flags", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*warns = 1;
		sprintf(st, 
"p15_process_private_key_object: In private key '%s':commonObjectAttributes.flags: Unrecoginzed flag format. Skipping flags", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto no_flags;
	}
	if ((coa->flags->buf[0] & 0xC0) != 0x80) {
		*warns = 1;
		sprintf(st, 
"p15_process_private_key_object: In private key '%s':commonObjectAttributes.flags: Incorrect object flags. Should be private and non-modifiable", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}

no_flags:
	/* *coa->authId */
	if (coa->authId) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':commonObjectAttributes: authId is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->userConsent */
	if (coa->userConsent) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':commonObjectAttributes: userConsent is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->accessControlRules  */
	if (coa->accessControlRules) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':commonObjectAttributes: accessControlRules is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* pk->classAttributes */
	if (is2012) cka = &pk2012->classAttributes;
	else cka = &pk2001->classAttributes;

	/* cka->iD */
	rc = p15_process_obj_id(ctx, &cka->iD, P15_OT_PRIVKEY, &key->id);
	switch (rc) {
	case 0:
		break;
	case P15_RC_INVALID_TOKEN:
		*errs = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.iD: iD has invalid format. Skipping the object", key->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		goto end;
		
	case P15_RC_PRIVATE_KEY_ID_IN_USE:
	case P15_RC_ID_IN_USE:
		*errs = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.iD: iD is not unique. Skipping the object", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = 0;
		goto end;

	default:
		sprintf(st, 
"p15_process_private_key_object: In private key '%s':p15_process_obj_id returned error",
			key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
	/* cka->usage */
	rc2 = p15_unpack_bitstring(&cka->usage, 2, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_private_key_object: In private key '%s':p15_unpack_bitstring returned error for cka->usage", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*errs = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.usage: Unrecognized key usage format. Ignoring the flags", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
	} else {
		key->key_usage = cka->usage.buf[0];
		key->key_usage <<= 8;
		key->key_usage |= (uint8_t)cka->usage.buf[1];
	}

	/* *cka->startDate */
	if (cka->startDate) {
		rc = p15_gt_to_utctime(cka->startDate, &key->start_date);
		if (rc == P15_RC_INVALID_TOKEN) {
			*errs = 1;
			sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.startDate: Invalid Generalized time. Ignoring the field", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			rc = 0;
		} else if (rc) {
			sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.startDate:p15_gt_to_utctime returned error", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* *cka->endDate */
	if (cka->endDate) {
		rc = p15_gt_to_utctime(cka->endDate, &key->end_date);
		if (rc == P15_RC_INVALID_TOKEN) {
			*errs = 1;
			sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.endDate: Invalid Generalized time. Ignoring the field", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			rc = 0;
		} else if (rc) {
			sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.endDate:p15_gt_to_utctime returned error", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* *cka->accessFlags */
	if (cka->accessFlags) {
		rc2 = p15_unpack_bitstring(cka->accessFlags, 1, 3);
		if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
			sprintf(st,
"p15_process_private_key_object: In private key '%s':p15_unpack_bitstring returned error for cka->accessFlags", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc2) {
			*errs = 1;
			sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.accessFlags: Unrecognized format of the field. Ignoring the flags", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		} else {
			key->key_access_flags = cka->accessFlags->buf[0];
			key->key_access_flags &= P15_KA_MASK;
			if (!key->key_access_flags) key->key_access_flags = 1;
		}
	}


	/* unsupported fields */

	if (cka->native) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.native: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}


	if (cka->keyReference) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':classAttributes.keyReference: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	sca = is2012 ? pk2012->subClassAttributes : pk2001->subClassAttributes ;
	if (!sca) goto no_subject;
	if (!sca->subjectName) goto no_subject;

	rc = p15_name_to_rdn(sca->subjectName, &key->subject);
	if (rc) {
		sprintf(st,
"p15_process_private_key_object: In private key '%s':subClassAttributes.subjectName:p15_name_to_rdn returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
no_subject:
	if (is2012) {
		ki2012 = pk2012->typeAttributes.keyInfo;
		if (!ki2012) {
			key->key_parameters = P15_PKIPAR_TC26_512_A;
			is512 = 1;
			goto no_keyinfo;
		}
	} else {
		ki2001 = pk2001->typeAttributes.keyInfo;
		if (!ki2001) {
			key->key_parameters = P15_PKIPAR_CPRO_A;
			is512 = 0;
			goto no_keyinfo;
		}
	}
	if ((is2012 ? ki2012->present : ki2001->present) != 
		(is2012 ? P15_KeyInfoPrivateGost12_PR_paramsAndOps12 :
		P15_KeyInfoPrivateGost01_PR_paramsAndOps01) ) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.keyInfo: Only paramsAndOps choice is supported. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	if (is2012 ? ki2012->choice.paramsAndOps12.supportedOperations :
		ki2001->choice.paramsAndOps01.supportedOperations) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.keyInfo.paramsAndOps.supportedOperations: The field is not supported. Skipping the field", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}
	if (is2012) pars2012 = &ki2012->choice.paramsAndOps12.parameters;
	else pars2001 = &ki2001->choice.paramsAndOps01.parameters;
	sprintf(st, "private key '%s': typeAttributes.keyInfo.paramsAndOps", key->label);
	if (is2012) rc = p15_process_key_params12(ctx, st, pars2012, &key->key_parameters, 
		warns, errs, errors);
	else rc = p15_process_key_params01(ctx, st, pars2001, &key->key_parameters, 
		warns, errs, errors);
	switch (rc) {
	case 0:
		break;
	case P15_RC_INTERNAL_ERROR:
		sprintf(st,
"p15_process_private_key_object:Private key '%s':p15_process_key_params returned internal error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	default:
		rc = 0;
		goto end;
	}
	is512 = p15_pkey_is_2012(key->key_parameters);
	
no_keyinfo:
	ovgpk = is2012 ? &pk2012->typeAttributes.value : &pk2001->typeAttributes.value;
	if (ovgpk->present != P15_ObjectValueGostPrivateKey_PR_direct_protected) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.value: Only direct-protected choice is supported. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	edata = &ovgpk->choice.direct_protected;

	/* ck pwkey */
	if (!pwkey) {
		*warns = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.value.direct-protected: Encrypted data will not be decrypted because of previous authentication failure. Skipping the key data", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto no_key_data;
	}

	/* decrypt */
	sprintf(st, "private key '%s':typeAttributes.value.direct-protected", 
		key->label);
	rc = p15_decrypt_enveloped_data(ctx, kekid, st, edata, &ebuf, &ebuflen, 
		pwkey, pwkmsk, warns, errs, errors);
	if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.value.direct-protected: Failed to decrypt the key data", key->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		no_auth = 1;
		goto no_key_data;
	}
	if (rc) {
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.value.direct-protected: p15_decrypt_enveloped_data returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* decode */
	rc = p15_asn_decode(ebuf, ebuflen, &asn_DEF_P15_GostPrivateKey, (void**)&gpk, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.value.direct-protected: p15_asn_decode returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	if (rc) {
		*errs = 1;
		no_auth = 1;
		rc = 0;
		sprintf(st,
"p15_process_private_key_object: In private key '%s':typeAttributes.value.direct-protected: p15_asn_decode returned error", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto no_key_data;
	}

	/* key data */
	if (is512 < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_private_key_object: In private key '%s': is512 < 0 (not set)", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	kmasks = gpk->size;
	if (kmasks & (is512 ? 0x3F : 0x1F)) {
		*errs = 1;
		no_auth = 1;
		sprintf(st,
"p15_process_private_key_object: In private key '%s': KeyValueMask must be multiple of %s",
			key->label, is512 ? "64" : "32");
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto no_key_data;
	}
	kmasks >>= is512 ? 6 : 5;

	key_masks = p15_malloc(kmasks * sizeof(uint8_t*));
	for (i = 0; i < kmasks; i++) {
		key_masks[i] = p15_malloc(P15_PAGE);
		memcpy(key_masks[i], gpk->buf + (i << (is512 ? 6 : 5)), is512 ? 64 : 32);
	}
	key->key_masks = key_masks;
	key->mask_count = kmasks;
	key_masks = NULL;

no_key_data:
	p15_add_object(ctx, P15_OT_PRIVKEY, key);
	key = NULL;
	if (!rc && no_auth) rc = P15_RC_NO_AUTH_AND_DECRYPT;
end:
	if (key_masks) {
		for (i = 0; i < kmasks; i++) p15_free(key_masks[i]);
		p15_free(key_masks);
	}
	p15_free_private_key(&key);
	ASN_STRUCT_FREE(asn_DEF_P15_GostPrivateKey, gpk);
	p15_free(ebuf);
	return rc;
}
/****************************************************************************/
static int p15_process_key_params01(p15_ctx_t* ctx, char* parent, 
	P15_Gost01KeyParameters_t* pars, int* keypar, 
	int* warns, int* errs, p15_info_t** errors)
/* ret: ok, int err, unsupported alg, inv token */
{
	char st[4096];
	int rc = 0;

	if (!ctx || !parent || !pars || !keypar || !warns || !errs) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
		"p15_process_prkey_params01");
		return P15_RC_INTERNAL_ERROR;
	}

	if (pars->present == P15_Gost01KeyParameters_PR_cryptoProParamSet) {
		rc = p15_oid_to_pkipar(&pars->choice.cryptoProParamSet, keypar, 1);
		if (rc == P15_RC_UNSUPPORTED_ALGORITHM) {
			*warns = 1;
			sprintf(st,
"p15_process_key_params01: In %s.parameters.cryptoProParamSet: OID is not recognized. Unsupported algorithm. Skipping the key object", parent);
			p15_add_error(errors, P15_WARNING, P15_RC_UNSUPPORTED_ALGORITHM, st);
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
			goto end;
		}
		if (rc) {
			sprintf(st,
"p15_process_key_params01: In %s.parameters.cryptoProParamSet: p15_oid_to_pkipar returned error",
				parent);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	} else if (pars->present == P15_Gost01KeyParameters_PR_privateKeyParamSet) {
		*warns = 1;
		sprintf(st,
"p15_process_key_params01: In %s.parameters: 'privateKeyParamSet' choice is not supported. Skipping the key object.", parent);
		p15_add_error(errors, P15_WARNING, P15_RC_UNSUPPORTED_ALGORITHM, st);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	} else if (pars->present == P15_Gost01KeyParameters_PR_secretKeyParamSet) {
		*errs = 1;
		sprintf(st,
"p15_process_key_params01: In %s.parameters: 'secretKeyParamSet' choice detected. Illegal choice. Skipping the key object", parent);
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc, st);
		goto end;
	} else {
		*warns = 1;
		sprintf(st,
"p15_process_key_params01: In %s.parameters: Unsupported choice detected. Skipping the key object", parent);
		rc = P15_RC_UNSUPPORTED_ALGORITHM;
		p15_add_error(errors, P15_WARNING, rc, st);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
end:
	return rc;
}
/****************************************************************************/
static int p15_process_key_params12(p15_ctx_t* ctx, char* parent, 
	P15_Gost12KeyParameters_t* pars, int* keypar, 
	int* warns, int* errs, p15_info_t** errors)
/* ret: ok, int err, unsupported alg, inv token */
{
	char st[4096];
	int rc = 0;

	if (!ctx || !parent || !pars || !keypar || !warns || !errs) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
		"p15_process_prkey_params12");
		return P15_RC_INTERNAL_ERROR;
	}

	if (pars->present == P15_Gost12KeyParameters_PR_gostR3410_2012ParamSet) {
		rc = p15_oid_to_pkipar(&pars->choice.gostR3410_2012ParamSet, keypar, 0);
		if (rc == P15_RC_UNSUPPORTED_ALGORITHM) {
			*warns = 1;
			sprintf(st,
"p15_process_key_params12: In %s.parameters.gostR3410-2012ParamSet: OID is not recognized. Unsupported algorithm. Skipping the key object", parent);
			p15_add_error(errors, P15_WARNING, P15_RC_UNSUPPORTED_ALGORITHM, st);
			p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
			goto end;
		}
		if (rc) {
			sprintf(st,
"p15_process_key_params12: In %s.parameters.gostR3410-2012ParamSet: p15_oid_to_pkipar returned error",
				parent);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	} else if (pars->present == P15_Gost12KeyParameters_PR_privateKeyParamSet) {
		*warns = 1;
		sprintf(st,
"p15_process_key_params12: In %s.parameters: 'privateKeyParamSet' choice is not supported. Skipping the key object.", parent);
		p15_add_error(errors, P15_WARNING, P15_RC_UNSUPPORTED_ALGORITHM, st);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	} else {
		*warns = 1;
		sprintf(st,
"p15_process_key_params12: In %s.parameters: Unsupported choice detected. Skipping the key object", parent);
		rc = P15_RC_UNSUPPORTED_ALGORITHM;
		p15_add_error(errors, P15_WARNING, rc, st);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
end:
	return rc;
}
/****************************************************************************/
static int p15_process_public_key_object(p15_ctx_t* ctx, void* pkobj,
	int* warns, int* errs, p15_info_t** errors)
/*ret: int err */
{
	int rc = 0;
	int rc2;
	P15_PublicKeyType_t* pkt = pkobj;
	p15_public_key_t* key = NULL;
	P15_PKCS15ObjectGost01PublicKey_t* pk2001 = NULL;
	P15_PKCS15ObjectGost12PublicKey_t* pk2012 = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CommonKeyAttributes_t* cka = NULL;
	P15_CommonPublicKeyAttributes_t* sca = NULL;
	char st[4096];
	P15_ObjectValueGostPublicKeyChoice_t* ovgpk = NULL;
	P15_KeyInfoPublicGost01_t* ki2001 = NULL;
	P15_KeyInfoPublicGost12_t* ki2012 = NULL;
	P15_Gost01KeyParameters_t* pars2001 = NULL;
	P15_Gost12KeyParameters_t* pars2012 = NULL;
	P15_GostPublicKeyChoice_t* gpkc = NULL;
	OCTET_STRING_t* rawkey = NULL;
	int is2012;
	int is1024 = -1; /* -1 - not defined */

	if (!ctx || !pkobj || !warns || !errs) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
		"p15_process_public_key_object");
		return P15_RC_INTERNAL_ERROR;
	}

	key = p15_malloc(sizeof(p15_public_key_t));

	if (pkt->present == P15_PublicKeyType_PR_publicGostR3410_2001Key) is2012 = 0;
	else if (pkt->present == P15_PublicKeyType_PR_publicGostR3410_2012Key) is2012 = 1;
	else {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_process_public_key_object:Unsupported public key type. Skipping the object");
		goto end;
	}

	if (is2012) {
		pk2012 = &pkt->choice.publicGostR3410_2012Key;
		coa = &pk2012->commonObjectAttributes;
	} else {
		pk2001 = &pkt->choice.publicGostR3410_2001Key;
		coa = &pk2001->commonObjectAttributes;
	}

	/* *coa->Label */
	if (!coa->label) goto no_label;
	if (!coa->label->buf || !coa->label->size) goto no_label;
	if (coa->label->size > P15_LABEL_MAX) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_LABEL_TOO_LONG,
"p15_process_public_key_object:Object label is too long. Skipping the label");
		goto no_label;
	}
	memcpy(key->label, coa->label->buf, coa->label->size);

no_label:
	/* *coa->flags */
	if (!coa->flags) {
		*warns = 1;
		sprintf(st, 
"p15_process_public_key_object: In public key '%s':commonObjectAttributes.flags: No object flags", 
			key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}

	rc2 = p15_unpack_bitstring(coa->flags, 1, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_public_key_object: In public key '%s':p15_unpack_bitstring returned error for coa->flags", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*warns = 1;
		sprintf(st, 
"p15_process_public_key_object: In public key '%s':commonObjectAttributes.flags: Unrecoginzed flag format. Skipping flags", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto no_flags;
	}
	if ((coa->flags->buf[0] & 0xC0) != 0x00) {
		*warns = 1;
		sprintf(st, 
"p15_process_public_key_object: In public key '%s':commonObjectAttributes.flags: Incorrect object flags. Should be non-private and non-modifiable", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}
no_flags:
	/* *coa->authId */
	if (coa->authId) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':commonObjectAttributes: authId is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->userConsent */
	if (coa->userConsent) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':commonObjectAttributes: userConsent is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->accessControlRules  */
	if (coa->accessControlRules) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':commonObjectAttributes: accessControlRules is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* pk->classAttributes */
	if (is2012) cka = &pk2012->classAttributes;
	else cka = &pk2001->classAttributes;

	/* cka->iD */
	rc = p15_process_obj_id(ctx, &cka->iD, P15_OT_PUBKEY, &key->id);
	switch (rc) {
	case 0:
		break;
	case P15_RC_INVALID_TOKEN:
		*errs = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.iD: iD has invalid format. Skipping the object", key->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		goto end;
		
	case P15_RC_PUBLIC_KEY_ID_IN_USE:
	case P15_RC_ID_IN_USE:
		*errs = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.iD: iD is not unique. Skipping the object", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = 0;
		goto end;

	default:
		sprintf(st, 
"p15_process_public_key_object: In public key '%s':p15_process_obj_id returned error",
			key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
	/* cka->usage */
	rc2 = p15_unpack_bitstring(&cka->usage, 2, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_public_key_object: In public key '%s':p15_unpack_bitstring returned error for cka->usage", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*errs = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.usage: Unrecognized key usage format. Ignoring the flags", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
	} else {
		key->key_usage = cka->usage.buf[0];
		key->key_usage <<= 8;
		key->key_usage |= (uint8_t)cka->usage.buf[1];
	}

	/* *cka->startDate */
	if (cka->startDate) {
		rc = p15_gt_to_utctime(cka->startDate, &key->start_date);
		if (rc == P15_RC_INVALID_TOKEN) {
			*errs = 1;
			sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.startDate: Invalid Generalized time. Ignoring the field", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			rc = 0;
		} else if (rc) {
			sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.startDate:p15_gt_to_utctime returned error", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* *cka->endDate */
	if (cka->endDate) {
		rc = p15_gt_to_utctime(cka->endDate, &key->end_date);
		if (rc == P15_RC_INVALID_TOKEN) {
			*errs = 1;
			sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.endDate: Invalid Generalized time. Ignoring the field", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			rc = 0;
		} else if (rc) {
			sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.endDate:p15_gt_to_utctime returned error", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	if (cka->native) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.native: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (cka->accessFlags) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.accessFlags: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (cka->keyReference) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':classAttributes.keyReference: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (is2012) sca = pk2012->subClassAttributes;
	else sca = pk2001->subClassAttributes;
	if (!sca) goto no_subject;
	if (!sca->subjectName) goto no_subject;

	rc = p15_name_to_rdn(sca->subjectName, &key->subject);
	if (rc) {
		sprintf(st,
"p15_process_public_key_object: In public key '%s':subClassAttributes.subjectName:p15_name_to_rdn returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
no_subject:
	if (is2012) {
		ki2012 = pk2012->typeAttributes.keyInfo;
		if (!ki2012) {
			key->key_parameters = P15_PKIPAR_TC26_512_A;
			is1024 = 1;
			goto no_keyinfo;
		}
	} else {
		ki2001 = pk2001->typeAttributes.keyInfo;
		if (!ki2001) {
			key->key_parameters = P15_PKIPAR_CPRO_A;
			is1024 = 0;
			goto no_keyinfo;
		}
	}
	if ((is2012 ? ki2012->present : ki2001->present) != (is2012 ?
		P15_KeyInfoPublicGost12_PR_paramsAndOpsPub12 :
		P15_KeyInfoPublicGost01_PR_paramsAndOpsPub01)) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':typeAttributes.keyInfo: Only paramsAndOps choice is supported. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	if (is2012 ? ki2012->choice.paramsAndOpsPub12.supportedOperations :
		ki2001->choice.paramsAndOpsPub01.supportedOperations) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':typeAttributes.keyInfo.paramsAndOps.supportedOperations: The field is not supported. Skipping the field", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (is2012) pars2012 = &ki2012->choice.paramsAndOpsPub12.parameters;
	else pars2001 = &ki2001->choice.paramsAndOpsPub01.parameters;
	
	sprintf(st, "public key '%s':typeAttributes.keyInfo.paramsAndOps.parameters",
		key->label);

	if (is2012) rc = p15_process_key_params12(ctx, st, pars2012, &key->key_parameters,
		warns, errs, errors);
	else rc = p15_process_key_params01(ctx, st, pars2001, &key->key_parameters,
		warns, errs, errors);
	switch (rc) {
	case 0:
		break;
	case P15_RC_INTERNAL_ERROR:
		sprintf(st,
"p15_process_public_key_object:Public key '%s':p15_process_key_params returned internal error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	default:
		rc = 0;
		goto end;
	}
	is1024 = p15_pkey_is_2012(key->key_parameters);
no_keyinfo:
	ovgpk = is2012 ? &pk2012->typeAttributes.value : &pk2001->typeAttributes.value ;
	if (ovgpk->present != P15_ObjectValueGostPublicKeyChoice_PR_direct) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':typeAttributes.value: Only 'direct' choice is supported. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	gpkc = &ovgpk->choice.direct;
	if (gpkc->present != P15_GostPublicKeyChoice_PR_raw) {
		*warns = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':typeAttributes.value.direct: Only 'raw' choice is supported for now. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	rawkey = &gpkc->choice.raw;

	if (is1024 < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_process_public_key_object: is1024 < 0 (not set)");
		goto end;
	}
	if (rawkey->size != (is1024 ? 128 : 64)) {
		*errs = 1;
		sprintf(st,
"p15_process_public_key_object: In public key '%s':The key should be %s-byte long. Skiping the key object", key->label, is1024 ? "128" : "64");
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto end;
	}

	/* key is in little-endian */
	memcpy(key->key, rawkey->buf, is1024 ? 128 : 64);

	p15_add_object(ctx, P15_OT_PUBKEY, key);
	key = NULL;
end:
	p15_free_public_key(&key);
	return rc;
}
/****************************************************************************/
static int p15_process_cert_object(p15_ctx_t* ctx, void* cobj, int trusted, 
	int* warns, int* errs, p15_info_t** errors)
/* ret int err */
{
	int rc = 0;
	int rc2;
	char st[4096];
	P15_CertificateType_t* ct = cobj;
	P15_PKCS15ObjectX509Certificate_t* x509 = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CommonCertificateAttributes_t* cca = NULL;
	P15_X509CertificateAttributes_t* ta = NULL;
	p15_certificate_t* cer = NULL;
	p15_der_t* issuer = NULL;
	p15_der_t* subject = NULL;
	p15_der_t* serial_number = NULL;
	p15_der_t* tmpder = NULL;
	P15_Certificate_t* cert = NULL;
	INTEGER_t *sn1, *sn2;
	uint8_t* ebuf = NULL;
	size_t ebuflen;
	p15_sbuf_t* sbuf = NULL;
	asn_enc_rval_t erv;
	p15_info_t* info = NULL;

	if (!ctx || !cobj || !warns || !errs) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
			"p15_process_cert_object");
		return P15_RC_INTERNAL_ERROR;
	}

	if (ct->present != P15_CertificateType_PR_x509Certificate) {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_process_cert_object: Unsupported type of certificate. Ignoring the object");
		return 0;
	}
	x509 = &ct->choice.x509Certificate;

	/* x509->commonObjectAttributes */
	coa = &x509->commonObjectAttributes;

	cer = p15_malloc(sizeof(p15_certificate_t));

	/* *coa->label */
	if (coa->label) {
		if (!coa->label->size) goto no_label;
		if (coa->label->size > P15_LABEL_MAX) {
			*errs = 1;
			p15_add_error(errors, P15_ERROR, P15_RC_LABEL_TOO_LONG,
"p15_process_cert_object: In CertificateType.commonObjectAttributes.label: Label is too long. Skipping the label");
			goto no_label;
		} else
			memcpy(cer->label, coa->label->buf, coa->label->size); 
	}
no_label:

	/* *coa->flags */
	if (!coa->flags) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s':commonObjectAttributes.flags: No object flags", 
			cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}
	rc2 = p15_unpack_bitstring(coa->flags, 1, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_cert_object: In certificate '%s':p15_unpack_bitstring returned error for coa->flags", cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s':commonObjectAttributes.flags: Unrecoginzed flag format. Skipping flags", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto no_flags;
	}
	if ((coa->flags->buf[0] & 0xC0) != 0x00) { /* non-private, non-modifiable */
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s':commonObjectAttributes.flags: Incorrect object flags. Should be private and non-modifiable", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}

no_flags:
	/* *coa->authId */
	if (coa->authId) {
		*warns = 1;
		sprintf(st,
"p15_process_cert_object: In certificate '%s':commonObjectAttributes: authId is not supported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->userConsent */
	if (coa->userConsent) {
		*warns = 1;
		sprintf(st,
"p15_process_cert_object: In certificate '%s':commonObjectAttributes: userConsent is not supported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->accessControlRules  */
	if (coa->accessControlRules) {
		*warns = 1;
		sprintf(st,
"p15_process_cert_object: In certificate '%s':commonObjectAttributes: accessControlRules is not supported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* x509->classAttributes */
	cca = &x509->classAttributes;

	/* cca->iD */
	rc = p15_process_obj_id(ctx, &cca->iD, trusted ? P15_OT_TRUSTED_CERT : P15_OT_CERT,
		&cer->id);
	switch (rc) {
	case 0:
		break;

	case P15_RC_INVALID_TOKEN:
		*errs = 1;
		sprintf(st,
"p15_process_cert_object: In certificate '%s':classAttributes.iD: iD has invalid format. Skipping the object", cer->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		goto end;
		
	case P15_RC_CERT_ID_IN_USE:
	case P15_RC_ID_IN_USE:
		*errs = 1;
		sprintf(st,
"p15_process_cert_object: In certificate '%s':classAttributes.iD: iD is not unique. Skipping the object", cer->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = 0;
		goto end;

	default:
		sprintf(st, 
"p15_process_cert_object: In certificate '%s':p15_process_obj_id returned error",
			cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* *cca->authority */
	if (cca->authority) {
		if (*cca->authority) cer->ca = 1;
	}

	/* cca unsupported fields */

	if (cca->identifier) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.classAttributes.identifier: The field is unsupported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (cca->certHash) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.classAttributes.certHash: The field is unsupported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (cca->trustedUsage) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.classAttributes.trustedUsage: The field is unsupported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (cca->identifiers) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.classAttributes.identifiers: The field is unsupported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (cca->implicitTrust) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.classAttributes.identifiers: The field is unsupported and ignored", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* x509->typeAttributes */
	ta = &x509->typeAttributes;
	
	/* *ta->subject */
	if (ta->subject) {
		rc = p15_name_to_rdn(ta->subject, &subject);
		if (rc) {
			sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.subject: p15_name_to_rdn returned error", cer->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* *ta->issuer */
	if (ta->issuer) {
		rc = p15_name_to_rdn(ta->issuer, &issuer);
		if (rc) {
			sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.issuer: p15_name_to_rdn returned error", cer->label);

			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* ta->value */
	if (ta->value.present != P15_ObjectValueCertificate_PR_direct) {
		*warns = 1;
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.value: Only 'direct' choice is supported. Skipping the certificate object", cer->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	cert = &ta->value.choice.direct;

	/* ck serial number */
	sn1 = &cert->tbsCertificate.serialNumber;
	sn2 = ta->serialNumber;
	if (sn2) {
		if (!sn1->size || !sn1->buf || !sn2->buf || sn1->size != sn2->size) rc = 1;
		if (!rc) { 
			if (memcmp(sn1->buf, sn2->buf, sn1->size)) rc = 1;
		}
		if (rc) {
			rc = 0;
			*errs = 1;
			sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.serialNumber: Only 'direct' choice is supported. Skipping the certificate object", cer->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		}
	}
	/* encode serial number back */
	sbuf = p15_add_sbuf(NULL, NULL, 0);
	erv = der_encode(&asn_DEF_P15_CertificateSerialNumber, 
		sn1, p15_consume_bytes, sbuf);
	if (erv.encoded < 0) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_cert_object: In certificate '%s': serialNumber encoding error.", 
			cer->label);
		info = p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		p15_add_error_type(info, erv);
		goto end;
	}
	serial_number = p15_new_der_from_sbuf(sbuf);
	p15_free_sbuf(&sbuf);
	if (!serial_number) {
		rc = P15_RC_INTERNAL_ERROR;
		sprintf(st,
"p15_process_cert_object: In certificate '%s': serialNumber int err 2.", 
			cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}

	/* ck subject */
	rc = p15_name_to_rdn(&cert->tbsCertificate.subject, &tmpder);
	if (rc) {
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.value.direct.tbsCertificate.subject: p15_name_to_rdn returned error", cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (!subject) {
		subject = tmpder;
		tmpder = NULL;
	} else {
		rc = p15_der_cmp(subject, tmpder);
		if (rc) {
			*errs = 1;
			rc = 0;
			sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.value.direct.tbsCertificate.subject: Mismatch with typeAttributes.subject. Skipping the certificate object", cer->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			goto end;
		}
	}
	p15_free_der(&tmpder);

	/* ck issuer */
	rc = p15_name_to_rdn(&cert->tbsCertificate.issuer, &tmpder);
	if (rc) {
		sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.value.direct.tbsCertificate.issuer: p15_name_to_rdn returned error", cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (!issuer) {
		issuer = tmpder;
		tmpder = NULL;
	} else {
		rc = p15_der_cmp(issuer, tmpder);
		if (rc) {
			*errs = 1;
			rc = 0;
			sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.value.direct.tbsCertificate.issuer: Mismatch with typeAttributes.issuer. Skipping the certificate object", cer->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			goto end;
		}
	}
	p15_free_der(&tmpder);

	cer->issuer = issuer;
	cer->subject = subject;
	cer->serial_number = serial_number;
	issuer = NULL;
	subject = NULL;
	serial_number = NULL;

	/* reencoding cert */
	rc = p15_asn_encode(&asn_DEF_P15_Certificate, cert, &ebuf, &ebuflen, errors);
	if (rc) {
	sprintf(st, 
"p15_process_cert_object: In certificate '%s'.typeAttributes.value.direct: p15_asn_encode returned error", cer->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	cer->ptr = ebuf;
	ebuf = NULL;
	cer->len = ebuflen;

	/* adding cert */
	p15_add_object(ctx, trusted ? P15_OT_TRUSTED_CERT : P15_OT_CERT, cer);
	cer = NULL;

end:
	p15_free(ebuf);
	p15_free_cert(&cer);
	p15_free_der(&issuer);
	p15_free_der(&subject);
	p15_free_der(&serial_number);
	p15_free_der(&tmpder);
	p15_free_sbuf(&sbuf);
	return rc;
}
/****************************************************************************/
static int p15_load_objects(p15_ctx_t* ctx, P15_PKCS15Token_t* ptoken, 
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid, int* warns, int* errs, 
	p15_info_t** errors)
/*ret: 0-ok, no_auth, int_err */
{
	int rc = 0;
	p15_obj_it_t* oit = NULL;
	P15_Label_t* label;
	uint8_t flags;
	OBJECT_IDENTIFIER_t* appoid = NULL;
	OBJECT_IDENTIFIER_t* dataoid = NULL;
	ANY_t* oany;
	P15_EnvelopedData_t* edata;
	int no_auth = 0;
	void* uobj = NULL;
	int ign = 0;
	char st[4096];
	int otype;

	if (!ctx || !ptoken || !errs || !warns ||
		(!pwkey && pwkmsk) || (pwkey && !pwkmsk) ||
		(pwkey && !kekid)) {
		p15_add_error(errors, P15_INTERNAL_ERROR, 
			P15_RC_INVALID_ARGS, "p15_load_objects");
		return P15_RC_INTERNAL_ERROR;
	}

	/* Iterating objects (except random container) */
	oit = p15_new_obj_it(P15_OIT_ANY);
	while (1) {
		uobj = NULL;
		rc = p15_obj_iterate(oit, ptoken, &uobj, &otype, warns, errors);
		/* ret: 0-ok (object returned), 1-no objects left, >1-internal error*/
		if (rc > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
				"p15_load_objects:p15_obj_iterate returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc) { /* no more objs */
			rc = 0;
			break;
		}

		switch (otype) {
		case P15_OIT_CERT:
			rc = p15_process_cert_object(ctx, uobj, 0, warns, errs, errors);
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_load_objects:p15_process_cert_object (non root certs) returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			break;

		case P15_OIT_RCERT:
			rc = p15_process_cert_object(ctx, uobj, 1, warns, errs, errors);
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_load_objects:p15_process_cert_object (root certs) returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			break;

		case P15_OIT_PRKEY:
			rc = p15_process_private_key_object(ctx, uobj, pwkey, pwkmsk, kekid,
				warns, errs, errors);
			if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
				no_auth = 1;
				rc = 0;
			}
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_objects:p15_process_private_key_object returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			break;

		case P15_OIT_DATA:
			rc = p15_iterate_data_object(uobj, &label, &flags, 
				&appoid, &dataoid, &oany, &edata, warns, errs, errors);
			if (rc > 1) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_load_objects:p15_iterate_data_object returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			if (rc) { /* not supported data obj - continue */
				rc = 0; 
				break; 
			}
			rc = p15_process_data_object(ctx, label, flags, appoid, dataoid,
				oany, edata, pwkey, pwkmsk, kekid,
				warns, errs, errors);
			if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
				no_auth = 1;
				rc = 0;
			}
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
				"p15_load_objects:p15_process_data_object returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			break;

		case P15_OIT_PUBKEY:
			rc = p15_process_public_key_object(ctx, uobj, warns, errs, errors);
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_objects:p15_process_public_key_object returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			break;

		case P15_OIT_SECKEY:
			rc = p15_process_secret_key_object(ctx, uobj, pwkey, pwkmsk, kekid,
				warns, errs, errors);
			if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
				no_auth = 1;
				rc = 0;
			}
			if (rc) {
				p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_objects:p15_process_secret_key_object returned error");
				rc = P15_RC_INTERNAL_ERROR;
				goto end;
			}
			break;

		default:
			sprintf(st, "p15_load_objects: Unexpected otype (%d)", otype);
			rc = P15_RC_INTERNAL_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			goto end;
		}
	}
	p15_free_obj_it(&oit);

	/* Unknown object categories report */
	oit = p15_new_obj_it(P15_OIT_UNKNOWN);
	do {
		rc = p15_obj_iterate(oit, ptoken, &uobj, &otype, warns, errors);
		ign++;
	} while (!rc);
	ign--;
	if (rc > 1) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_objects:p15_obj_iterate returned error for unknown objs");
		goto end;
	}
	rc = 0;
	if (ign) {
		*warns = 1;
		sprintf(st, 
"p15_load_objects: %d unsupported object group(s) were skipped", ign);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	if (!rc && no_auth) rc = P15_RC_NO_AUTH_AND_DECRYPT;
end:
	p15_free_obj_it(&oit);
	return rc;
}
/****************************************************************************/
int p15_load_token(p15_ctx_t* ctx, uint8_t* token, size_t token_len, 
	uint8_t* password, int passlen, p15_info_t** errors)
{
	int rc = 0;
	int rc2;
	int warns = 0;
	int no_auth = 0;
	int terr = 0; /* token_error */
	int wrong_pass = 0;
	P15_AuthenticatedData_t* authdata = NULL;
	long v;
	uint8_t* pwkey = NULL;
	uint8_t* pwkmsk = NULL;
	uint8_t* akey = NULL;
	uint8_t* akmsk = NULL;
	p15_id_t* kekid = NULL;
	P15_PKCS15Token_t* ptoken = NULL;
	uint8_t mac[64];
	uint8_t macc[64];
	uint8_t hash[64];
	uint8_t hashc[64];
	void* hh01 = NULL;
	gostc_hash12_ctx_t hh12;
	gostc_hmac12_ctx_t hm12;
	uint8_t key[32];
	int i;
	uint8_t* ebuf = NULL;
	size_t ebuflen;
	P15_RandomInitValue_t* rndiv = NULL;
	int hmac2012;
	int hash2012;
	char st[1024];

	if (!ctx || !token || !token_len || (!password && passlen)) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc; 
	}

	/* free old data */
	p15_free_ctx_content(ctx);

	/* Decoding */

	rc = p15_asn_decode(token, token_len, &asn_DEF_P15_AuthenticatedData, 
		(void**)&authdata, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
		"p15_load_token:p15_asn_decode returned error for authdata");
		goto end;
	}
	if (rc) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
		"p15_load_token: Failed to parse AuthenticatedData structure");
		goto end;
	}

	/* authdata->version - must be 0 */
	rc = asn_INTEGER2long(&authdata->version, &v);
	if (!rc && v) rc = 1;
	if (rc) { 
		terr = 1; 
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
		"p15_load_token: AuthenticatedData has incompatible version");
		rc = 0; 
	}

	/* *authdata->originatorInfo */
	if (authdata->originatorInfo) {
		warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
		"p15_load_token: Ignoring AuthenticatedData.originatorInfo");
	}

	/* authdata->encapContentInfo.eContentType */
	/*-- should be 1.2.840.113549.1.15.3.1 */

	rc = p15_oid_str_eq(P15_OID_CONTENT_TYPE, 
		&authdata->encapContentInfo.eContentType);
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_load_token:p15_oid_str_eq failed for authdata->encapContentInfo.eContentType");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
"p15_load_token:AuthenticatedData.encapContentInfo.eContentType contains unrecognized OID");
		goto end;
	}

	/* *authdata->encapContentInfo.eContent */
	if (!authdata->encapContentInfo.eContent) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
		"p15_load_token:AuthenticatedData.encapContentInfo.eContent is absent");
		goto end;
	}

	rc = p15_asn_decode(authdata->encapContentInfo.eContent->buf, 
		authdata->encapContentInfo.eContent->size, &asn_DEF_P15_PKCS15Token, 
		(void**)&ptoken, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
"p15_load_token:p15_asn_decode returned error for authdata->encapContentInfo.eContent");
		goto end;
	}
	if (rc) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
		"p15_load_token: Failed to parse AuthenticatedData.encapContentInfo.eContent");
		goto end;
	}

	/* ptoken->version */
	rc = asn_INTEGER2long(&ptoken->version, &v);
	if (!rc && v) rc = 1;
	if (rc) { 
		terr = 1; 
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc,
		"p15_load_token: PKCS15Token has incompatible version");
		rc = 0; 
	}

	/* *ptoken->keyManagementInfo */
/* p15_load_kmi also performs the detection of the container format P15-2001/2012 */
	rc = p15_load_kmi(ctx, ptoken->keyManagementInfo, password, passlen, 
		&kekid, &pwkey, &pwkmsk, errors);
	/* ret 0, 1-warns, P15_RC_INTERNAL_ERROR, errors */
	switch (rc) {
		case 0: break;
		case 1: rc = 0; warns = 1; break;
		case P15_RC_INTERNAL_ERROR:
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
				"p15_load_token:p15_load_kmi");
			goto end;
		default: rc = 0; terr = 1; break;
	}

	if (!pwkey || !pwkmsk || !kekid) {
		no_auth = 1;
		goto no_auth_clr;
	}

	/* authdata->recipientInfos */
	rc = p15_load_kekris(ctx, "AuthenticatedData", &authdata->recipientInfos,
		kekid, pwkey, pwkmsk, &akey, &akmsk, 1, errors);
/* ret 0, 1-warns, P15_RC_WRONG_PASSWORD, P15_RC_INTERNAL_ERROR, P15_RC_NO_AUTH_AND_DECRYPT, */
/* other not important errors */
	switch (rc) {
		case 0: break;
		case 1: warns = 1; rc = 0; break;
		case P15_RC_WRONG_PASSWORD:
			rc = 0;
			wrong_pass = 1; 
			goto no_auth_clr;
		case P15_RC_NO_AUTH_AND_DECRYPT:
			rc = 0;
			no_auth = 1;
			goto no_auth_clr;
		case P15_RC_INTERNAL_ERROR: 
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
	"p15_load_token:p15_load_kekris returned error for AuthenticatedData");
			goto end;
		default:
			terr = 1; rc = 0;
	}
	
	/* *authdata->digestAlgorithm */
	if (!authdata->digestAlgorithm) {
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
		"p15_load_token: AuthenticatedData.digestAlgorithm is absent");
		no_auth = 1;
		goto no_auth_clr;
	}

	/* authdata->digestAlgorithm->algorithm */
	rc = p15_oid_str_eq(P15_OID_GOST_R3411_94, &authdata->digestAlgorithm->algorithm);
	if (!rc) hash2012 = 0;
	else if (rc == 1) {
		rc = p15_oid_str_eq(P15_OID_GOST_R3411_2012_512,
			&authdata->digestAlgorithm->algorithm);
		if (!rc) hash2012 = 1;
	}
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_load_token:p15_oid_str_eq returned error for authdata->digestAlgorithm->algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) {
		rc = 0;
		no_auth = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, 
"p15_load_token:AuthenticatedData.digestAlgorithm.algorithm contains unrecognized OID");
		goto no_auth_clr;
	}
	rc2 = p15_ck_format(ctx, hash2012, errors);
	if (rc2 == P15_RC_INVALID_CONTAINER_FORMAT) {
		terr = 1;
		p15_add_error(errors, P15_ERROR, rc2, 
"p15_load_token:In AuthenticatedData.digestAlgorithm.algorithm: Unexpected hash algorithm for this container format.");
	} else if (rc2) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2,
"p15_load_token:AuthenticatedData.digestAlgorithm.algorithm: p15_ck_format returned error.");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* *authdata->digestAlgorithm->parameters */
	if (authdata->digestAlgorithm->parameters) {
		if (p15_ck_ANY_NULL(authdata->digestAlgorithm->parameters)) {
			p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, 
"p15_load_token:AuthenticatedData.digestAlgorithm.parameters are present and not NULL");
			no_auth = 1;
			goto no_auth_clr;
		}
	}

	/* authdata->macAlgorithm.algorithm */
	rc = p15_oid_str_eq(P15_OID_HMAC_GOST_94, &authdata->macAlgorithm.algorithm);
	if (!rc) hmac2012 = 0;
	else if (rc == 1) {
		rc = p15_oid_str_eq(P15_OID_HMAC_GOST_2012, &authdata->macAlgorithm.algorithm);
		if (!rc) hmac2012 = 1;
	}
	if (rc > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_load_token:p15_oid_str_eq returned error for authdata->macAlgorithm.algorithm");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc) {
		rc = 0;
		no_auth = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, 
"p15_load_token:AuthenticatedData.macAlgorithm.algorithm contains unrecognized OID");
		goto no_auth_clr;
	}
	rc2 = p15_ck_format(ctx, hmac2012, errors);
	if (rc2 == P15_RC_INVALID_CONTAINER_FORMAT) {
		terr = 1;
		p15_add_error(errors, P15_ERROR, rc2, 
"p15_load_token:In AuthenticatedData.macAlgorithm.algorithm: Unexpected HMAC algorithm for this container format.");
	} else if (rc2) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2,
"p15_load_token:AuthenticatedData.macAlgorithm.algorithm: p15_ck_format returned error.");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* authdata->macAlgorithm.parameters */
	if (authdata->macAlgorithm.parameters) {
		if (p15_ck_ANY_NULL(authdata->macAlgorithm.parameters)) {
			p15_add_error(errors, P15_ERROR, P15_RC_UNSUPPORTED_ALGORITHM, 
"p15_load_token:AuthenticatedData.macAlgorithm.parameters are present and not NULL");
			no_auth = 1;
			goto no_auth_clr;
		}
	}

	/* *authdata->unauthAttrs */
	if (authdata->unauthAttrs) {
		warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
			"p15_load_token: AuthenticatedData.unauthAttrs are ignored");
	}

	/* authdata->mac */
	if (authdata->mac.size != (hmac2012 ? 64 : 32)) {
		sprintf(st,"p15_load_token: AuthenticatedData.mac is not %s bytes long",
			hmac2012 ? "64" : "32");
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		no_auth = 1;
		goto no_auth_clr;
	}
	memcpy(mac, authdata->mac.buf, hmac2012 ? 64 : 32);

	/* *authdata->authAttrs */
	if (!authdata->authAttrs) {
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN,
"p15_load_token: AuthenticatedData.authAttrs are absent. Authentication failed");
		no_auth = 1;
		goto no_auth_clr;
	}
	rc = p15_load_aattrs(hash2012, authdata->authAttrs, hash, errors);
	/* ret: 0, 1-warns, int err, no_auth, other errs - terr  */
	switch (rc) {
		case 0: break;
		case 1: warns = 1; rc = 0; break;
		case P15_RC_NO_AUTH_AND_DECRYPT:
			rc = 0;
			no_auth = 1;
			goto no_auth_clr;
		case P15_RC_INTERNAL_ERROR: 
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_token:p15_load_aattrs returned error");
			goto end;
		default:
			terr = 1; rc = 0;
	}

	/* ck hash */
	if (hash2012) {
		rc2 = gostc_hash12_init(&hh12, 0);
		if (!rc2) rc2 = gostc_hash12_do(&hh12, 
			authdata->encapContentInfo.eContent->buf,
			authdata->encapContentInfo.eContent->size);
		if (!rc2) rc2 = gostc_hash12_finish(&hh12, hashc);
		if (rc2) {
			p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_CRYPTOLIB_HASH,
			"p15_load_token:One of gostc_hash12_* functions returned error.");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	} else {
		hh01 = gostc_hash_open();
		gostc_hash_do(hh01, authdata->encapContentInfo.eContent->buf,
			authdata->encapContentInfo.eContent->size);
		gostc_hash_close(hh01, hashc);
	}
	if (memcmp(hashc, hash, hash2012 ? 64 : 32)) {
		p15_add_error(errors, P15_ERROR, P15_RC_NO_AUTH_AND_DECRYPT,
"p15_load_token: In AuthenticatedData: Calculated message digest does not match the digest in authAttrs. Authentication failed");
		no_auth = 1;
		goto no_auth_clr;
	}

	/* re-encode authAttrs */
	rc = p15_asn_encode(&asn_DEF_P15_AuthAttributes, authdata->authAttrs, &ebuf,
		&ebuflen, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
		"p15_load_token:p15_asn_encode returned error for AuthAttributes");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* ck mac */
	for (i = 0; i < 32; i++) key[i] = akey[i] ^ akmsk[i];
	if (hmac2012) {
		rc2 = gostc_hmac12_init(&hm12, key, 32);
		gostc_cleanse(key, 32);
		if (!rc2) rc2 = gostc_hmac12_do(&hm12, ebuf, ebuflen);
		if (!rc2) rc2 = gostc_hmac12_finish(&hm12, macc);
		if (rc2) {
			p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_CRYPTOLIB_HASH,
			"p15_load_token:One of gostc_hmac12_* functions returned error.");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	} else {
		hh01 = gostc_hmac_open(key, 32);
		gostc_cleanse(key, 32);
		gostc_hmac_do(hh01, ebuf, ebuflen);
		gostc_hmac_close(hh01, macc);
	}
	if (memcmp(mac, macc, hmac2012 ? 64 : 32)) {
		p15_add_error(errors, P15_ERROR, P15_RC_NO_AUTH_AND_DECRYPT,
"p15_load_token: In AuthenticatedData: Calculated mac on AuthAttrs does not match AuthenticatedData.mac. Authentication failed");
		no_auth = 1;
		goto no_auth_clr;
	}
	
	goto no_auth_cont;

no_auth_clr:
	p15_free(pwkey);
	p15_free(pwkmsk);
	pwkey = NULL;
	pwkmsk = NULL;
no_auth_cont:
	p15_free(akey);
	p15_free(akmsk);
	akey = NULL;
	akmsk = NULL;

	/* read random container - ptoken->pkcs15Objects[...].DataObjects (random only) */
	if (!pwkey) {
		warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_NO_RANDOM_CONTAINER,
"p15_load_token: Random init value will not be read from the token because of authentication failure");
		goto rng_skipped;
	}
	
	rc = p15_get_random_container(ctx, ptoken, kekid, pwkey, pwkmsk, &rndiv,
		&warns, &terr, errors);
	switch (rc) {
	case 0: break;
	case P15_RC_NO_RANDOM_CONTAINER:
		warns = 1;
		p15_add_error(errors, P15_WARNING, rc, 
			"p15_load_token: No random init value container found in the token");
		rc = 0;
		goto rng_skipped;
	case P15_RC_NO_AUTH_AND_DECRYPT:
		warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_NO_RANDOM_CONTAINER, 
			"p15_load_token: Decryption of random init value container failed");
		no_auth = 1;
		rc = 0;
		goto rng_skipped;
	default:
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_token:p15_get_random_container returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* init random */
	if (!ctx->rnd_init) {
		rc = p15_init_random(ctx, rndiv->randomInit.buf, 
			rndiv->randomInit.buf + 32);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_load_token:p15_init_random returned error");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* load extra random */
	p15_free(ctx->rnd_extra);
	ctx->rnd_extra = NULL;
	ctx->rnd_extra_len = 0;
	ctx->rnd_new_extra_len = 0;
	if (rndiv->moreRandom) {
		ctx->rnd_extra_len = rndiv->moreRandom->size;
		ctx->rnd_extra = p15_malloc(ctx->rnd_extra_len);
		memcpy(ctx->rnd_extra, rndiv->moreRandom->buf, ctx->rnd_extra_len);
		ctx->rnd_new_extra_len = ctx->rnd_extra_len; 
		/* to write back the number of bytes when save_token after load_token */
	}

rng_skipped:
	/* ck if rng init */
	if (!ctx->rnd_init) {
		if (wrong_pass) goto wrong_pass_no_random;
		rc = P15_RC_RNG_NOT_INITIALIZED;
		p15_add_error(errors, P15_ERROR, rc,
"p15_load_token: Random Generator is not pre-initialized and no RandomInitValue object found in the container");
		goto end;
	}

	/* remask a key */
	if (pwkey) {
		rc = p15_get_random(ctx, key, 32);
		for (i = 0; i < 32; i++) {
			pwkey[i] ^= key[i];
			pwkmsk[i] ^= key[i];
		}
		gostc_cleanse(key, 32);
	}

	/* ptoken->pkcs15Objects (except random) */
	rc = p15_load_objects(ctx, ptoken, pwkey, pwkmsk, kekid, &warns, &terr, errors);
	if (rc == P15_RC_NO_AUTH_AND_DECRYPT) { rc = 0; no_auth = 1;}
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_token:p15_load_objects returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
	/* Checking if Factor-TS version detected */
	if (!ctx->fts_ver_detected) {
		p15_add_error(errors, P15_WARNING, P15_RC_INV_FACTOR_TS_VERSION, 
"p15_load_token: Factor-TS version object not found. Possibly incompatible token.");
		warns = 1;
	}

wrong_pass_no_random:
	/* Setting remask flag for private keys */
	rc = p15_remask_setup(ctx, NULL, NULL, 1, 0, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_load_token:p15_remask_setup returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	if (!rc && wrong_pass) {
		rc = P15_RC_WRONG_PASSWORD;
		p15_add_error(errors, P15_ERROR, rc, NULL);
	}
	if (!rc && no_auth) {
		rc = P15_RC_NO_AUTH_AND_DECRYPT;
		p15_add_error(errors, P15_ERROR, rc, NULL);
	}

	if (!rc && terr) {
		rc = P15_RC_INVALID_TOKEN;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_load_token: The token contains some erroneous or unsupported structures");
	}
end:
	gostc_hash12_clear(&hh12);
	gostc_hmac12_clear(&hm12);
	gostc_cleanse(key, 32);
	p15_free(akey);
	p15_free(akmsk);
	p15_free(pwkey);
	p15_free(pwkmsk);
	p15_free(ebuf);
	p15_free_id(&kekid);
	ASN_STRUCT_FREE(asn_DEF_P15_AuthenticatedData, authdata);
	ASN_STRUCT_FREE(asn_DEF_P15_PKCS15Token, ptoken);
	ASN_STRUCT_FREE(asn_DEF_P15_RandomInitValue, rndiv);

	if (!rc && warns) rc = 1;
	return rc;
}
/****************************************************************************/
static int p15_pack_bitstring(BIT_STRING_t* bstr)
/* ret: int err */
{
	int rc = 0;
	uint8_t m;

	if (!bstr) {rc = P15_RC_INTERNAL_ERROR; goto end; }

	if (!bstr->size) {
		bstr->bits_unused = 0;
		goto end;
	}
	if (!bstr->buf || bstr->bits_unused > 7 || (bstr->bits_unused < 0)) { 
		rc = P15_RC_INTERNAL_ERROR; 
		goto end; 
	}

	/* zeroing already unused bits */
	if (bstr->bits_unused) {
		m = 0xFF;
		m <<= bstr->bits_unused;
		bstr->buf[bstr->size - 1] &= m;
	}

	/* refining bytes */
	do {
		m = bstr->buf[bstr->size - 1];
		if (m) { /* calculating unused bits */
			bstr->bits_unused = 0;
			while (!(m & 1)) {
				bstr->bits_unused++;
				m >>= 1;
			}
			break;
		} else { /* eating trailing zero byte */
			bstr->size--;
			bstr->bits_unused = 0;
		}
	} while (bstr->size);

	if (!bstr->size) {
		p15_free(bstr->buf);
		bstr->buf = NULL;
	}
end:
	return rc;
}
/****************************************************************************/
static int p15_unpack_bitstring(BIT_STRING_t* bstr, int size, int ubits)
/* ret: int err, inv bit str*/
{
	int rc = 0;
	uint8_t* buf = NULL;

	if (!bstr || size < 0 || ubits < 0 || ubits > 7) {
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (	(bstr->size > size) || 
		(bstr->size == size && bstr->bits_unused < ubits)) { 

		rc = P15_RC_INVALID_BIT_STRING; 
		goto end; 
	}
	if (!bstr->size) {
		p15_free(bstr->buf);
		bstr->buf = p15_malloc(size);
		bstr->size = size;
		bstr->bits_unused = ubits;
		goto end;
	}
	if (!bstr->buf) {
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	buf = p15_malloc(size);
	memcpy(buf, bstr->buf, bstr->size);
	p15_free(bstr->buf);
	bstr->buf = buf;
	bstr->size = size;
	bstr->bits_unused = ubits;
end:
	return rc;
}
/****************************************************************************/
int p15_set_format(p15_ctx_t* ctx, int format)
{
	p15_object_t* o = NULL;
	p15_private_key_t* prkey;
	p15_public_key_t* pubkey;
	int paramset;

	if (!ctx || format < P15_FORMAT_ANY || format > P15_FORMAT_2012) 
		return P15_RC_INVALID_ARGS;
	
	if (format == P15_FORMAT_2012) { ctx->p15_format = format; return 0; }

	/* Checking presence of 2012 pkeys */
	for (o = ctx->objects; o; o = o->next) {

		if (o->type == P15_OT_SECKEY) 
			return P15_RC_2012_KEYS_EXIST;
		
		if (o->type == P15_OT_PRIVKEY) {
			prkey = (p15_private_key_t*)o->obj;
			paramset = prkey->key_parameters;
		} else if (o->type == P15_OT_PUBKEY) {
			pubkey = (p15_public_key_t*)o->obj;
			paramset = pubkey->key_parameters;
		} else continue;

		if (p15_pkey_is_2012(paramset)) 
			return P15_RC_2012_KEYS_EXIST;
	}
	ctx->p15_format = format;
	return 0;
}
/****************************************************************************/
static int p15_pkey_is_2012(int keypar)
{
	if (keypar >= P15_PKIPAR_TC26_512_TEST) return 1;
	return 0;
}
/****************************************************************************/
static int p15_ck_format(p15_ctx_t* ctx, int is2012, p15_info_t** errors)
/* ret: 0, inv_cont_format, int_err */
{
	int rc = 0;

	if (!ctx) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS, 
			"p15_ck_format");
		return P15_RC_INTERNAL_ERROR;
	}
	switch (ctx->p15_format) {

	case P15_FORMAT_ANY:
		if (is2012) ctx->p15_format = P15_FORMAT_2012;
		else ctx->p15_format = P15_FORMAT_2001;
		break;

	case P15_FORMAT_2001:
		if (!is2012) break;
		rc = P15_RC_INVALID_CONTAINER_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_ck_format: The container format has been forced to P15-2001, but a P15-2012 construct was met in the token");
		break;

	case P15_FORMAT_2012:
		if (is2012) break;
		rc = P15_RC_INVALID_CONTAINER_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_ck_format: The container format has been forced to P15-2012, but a P15-2001 construct was met in the token");
		break;
	
	default:
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_ck_format: Unexpected value of ctx->p15_container");
	}

	return rc;
}
/****************************************************************************/
int p15_add_secret_key(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	p15_info_t** errors,
	int mask_count, uint8_t* key, ...)
{
	va_list masks;
	uint8_t** mask_ar = NULL;
	int i;
	int rc = 0;
	uint8_t* mask;

	if (!ctx || key_params < P15_89UZ_CPRO_A || !key ||
			key_params > P15_89UZ_CPRO_Z ||
			mask_count < 1 || mask_count > P15_KEY_MASKS_MAX) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

	if (ctx->p15_format == P15_FORMAT_2001) {
		rc = P15_RC_2012_KEY_VS_2001_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_add_secret_key: Unable to add a GOST 28147-89 secret key to P15-2001 container. (Feature not implemented).");
		goto end;
	}

	mask_ar = p15_malloc(mask_count * sizeof(uint8_t*));
	mask_ar[0] = key;
	if (mask_count > 1) {
		va_start(masks, key);
		for (i = 1; i < mask_count; i++) {
			mask = va_arg(masks, uint8_t*);
			if (!mask) {
				va_end(masks);
				rc = P15_RC_INVALID_ARGS;
				p15_add_error(errors, P15_ERROR, rc, 
"p15_add_secret_key: One of the masks is NULL pointer");
				goto end;
			}
			mask_ar[i] = mask;
		}
		va_end(masks);
		mask = NULL;
	}
	rc = p15_add_secret_key2(ctx, id, newid, label, key_usage, key_access_flags, 
		key_params, start_date, end_date, mask_count, mask_ar, errors);
end:
	p15_free(mask_ar);
	return rc;	
}
/****************************************************************************/
int p15_add_secret_key2(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	int mask_count, uint8_t** key_masks,
	p15_info_t** errors)
{
	int rc = 0;
	p15_secret_key_t* skey = NULL;
	int l, i;
	p15_utctime_t* tm;
	int rc2;

	if (!ctx || key_params < P15_89UZ_CPRO_A || !key_masks ||
			key_params > P15_89UZ_CPRO_Z ||
			mask_count < 1 || mask_count > P15_KEY_MASKS_MAX) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_ERROR, rc, NULL);
		return rc;
	}

	if (ctx->p15_format == P15_FORMAT_2001) {
		rc = P15_RC_2012_KEY_VS_2001_FORMAT;
		p15_add_error(errors, P15_ERROR, rc, 
"p15_add_secret_key: Unable to add a GOST 28147-89 secret key to P15-2001 container. (Feature not implemented)");
		goto end;
	}

#ifdef TESTOUT
	printf("p15_add_secret_key: Adding key '%s':\n", label ? label: "");
	printf("key usage: ");
	p15_print_key_usage(key_usage);
	printf("\nkey access: ");
	p15_print_key_access(key_access_flags);
	printf("\n");
	printf("key parameters (1-5 - cproA,B,C,D,Z): %d\n", 
		key_params);
	if (start_date) {
		printf("start date: %d-%.2d-%.2d %d:%.2d:%.2d\n", 
			start_date->year,
			start_date->month,
			start_date->day,
			start_date->hours,
			start_date->minutes,
			start_date->seconds);
	} else printf("start date: absent\n");
	if (end_date) {
		printf("end date: %d-%.2d-%.2d %d:%.2d:%.2d\n", 
			end_date->year,
			end_date->month,
			end_date->day,
			end_date->hours,
			end_date->minutes,
			end_date->seconds);
	} else printf("end date: absent\n");
	if (mask_count > 1) printf("key: masked\n");
	else {
		printf("key (little-endian):\n");
		p15_dump(key_masks[0], 32);
	}
	printf("\n");
#endif

	/* Checking ID */
	if (id) {
		rc2 = p15_ck_id_exist(ctx, id);
		if (rc2 > 1) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc2, 
				"p15_add_secret_key: p15_ck_id_exist returned error.");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc2) {
			rc = P15_RC_ID_IN_USE;
			p15_add_error(errors, P15_ERROR, rc,
				"p15_add_secret_key: Given ID is already in use in this ctx");
			goto end;
		}
	}

	skey = p15_malloc(sizeof(p15_secret_key_t));

	if (label) {
		l = strnlen(label, P15_LABEL_MAX + 1);
		if (l > P15_LABEL_MAX) {
			rc = P15_RC_LABEL_TOO_LONG;
			p15_add_error(errors, P15_ERROR, rc, NULL);
			goto end;
		}
		strcpy(skey->label, label);
	}

	skey->mask_count = mask_count;
	l = sizeof(uint8_t*) * mask_count;
	skey->key_masks = p15_malloc(l);
	for (i = 0; i < mask_count; i++) 
		skey->key_masks[i] = p15_malloc(P15_PAGE);
	for (i = 0; i < mask_count; i++) 
		memcpy(skey->key_masks[i], key_masks[i], 32);

	skey->key_usage = key_usage;
	skey->key_parameters = key_params;
	key_access_flags &= P15_KA_MASK;
	if (key_access_flags > 1) 
		key_access_flags &= 0xFE;
	skey->key_access_flags = key_access_flags;

	if (start_date) {
		tm = start_date;
		skey->start_date = p15_new_time(tm->year, tm->month, tm->day,
			tm->hours, tm->minutes, tm->seconds);
	}

	if (end_date) {
		tm = end_date;
		skey->end_date = p15_new_time(tm->year, tm->month, tm->day,
			tm->hours, tm->minutes, tm->seconds);
	}

	/* Handling IDs */
	if (p15_set_node_id(ctx, &skey->id, id, newid)) {
		rc = P15_RC_INTERNAL_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_add_private_key: error in p15_set_node_id");
		goto end;
	}

	/* Inserting */
	p15_add_object(ctx, P15_OT_SECKEY, skey);
	skey = NULL;

	/* Forcing 2012 format */
	ctx->p15_format = P15_FORMAT_2012;
end:
	p15_free_secret_key(&skey);
	return rc;
}
/****************************************************************************/
static int p15_ck_id_exist(p15_ctx_t* ctx, p15_id_t* id)
/* 0-not exist(ok), 1-exists(bad), >1 - P15_RC_ const */
{
	p15_object_t* obj = NULL;
	p15_id_t* id2 = NULL;
	int rc2;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;
	
	obj = ctx->objects;
	while (obj) {
		id2 = NULL;
		switch (obj->type) {
		case P15_OT_PRIVKEY:
			id2 = &((p15_private_key_t*)obj->obj)->id;
			break;

		case P15_OT_PUBKEY:
			id2 = &((p15_public_key_t*)obj->obj)->id;
			break;

		case P15_OT_CERT:
		case P15_OT_TRUSTED_CERT:
			id2 = &((p15_certificate_t*)obj->obj)->id;
			break;

		case P15_OT_CRL:
			id2 = &((p15_crl_t*)obj->obj)->id;
			break;

		case P15_OT_SECKEY:
			id2 = &((p15_secret_key_t*)obj->obj)->id;
			break;

		case P15_OT_OIDDO:
			break;

		default: 
			return P15_RC_INTERNAL_ERROR;
		}
		if (id2) {
			rc2 = p15_id_cmp(id, id2);
			if (!rc2) return 1; /* id not unique */
			if (rc2 > 1) return P15_RC_INTERNAL_ERROR;
		}
		obj = obj->next;
	}
	return 0;
}
/****************************************************************************/
int p15_rm_secret_key(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_secret_key_t *key;
	int found = 0;
	p15_object_t *o, *o2;

	if (!ctx || !id) return P15_RC_INVALID_ARGS;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_SECKEY) {
			key = (p15_secret_key_t*)o->obj;
			if (!p15_id_cmp(&key->id, id)) {
				found = 1;
				o2 = o->next;
				p15_rm_object(ctx, o);
				o = o2;
				continue;
			}
		}
		o = o->next;
	}
	if (found) return 0;
	return P15_RC_SECRET_KEY_NOT_FOUND;
}
/****************************************************************************/
p15_secret_key_t* p15_get_secret_key(p15_ctx_t* ctx, p15_id_t* id)
{
	p15_secret_key_t* key;
	p15_object_t* o;

	if (!ctx || !id) return NULL;

	o = ctx->objects;
	while (o) {
		if (o->type == P15_OT_SECKEY) {
			key = (p15_secret_key_t*)o->obj;
			if (!p15_id_cmp(id, &key->id)) return key;
		}
		o = o->next;
	}
	return NULL;
}
/****************************************************************************/
p15_secret_key_t* p15_get_next_secret_key(p15_ctx_t* ctx, p15_object_t** it)
{
	p15_object_t* o;

	if (!ctx || !it) return NULL;
	o = *it;
	if (!o) 
		o = ctx->objects;
	else
		o = o->next;
	while (o) {
		if (o->type == P15_OT_SECKEY) 
			break;
		o = o->next;
	}
	*it = o;
	if (o) return (p15_secret_key_t*)o->obj;
	return NULL;
}
/****************************************************************************/
p15_secret_key_t* p15_copy_secret_key(p15_secret_key_t* ok)
{
	p15_secret_key_t* nk = NULL;
	unsigned int i;

	if (!ok) return NULL;
	nk = p15_malloc(sizeof(p15_secret_key_t));
	if (ok->label[0])
		strcpy(nk->label, ok->label);
	nk->id = ok->id;
	if (!ok->key_masks || !ok->mask_count) goto err;
	nk->mask_count = ok->mask_count;
	nk->to_remask = ok->to_remask;
	nk->key_masks = p15_malloc(ok->mask_count * sizeof(uint8_t*));
	for (i = 0; i < ok->mask_count; i++) {
		nk->key_masks[i] = p15_malloc(P15_PAGE);
		memcpy(nk->key_masks[i], ok->key_masks[i], 32);
	}
	nk->key_usage = ok->key_usage;
	nk->key_access_flags = ok->key_access_flags;
	nk->key_parameters = ok->key_parameters;

	if (ok->start_date) {
		nk->start_date = p15_new_time(ok->start_date->year, ok->start_date->month, 
			ok->start_date->day, ok->start_date->hours, ok->start_date->minutes, 
			ok->start_date->seconds);
		if (!nk->start_date) goto err;
	}
	if (ok->end_date) {
		nk->end_date = p15_new_time(ok->end_date->year, ok->end_date->month, 
			ok->end_date->day, ok->end_date->hours, ok->end_date->minutes, 
			ok->end_date->seconds);
		if (!nk->end_date) goto err;
	}

	return nk;

err:
	p15_free_secret_key(&nk);
	return NULL;	
}
/****************************************************************************/
void p15_free_secret_key(p15_secret_key_t** pkey)
{
	p15_secret_key_t* key;
	uint8_t** pkm;
	size_t i;

	if (!pkey) return;
	key = *pkey;
	if (!key) return;
	
	if (key->key_masks) {
		pkm = key->key_masks;
		for (i = 0; i < key->mask_count; i++) {
			p15_free(pkm[i]);
		}
		p15_free(pkm);
	}
	p15_free_time(&key->start_date);
	p15_free_time(&key->end_date);

	p15_free(key);
	*pkey = NULL;
}
/****************************************************************************/
static int p15_mk_seckey_obj(p15_ctx_t* ctx, P15_SecretKeys_t* skeys,
	p15_secret_key_t* sk, uint8_t* pwkey, uint8_t* pwkmsk, int* warns,
	p15_info_t** errors)
/* ret int err, P15_RC_NO_SECRET_KEY_DATA */
{
	int rc = 0;
	P15_SecretKeyType_t* skt = NULL;
	P15_GostSecretKey_t* gsk = NULL;
	P15_PKCS15ObjectGostSecretKey_t* ogsk = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_CommonKeyAttributes_t* cka = NULL;
	P15_ObjectValueGostSecretKey_t* ovgsk = NULL;
	P15_EnvelopedData_t* edata = NULL;
	P15_GostSecretKeyValue_t* gskv = NULL;
	P15_KeyInfoSecretGost_t* ki = NULL;
	P15_GostSecretKeyParameters_t* pars = NULL;
	char st[4096];
	int i;
	uint8_t* ebuf = NULL;
	size_t ebuflen = 0;
	uint8_t ka = 0;
	uint8_t* keysbuf = NULL;

	if (!ctx || !skeys || !sk || !pwkey || !pwkmsk || !warns) {
		rc = P15_RC_INVALID_ARGS;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, "p15_mk_seckey_obj");
		rc = P15_RC_INTERNAL_ERROR;
		return rc;
	}

	if (!sk->key_masks || sk->mask_count < 1) rc = 1;
	if (!rc) { if (!sk->key_masks[0]) rc = 1; }
	if (rc)	{
		rc = P15_RC_NO_SECRET_KEY_DATA;
		strcpy(st, "p15_mk_seckey_obj: No key/mask material set for secret key '");
		strcat(st, sk->label[0] ? sk->label : "NO_LABEL");
		strcat(st, "'");
		p15_add_error(errors, P15_ERROR, rc, st);
		return rc;
	}

	/* remasking */
	if (sk->to_remask) {
		rc = p15_remask_secret_key(ctx, sk, sk->mask_count);
		if (rc) {
			sprintf(st, 
"p15_mk_seckey_obj: p15_remask_secret_key returned error for secret key '%s'", 
				sk->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			return rc;
		}
	}

#ifdef TESTOUT
	printf("p15_mk_seckey_obj: Making secret key '%s' object:\n", sk->label);
	printf("Choosing SecretKeyType.gostKey choice.\n");
#endif
	skt = p15_malloc(sizeof(P15_SecretKeyType_t));
	skt->present = P15_SecretKeyType_PR_gostKey;
	gsk = &skt->choice.gostKey;

	/* gsk->keyTypeGost */
#ifdef TESTOUT
	printf("p15_mk_seckey_obj: Setting gostKey.keyTypeGost to %s\n", P15_OID_GOST_89);
#endif
	rc = p15_str_to_oid(P15_OID_GOST_89, &gsk->keyTypeGost);
	if (rc) {
		sprintf(st,
"p15_mk_seckey_obj: p15_str_to_oid(GOST_89) returned error for secret key '%s'",
			sk->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* gsk->keyAttr */
	ogsk = &gsk->keyAttr;

	/* ogsk->commonObjectAttributes */
	coa = &ogsk->commonObjectAttributes;

	/* *coa->label */
	if (sk->label[0]) {
		coa->label = p15_malloc(sizeof(P15_Label_t));
		if (OCTET_STRING_fromBuf(coa->label, sk->label, strlen(sk->label))) {
			rc = P15_RC_ASN1C_ERROR;
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_seckey_obj: OCTET_STRING_fromBuf error for label");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("gostKey.keyAttr.commonObjectAttributes.label:\n");
		p15_dump(coa->label->buf, coa->label->size);
#endif
	}
#ifdef TESTOUT
	printf("gostKey.keyAttr.commonObjectAttributes.flags: 0x80, len=1, unused_bits=6\n");
#endif

	/* *coa->flags */
	coa->flags = p15_malloc(sizeof(BIT_STRING_t));
	coa->flags->buf = p15_malloc(1);
	coa->flags->buf[0] = 0x80; /* private, non-modifiable */
	coa->flags->size = 1;
	coa->flags->bits_unused = 6;
	rc = p15_pack_bitstring(coa->flags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_seckey_obj:p15_pack_bitstring: error for coa->flags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* ogsk->classAttributes CommonKeyAttributes, */
	cka = &ogsk->classAttributes;

	/* cka->iD */
	rc = p15_id_to_ostr(&sk->id, &cka->iD);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_seckey_obj:p15_id_to_ostr returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* cka->usage */
	cka->usage.buf = p15_malloc(2);
	cka->usage.size = 2;
	cka->usage.bits_unused = 6;
	cka->usage.buf[0] = sk->key_usage >> 8;
	cka->usage.buf[1] = (uint8_t)sk->key_usage;
	rc = p15_pack_bitstring(&cka->usage);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_seckey_obj:p15_pack_bitstring: error for cka->usage");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("gostKey.keyAttr.classAttributes.iD:\n");
	p15_dump(cka->iD.buf, cka->iD.size);
	printf("gostKey.keyAttr.classAttributes.usage: len=2, unused_bits=6:\n");
	p15_dump(cka->usage.buf, cka->usage.size);
#endif
	/* *cka->accessFlags */
	if (!sk->key_access_flags) goto start_date;
	ka = sk->key_access_flags;
	ka &= P15_KA_MASK & 0xFE;
	cka->accessFlags = p15_malloc(sizeof(P15_KeyAccessFlags_t));
	cka->accessFlags->buf = p15_malloc(1);
	cka->accessFlags->size = 1;
	cka->accessFlags->bits_unused = 3;
	cka->accessFlags->buf[0] = ka;
	rc = p15_pack_bitstring(cka->accessFlags);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_seckey_obj:p15_pack_bitstring: error for cka->accessFlags");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

#ifdef TESTOUT
	printf("gostKey.keyAttr.classAttributes.accessFlags: len=1, unused_bits=3: 0x%.2X\n", ka);
#endif

	/* *cka->startDate */
start_date:
	if (sk->start_date) {
		cka->startDate = p15_malloc(sizeof(GeneralizedTime_t));
		rc = p15_utctime_to_gt(cka->startDate, sk->start_date);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_seckey_obj:p15_utctime_to_gt returned error for sk->start_date");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("gostKey.keyAttr.classAttributes.startDate:\n");
		p15_dump(cka->startDate->buf, cka->startDate->size);
#endif
	}

	/* *cka->endDate */
	if (sk->end_date) {
		cka->endDate = p15_malloc(sizeof(GeneralizedTime_t));
		rc = p15_utctime_to_gt(cka->endDate, sk->end_date);
		if (rc) {
			p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_seckey_obj:p15_utctime_to_gt returned error for sk->end_date");
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
#ifdef TESTOUT
		printf("gostKey.keyAttr.classAttributes.endDate:\n");
		p15_dump(cka->endDate->buf, cka->endDate->size);
#endif
	}

	/* ogsk->subClassAttributes - absent */ 
	
	/* ogsk->typeAttributes P15-GostSecretKeyAttributes */

	/* ovgsk */
	ovgsk = &ogsk->typeAttributes.value;

	ovgsk->present = P15_ObjectValueGostSecretKey_PR_direct_protected;
	edata = &ovgsk->choice.direct_protected;
	
	/* GostSecretKeyValue */
	gskv = p15_malloc(sizeof(P15_GostSecretKeyValue_t));
	keysbuf = p15_malloc(sk->mask_count << 5);

	for (i = 0; i < (int)sk->mask_count; i++) 
		memcpy(keysbuf + (i << 5), &sk->key_masks[i][0], 32);
	if (OCTET_STRING_fromBuf(gskv, (char*)keysbuf, sk->mask_count << 5)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_seckey_obj: OCTET_STRING_fromBuf failed for keysbuf");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* Encoding */
	rc = p15_asn_encode(&asn_DEF_P15_GostSecretKeyValue, gskv, &ebuf, &ebuflen, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
			"p15_mk_seckey_obj:p15_asn_encode returned error for gskv");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
#ifdef TESTOUT
	printf("Encoded GostSecretKeyValue (%d bytes):\n", (int)ebuflen);
	p15_dump(ebuf, ebuflen);
	printf(
"Wrapping GostSecretKeyValue to EnvelopedData (gostKey.keyAttr.typeAttributes.value.direct-protected:\n");
#endif

	/* mk envdata */
	rc = p15_mk_enveloped_data(ctx, edata, ebuf, ebuflen, pwkey, pwkmsk, errors);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_seckey_obj: p15_mk_enveloped_data failed for gskv");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* ki - ogsk->typeAttributes */
	ki = p15_malloc(sizeof(P15_KeyInfoSecretGost_t));
	ogsk->typeAttributes.keyInfo = ki;

#ifdef TESTOUT
	printf("Making gostKey.keyAttr.typeAttributes.keyInfo.paramsAndOps.parameters:\n");
	printf("pars.cryptoProParamSet (1-4 - cproA,B,C,D): %d\n", sk->key_parameters);
	printf("\n");
#endif
	ki->present = P15_KeyInfoSecretGost_PR_paramsAndOpsSKey;
	pars = &ki->choice.paramsAndOpsSKey.parameters;
	pars->present = P15_GostSecretKeyParameters_PR_cryptoProParamSet;

	/* pars->CryptoProParamSet */
	
	rc = p15_encuz_to_oid(sk->key_parameters, &pars->choice.cryptoProParamSet);
	if (rc) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc,
"p15_mk_seckey_obj:p15_encuz_to_oid failed for pars->cryptoProParamSet");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	if (asn_sequence_add(&skeys->choice.objectss.list, skt)) {
		rc = P15_RC_ASN1C_ERROR;
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, 
			"p15_mk_seckey_obj: asn_sequence_add failed for skt");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	skt = NULL;

end:
	ASN_STRUCT_FREE(asn_DEF_P15_SecretKeyType, skt);
	ASN_STRUCT_FREE(asn_DEF_P15_GostSecretKeyValue, gskv);
	p15_free(keysbuf);
	p15_free(ebuf);
	return rc;
}
/****************************************************************************/
static int p15_process_secret_key_object(p15_ctx_t* ctx, void* skobj,
	uint8_t* pwkey, uint8_t* pwkmsk, p15_id_t* kekid, int* warns, int* errs, 
	p15_info_t** errors)
/*ret: int err, no auth */
{
	int rc = 0;
	int rc2;
	uint8_t* ebuf = NULL;
	int ebuflen;
	char st[4096];
	int no_auth = 0;
	int kmasks = 0;
	uint8_t** key_masks = NULL;
	int i;
	p15_secret_key_t* key = NULL;
	P15_SecretKeyType_t* skt = skobj;
	P15_GostSecretKey_t* gsk = NULL;
	P15_CommonObjectAttributes_t* coa = NULL;
	P15_PKCS15ObjectGostSecretKey_t* ogsk = NULL;
	P15_CommonKeyAttributes_t* cka = NULL;
	P15_GostSecretKeyAttributes_t* gska = NULL;
	P15_KeyInfoSecretGost_t* ki = NULL;
	P15_GostSecretKeyParameters_t* pars = NULL;
	P15_ObjectValueGostSecretKey_t* ovgsk = NULL;
	P15_EnvelopedData_t* edata = NULL;
	P15_GostSecretKeyValue_t* gskv = NULL;

	if (!ctx || !skobj || !warns || !errs || (!pwkey && pwkmsk) ||
		(pwkey && !pwkmsk) || (pwkey && !kekid)) {
		p15_add_error(errors, P15_INTERNAL_ERROR, P15_RC_INVALID_ARGS,
		"p15_process_secret_key_object");
		return P15_RC_INTERNAL_ERROR;
	}

	key = p15_malloc(sizeof(p15_secret_key_t));

	if (skt->present != P15_SecretKeyType_PR_gostKey) {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_process_secret_key_object:Unsupported secret key type. Skipping the object");
		goto end;
	}
	rc2 = p15_ck_format(ctx, 1, errors);
	if (rc2 == P15_RC_INVALID_CONTAINER_FORMAT) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, rc2, 
"p15_process_secret_key_object: SecretKeyType object is supported only for P15-2012 container format");
	} else if (rc2) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, 
			"p15_process_secret_key_object: p15_ck_format returned error");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	gsk = &skt->choice.gostKey;

	/* gsk->keyTypeGost */
	rc2 = p15_oid_str_eq(P15_OID_GOST_89, &gsk->keyTypeGost);
	if (rc2 > 1) {
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2,
"p15_process_secret_key_object: p15_oid_str_eq returned error for gsk->keyTypeGost");
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*warns = 1;
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE,
"p15_process_secret_key_object:Unsupported SecretKeyType.gostKey.keyTypeGost OID. Skipping the object");
		goto end;
	}

	/* gsk->keyAttr */

	ogsk = &gsk->keyAttr;

	/* ogsk->commonObjectAttributes */

	coa = &ogsk->commonObjectAttributes;

	/* *coa->Label */
	if (!coa->label) goto no_label;
	if (!coa->label->buf || !coa->label->size) goto no_label;
	if (coa->label->size > P15_LABEL_MAX) {
		*errs = 1;
		p15_add_error(errors, P15_ERROR, P15_RC_LABEL_TOO_LONG,
"p15_process_secret_key_object:Object label is too long. Skipping the label");
		goto no_label;
	}
	memcpy(key->label, coa->label->buf, coa->label->size);

no_label:
	/* *coa->flags */
	if (!coa->flags) {
		*warns = 1;
		sprintf(st, 
"p15_process_secret_key_object: In secret key '%s':commonObjectAttributes.flags: No object flags", 
			key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}
	rc2 = p15_unpack_bitstring(coa->flags, 1, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':p15_unpack_bitstring returned error for coa->flags", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	if (rc2) {
		*warns = 1;
		sprintf(st, 
"p15_process_secret_key_object: In secret key '%s':commonObjectAttributes.flags: Unrecoginzed flag format. Skipping flags", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto no_flags;
	}
	if ((coa->flags->buf[0] & 0xC0) != 0x80) {
		*warns = 1;
		sprintf(st, 
"p15_process_secret_key_object: In secret key '%s':commonObjectAttributes.flags: Incorrect object flags. Should be private and non-modifiable", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_INV_OBJ_FLAGS, st);
		goto no_flags;
	}
no_flags:
	/* *coa->authId */
	if (coa->authId) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':commonObjectAttributes: authId is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->userConsent */
	if (coa->userConsent) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':commonObjectAttributes: userConsent is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *coa->accessControlRules  */
	if (coa->accessControlRules) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':commonObjectAttributes: accessControlRules is not supported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}


	/* ogsk->classAttributes */
	cka = &ogsk->classAttributes;

	/* cka->iD */
	rc = p15_process_obj_id(ctx, &cka->iD, P15_OT_SECKEY, &key->id);
	switch (rc) {
	case 0:
		break;
	case P15_RC_INVALID_TOKEN:
		*errs = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.iD: iD has invalid format. Skipping the object", key->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		goto end;
		
	case P15_RC_ID_IN_USE:
		*errs = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.iD: iD is not unique. Skipping the object", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		rc = 0;
		goto end;

	default:
		sprintf(st, 
"p15_process_secret_key_object: In secret key '%s':p15_process_obj_id returned error",
			key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	/* cka->usage */
	rc2 = p15_unpack_bitstring(&cka->usage, 2, 6);
	if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':p15_unpack_bitstring returned error for cka->usage", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	if (rc2) {
		*errs = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.usage: Unrecognized key usage format. Ignoring the flags", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
	} else {
		key->key_usage = cka->usage.buf[0];
		key->key_usage <<= 8;
		key->key_usage |= (uint8_t)cka->usage.buf[1];
	}

	/* *cka->startDate */
	if (cka->startDate) {
		rc = p15_gt_to_utctime(cka->startDate, &key->start_date);
		if (rc == P15_RC_INVALID_TOKEN) {
			*errs = 1;
			sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.startDate: Invalid Generalized time. Ignoring the field", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			rc = 0;
		} else if (rc) {
			sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.startDate:p15_gt_to_utctime returned error", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* *cka->endDate */
	if (cka->endDate) {
		rc = p15_gt_to_utctime(cka->endDate, &key->end_date);
		if (rc == P15_RC_INVALID_TOKEN) {
			*errs = 1;
			sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.endDate: Invalid Generalized time. Ignoring the field", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
			rc = 0;
		} else if (rc) {
			sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.endDate:p15_gt_to_utctime returned error", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
	}

	/* *cka->accessFlags */
	if (cka->accessFlags) {
		rc2 = p15_unpack_bitstring(cka->accessFlags, 1, 3);
		if (rc2 && rc2 != P15_RC_INVALID_BIT_STRING) {
			sprintf(st,
"p15_process_secret_key_object: In secret key '%s':p15_unpack_bitstring returned error for cka->accessFlags", key->label);
			p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
			rc = P15_RC_INTERNAL_ERROR;
			goto end;
		}
		if (rc2) {
			*errs = 1;
			sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.accessFlags: Unrecognized format of the field. Ignoring the flags", key->label);
			p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		} else {
			key->key_access_flags = cka->accessFlags->buf[0];
			key->key_access_flags &= P15_KA_MASK;
			if (!key->key_access_flags) key->key_access_flags = 1;
		}
	}

	/* unsupported fields */

	if (cka->native) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.native: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}


	if (cka->keyReference) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':classAttributes.keyReference: The field is unsupported and ignored", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* *ogsk->subClassAttributes */

	if (ogsk->subClassAttributes) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s': subClassAttributes structure is ignored",
			key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* ogsk->typeAttributes */

	gska = &ogsk->typeAttributes;

	/* *gska->keyInfo */
	ki = gska->keyInfo;
	if (!ki) {
		key->key_parameters = P15_89UZ_CPRO_A;
		goto no_keyinfo;
	}
	if (ki->present != P15_KeyInfoSecretGost_PR_paramsAndOpsSKey) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.keyInfo: Only paramsAndOps choice is supported. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	if (ki->choice.paramsAndOpsSKey.supportedOperations) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.keyInfo.paramsAndOps.supportedOperations: The field is not supported. Skipping the field", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
	}

	/* parameters */
	pars = &ki->choice.paramsAndOpsSKey.parameters;

	if (pars->present != P15_GostSecretKeyParameters_PR_cryptoProParamSet) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.keyInfo.paramsAndOps.parameters: Only cryptoProParamSet choice is supported. Skipping the object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}

	rc2 = p15_oid_to_encuz(&pars->choice.cryptoProParamSet, &key->key_parameters);
	if (rc2 == P15_RC_UNSUPPORTED_ALGORITHM) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.keyInfo.paramsAndOps.parameters.cryptoProParamSet: Unsupported OID. Skipping the object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	if (rc2) {
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.keyInfo.paramsAndOps.parameters.cryptoProParamSet: p15_oid_to_encuz returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc2, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}
	
no_keyinfo:
	/* gska->value */
	ovgsk = &gska->value;

	if (ovgsk->present != P15_ObjectValueGostSecretKey_PR_direct_protected) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.value: Only direct-protected choice is supported. Skipping the key object", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto end;
	}
	edata = &ovgsk->choice.direct_protected;

	/* ck pwkey */
	if (!pwkey) {
		*warns = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.value.direct-protected: Encrypted data will not be decrypted because of previous authentication failure. Skipping the key data", key->label);
		p15_add_error(errors, P15_WARNING, P15_RC_IGNORED_STRUCTURE, st);
		goto no_key_data;
	}

	/* decrypt */
	sprintf(st, "private key '%s':typeAttributes.value.direct-protected", 
		key->label);
	rc = p15_decrypt_enveloped_data(ctx, kekid, st, edata, &ebuf, &ebuflen, 
		pwkey, pwkmsk, warns, errs, errors);
	if (rc == P15_RC_NO_AUTH_AND_DECRYPT) {
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.value.direct-protected: Failed to decrypt the key data", key->label);
		p15_add_error(errors, P15_ERROR, rc, st);
		rc = 0;
		no_auth = 1;
		goto no_key_data;
	}
	if (rc) {
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.value.direct-protected: p15_decrypt_enveloped_data returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		rc = P15_RC_INTERNAL_ERROR;
		goto end;
	}

	/* decode */
	rc = p15_asn_decode(ebuf, ebuflen, &asn_DEF_P15_GostSecretKeyValue, (void**)&gskv, errors);
	if (rc == P15_RC_INTERNAL_ERROR) {
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.value.direct-protected: p15_asn_decode returned error", key->label);
		p15_add_error(errors, P15_INTERNAL_ERROR, rc, st);
		goto end;
	}
	if (rc) {
		*errs = 1;
		no_auth = 1;
		rc = 0;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s':typeAttributes.value.direct-protected: p15_asn_decode returned error", key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto no_key_data;
	}

	/* key data */
	kmasks = gskv->size;
	if (kmasks & 0x1F) {
		*errs = 1;
		no_auth = 1;
		sprintf(st,
"p15_process_secret_key_object: In secret key '%s': KeyValueMask must be multiple of 32",
			key->label);
		p15_add_error(errors, P15_ERROR, P15_RC_INVALID_TOKEN, st);
		goto no_key_data;
	}
	kmasks >>= 5;

	key_masks = p15_malloc(kmasks * sizeof(uint8_t*));
	for (i = 0; i < kmasks; i++) {
		key_masks[i] = p15_malloc(P15_PAGE);
		memcpy(key_masks[i], gskv->buf + (i << 5), 32);
	}
	key->key_masks = key_masks;
	key->mask_count = kmasks;
	key_masks = NULL;

no_key_data:
	p15_add_object(ctx, P15_OT_SECKEY, key);
	key = NULL;
	if (!rc && no_auth) rc = P15_RC_NO_AUTH_AND_DECRYPT;
end:
	if (key_masks) {
		for (i = 0; i < kmasks; i++) p15_free(key_masks[i]);
		p15_free(key_masks);
	}
	p15_free_secret_key(&key);
	p15_free(ebuf);
	ASN_STRUCT_FREE(asn_DEF_P15_GostSecretKeyValue, gskv);
	return rc;
}
/****************************************************************************/

